
package interfaceProofTree;

import java.awt.event.*;
import javax.swing.event.*;
import java.awt.*; // para borderLayout
import javax.swing.*; // JPanel (entre otros)

import proofTree.ProofTree;

public class DDTWindow extends JFrame {
    
     // the window consists of a menu and a SplitPanel 
     JMenuBar menuBar;
     NavigatorPanel panel; 

     ProofTree tree;
     Oracle oracle;
     
     // icons and images  
     ImageIcon nonvalidIcon;
     ImageIcon validIcon;
     ImageIcon buggyIcon;
     ImageIcon dontknowIcon;
     ImageIcon trustedIcon;
     Image ddtImage;
   
   public DDTWindow(String path, String file, String fileEntailment) {
     
     // loading the icons
     loadIcons(path);
     setIconImage(ddtImage);
 
     // nodes counter
     StatusPanel sPanel=new StatusPanel(nonvalidIcon, validIcon, buggyIcon,
                                       dontknowIcon, trustedIcon);


     // proof tree
     tree = new ProofTree(this,path,file,fileEntailment,sPanel);

    // Set the icons for  nodes.
    tree.setCellRenderer(new MyRenderer(nonvalidIcon, validIcon, buggyIcon,
                                         dontknowIcon, trustedIcon));

      // the bottom panel
     oracle = new Oracle(Oracle.NO_STRATEGY, nonvalidIcon, validIcon,
                                             dontknowIcon, trustedIcon,this,tree);


     // the window contains a panel and a menu
     panel =  new NavigatorPanel(tree, sPanel, oracle);

     getContentPane().add( panel, BorderLayout.CENTER);
  
     // menu bar 
     menuBar =  createMenu(tree,oracle);
     setJMenuBar(menuBar);

     // closing event 
     addWindowListener(new WindowAdapter(){ public void windowClosing(WindowEvent evt) { System.exit(0); } } );

   }

//////////////////////////////////// splitpanel

   public void closeOraclePanel() {
     panel.setDividerLocation(panel.getMaximumDividerLocation());
   }

   public void openOraclePanel() { 
     int center = (panel.getMinimumDividerLocation() +
                         panel.getMaximumDividerLocation() )/2;
     panel.setDividerLocation(center);
   }

  

 ////////////////////////////////////// loading icons ////////////////////////////////////////////
 private void loadIcons(String path) {
     MediaTracker tracker;

     // load the icons
     nonvalidIcon = new ImageIcon(path+"/nonvalid.gif");
     validIcon    = new ImageIcon(path+"/valid.gif");
     buggyIcon    = new ImageIcon(path+"/buggy.gif");
     dontknowIcon = new ImageIcon(path+"/dontknow.gif");
     trustedIcon  = new ImageIcon(path+"/trusted.gif");

     // load the main icon
     ddtImage  = Toolkit.getDefaultToolkit().getImage(path+"/ddt.gif");     
     tracker = new MediaTracker(this);
     tracker.addImage(ddtImage,0);
     try { tracker.waitForAll(); } catch(Exception e) { e.printStackTrace();};

     //scale the icons
     trustedIcon.setImage(trustedIcon.getImage().getScaledInstance(12,11,Image.SCALE_SMOOTH));
     buggyIcon.setImage(buggyIcon.getImage().getScaledInstance(18,16,Image.SCALE_SMOOTH));
     dontknowIcon.setImage(dontknowIcon.getImage().getScaledInstance(14,14,Image.SCALE_SMOOTH));

 }

 ////////////////////////////////////////////////// menu 
 private JMenuBar createMenu(ProofTree tree, Oracle oracle ) {
 	 final Boolean bT = new Boolean(true);
 	 final Boolean bF = new Boolean(false);

        JMenuBar menuBar = new JMenuBar();
        //menuBar.setBorder( new BevelBorder(BevelBorder.LOWERED) );

	ActionListener itemListener = new ItemListener(tree,oracle);
	DDTMenuListener  menuListener = new DDTMenuListener(tree,oracle);

	 Object fileMenuData[][] = {
	 	// menu name, short-cut
		{"File",new Character('F'),null},
		// menu name, short-cut, listener, enabled
		{"Open",	new Character('O'),itemListener,bT},
		{"Reload",new Character('R'),itemListener,bT},
		{},
		{"Save",	new Character('S'),itemListener,bT},
		{"Save As",new Character('A'),itemListener,bT},
		{},
		{"Exit",new Character('X'),itemListener,bT}
	};

	 Object nodeMenuData[][] = {
	 	// menu name, short-cut
		{"Node",new Character('N'),menuListener},
		// menu name, short-cut, listener, enabled
		{"Valid",	new Character('V'),itemListener,bT,validIcon},
		{"Non-Valid",new Character('N'),itemListener,bT,nonvalidIcon},
		{"Unknown",	new Character('U'),itemListener,bT,dontknowIcon},
		{"Trusted",new Character('T'),itemListener,bT,trustedIcon},
		{"Buggy",new Character('B'),itemListener,bT,buggyIcon}
	};

	Object treeMenuData[][] = {
	 	// menu name, short-cut
		{"Tree",new Character('T'),null},
		// menu name, short-cut, listener, enabled
		{"Expand Tree",new Character('E'),itemListener,bT},
		{"Collapse Tree",new Character('C'),itemListener,bT},
		{},
		{"Remove Trusted Nodes",new Character('R'),itemListener,bT},
		{"Remove Valid &Trusted Nodes",new Character('V'),itemListener,bT},
		{},
		{"Remove Selected Node Subtree",new Character('S'),itemListener,bT},
		{"Make Selected Node New Root",new Character('M'),itemListener,bT}

	};

	Object selectMenuData[][] = {
	 	// menu name, short-cut
		{"Select",new Character('S'),null},
		// menu name, short-cut, listener, enabled
		{"Deepest Non-Valid Node",new Character('D'),itemListener,bT},
		{"Deepest Desc. with the Same Rule",new Character('R'),itemListener,bT},
		{"Center of the Tree",new Character('C'),itemListener,bT}

	};

	Object infoMenuData[][] = {
	 	// menu name, short-cut
		{"Information",new Character('I'),null},
		// menu name, short-cut, listener, enabled
		{"Selected Node",new Character('N'),itemListener,bT},
		{"Tree",new Character('T'),itemListener,bT},
		{},
		{"History",new Character('H'),itemListener,bT},
		{"Clear History",new Character('C'),itemListener,bT}
	};

	Object strategyMenuData[][] = {
	 	// menu name, short-cut
		{"Strategy",new Character('R'),menuListener},
		// menu name, short-cut, listener, enabled
		{"No Strategy",new Character('N'),itemListener,bT,null},
		{"Top-Down",new Character('T'),itemListener,bT,null},
		{"Divide & Query",new Character('D'),itemListener,bT,null},
		{"Combine Strategies",new Character('C'),itemListener,bT,null}
	};

	Object helpMenuData[][] = {
	 	// menu name, short-cut
		{"Help",new Character('H'),null},
		// menu name, short-cut, listener, enabled
		{"About",new Character('A'),itemListener,bT}
	};

        //  Create the menus
        JMenu fileMenu    = createMenu(fileMenuData);
	JMenu nodeMenu    = createMenu(nodeMenuData);
        JMenu treeMenu    = createMenu(treeMenuData);
        JMenu selectMenu  = createMenu(selectMenuData);
        JMenu strategyMenu = createMenu(strategyMenuData);
        JMenu infoMenu = createMenu(infoMenuData);
        JMenu helpMenu = createMenu(helpMenuData);

        //  Add the menus to the menu bar.
        menuBar.add(fileMenu);
        menuBar.add(nodeMenu);
        menuBar.add(treeMenu);
        menuBar.add(selectMenu);
        menuBar.add(strategyMenu);
        menuBar.add(infoMenu);
        menuBar.add(helpMenu);

     return menuBar;
 }

 private JMenu createMenu(Object [] [] menuData) {
  JMenu menu = new JMenu();
  ButtonGroup g = new ButtonGroup();

  menu.setText( (String) menuData[0][0]) ;
  menu.setMnemonic( ((Character)menuData[0][1]).charValue() );
  if  (menuData[0][2] != null)
	  menu.addMenuListener( (MenuListener)menuData[0][2]);

  for (int i=1; i< menuData.length; i++) {
  	if (menuData[i].length == 0)
		menu.add(new JSeparator() );
	else {
	        JMenuItem menuItem =  createMenuItem(menuData[i],g);
		menu.add( menuItem );
        }
  } // for
  return menu;
 } // createMenu

 private JMenuItem createMenuItem(Object [] data, ButtonGroup g) {

 JMenuItem m = null;

  if (data.length == 4)
       m = new JMenuItem();
  else
        m =  new JCheckBoxMenuItem();


  m.setText((String)data[0]);
  m.setMnemonic( ((Character) data[1]).charValue() );
  m.addActionListener( (ActionListener)data[2]);
  m.setEnabled(  ( (Boolean)data[3]).booleanValue() );

    if (data.length == 5)
        m.setIcon(  (Icon)data[4] );

  return m;
 } // createMenuItem
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




}