

package interfaceProofTree;
import javax.swing.*; // JPanel (entre otros)
import javax.swing.tree.*;

import proofTree.NodeInfo;
import proofTree.ProofTree;

import java.awt.event.*;
import java.awt.*; // para borderLayout
import java.util.*; // para Vector
import java.io.*;



public class ItemListener implements ActionListener {

 ProofTree tree;
 Oracle oracle;

 public ItemListener(ProofTree theTree, Oracle theOracle) {
   tree = theTree;
   oracle = theOracle;
 } // constructor

 public void actionPerformed( ActionEvent evt ) {
  String selected = ( (JMenuItem)evt.getSource()).getText();



  if (selected == "Open")
  	actionOpen();

  if (selected == "Reload")
  	actionReload();
  if (selected == "Save")
  	actionSave();
  if (selected == "Save As")
  	actionSaveAs();

  if (selected == "Exit")
  	System.exit(0);

  if (selected == "Valid")
      actionState(NodeInfo.VALID);
  if (selected == "Non-Valid")
      actionState(NodeInfo.NONVALID);
  if (selected == "Unknown")
      actionState(NodeInfo.DONTKNOW);
  if (selected == "Trusted")
      actionState(NodeInfo.TRUSTED);
  if (selected == "Buggy")
      actionState(NodeInfo.BUGGY);

  if (selected == "Expand Tree")
  	actionExpand();
  if (selected == "Collapse Tree")
  	actionCollapse();
  if (selected == "Remove Trusted Nodes")
  	actionRemoveTrusted();
  if (selected == "Remove Valid &Trusted Nodes")
  	actionRemoveValid();
   if (selected == "Remove Selected Node Subtree")
	 actionRemoveSubtree();
   if (selected == "Make Selected Node New Root")
	 actionRoot();

   if (selected == "Deepest Non-Valid Node")
	 actionDeepest();
   if (selected == "Deepest Desc. with the Same Rule")
	 actionSameRule();
   if (selected == "Center of the Tree")
	 actionDivide();

  if (selected == "Selected Node")
        actionSelectedNode();
  if (selected == "Tree")
  	actionTree();
  if (selected == "History")
  	actionLastChange();
  if (selected == "Clear History")
  	tree.initializeLastAction();

  if (selected == "No Strategy")
      actionNoStrategy();
  if (selected == "Top-Down")
      actionTopDown();
  if (selected == "Divide & Query")
      actionDivideAndQuery();
  if (selected == "Combine Strategies")
      actionCombined();

  if (selected == "About")
  	actionAbout();

  tree.repaint();
//  System.out.println(selected);
 } // actionPerformed

 // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



 private void actionReload() {
   tree.open(tree.getFileName());
 } // actionReload


public void actionSave() {
 tree.save(tree.getFileName());
} // actionSave

private void actionOpen() {
    String name = chooseFile("Open");
    if (name != null)
        tree.open(name);
       oracle.initiateNavigation();

} // actionOpen

private void actionSaveAs() {
    String name = chooseFile("Save As");
    if (name != null)
         tree.save(name);
} // actioSaveAs

private String chooseFile(String title) {
   String name = null;
   JFileChooser chooser = new JFileChooser();


//   ExampleFileFilter filter = new ExampleFileFilter();

//   filter.addExtension("xml");
//   filter.setDescription("xml files");
//   chooser.setFileFilter(filter);
   chooser.setDialogTitle(title);
   chooser.setFileHidingEnabled(true);
   chooser.setCurrentDirectory(new File("."));
   int returnVal = chooser.showOpenDialog(null);
   if (returnVal == JFileChooser.APPROVE_OPTION) {
       name = chooser.getSelectedFile().getAbsolutePath();
   }
   return name;
} // choose file

//////////////////////////////////////////////////////////////////////////////////////////////////////
 private void actionState(int newState) {
   // obtaing the selected node
  DefaultMutableTreeNode node = (DefaultMutableTreeNode)
                                  tree.getLastSelectedPathComponent();
  if (node == null) {
      JOptionPane.showMessageDialog(null,"There is no selected node","DDT Information",JOptionPane.ERROR_MESSAGE);
       return;
   }
   NodeInfo nodeInfo =(NodeInfo)(node.getUserObject());

   if (nodeInfo.getNodeState() != NodeInfo.BUGGY && newState == NodeInfo.BUGGY) {
           JOptionPane.showMessageDialog(null,"Nodes become buggy automatically","DDT Information",JOptionPane.ERROR_MESSAGE);
           return;
        }

   if (nodeInfo.getBasicFact().toString().equals("root") ) {
           JOptionPane.showMessageDialog(null,"The state of the root cannot be changed","DDT Information",JOptionPane.ERROR_MESSAGE);
           return;
        }

   int preState = nodeInfo.getNodeState();
   nodeInfo.setNodeState(newState);
   tree.changedState(node,preState,newState,"(by user)");

 } // actionState

 private void actionLastChange() {
    String msg = "\n"+tree.getLastAction()+"\nTotal number of modifications: "+tree.getLastActionNumber()+".\n";
   JFrame frame = new JFrame("Nodes modified  after the last change of state");

    // prepare the  panel
    JPanel panel=new JPanel();
    panel.setLayout(new BorderLayout());
    JTextArea text = new JTextArea(msg);
    text.setEditable(false);
    Font fuente = new Font("Helvetica", Font.PLAIN, 15);
    text.setFont(fuente);
    JScrollPane scrollPane = new JScrollPane(text);
    panel.add(scrollPane, BorderLayout.CENTER);

    // button
    JButton button = new JButton("Close");
    button.addActionListener(  new ButtonListener(frame) );
    JPanel panelButton=new JPanel();
    panelButton.setLayout(new FlowLayout());
    panelButton.add(button);
    panel.add(panelButton, BorderLayout.SOUTH);

    // frame
    frame.getContentPane().add(panel);
    frame.setSize(400,200);
    frame.setVisible(true);
    frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

 } // actionLastChange

 //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  private void actionTree() {
     JFrame frame = new JFrame("Information about the proof tree");

     Font fuenteBold = new Font("Helvetica", Font.BOLD, 15);
     Font fuente        = new Font("Helvetica", Font.PLAIN, 15);

    // prepare the  panel
    JPanel panel=new JPanel();
    panel.setLayout(new BorderLayout());
    JTextArea text = new JTextArea();
    text.setEditable(false);
    JScrollPane scrollPane = new JScrollPane(text);
    panel.add(scrollPane, BorderLayout.CENTER);

    // button
    JButton button = new JButton("Close");
    button.addActionListener(  new ButtonListener(frame) );
    JPanel panelButton=new JPanel();
    panelButton.setLayout(new FlowLayout());
    panelButton.add(button);
    panel.add(panelButton, BorderLayout.SOUTH);

    // frame
    frame.getContentPane().add(panel);
    frame.setLocation(50,50);
    frame.setSize(500,300);
    frame.setVisible(true);
    frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

    // information displayed
    text.setFont(fuenteBold);
    text.append("\n");
    text.append(" Name of the file: "+tree.getFileName());
    text.append("\n\n");

    text.append(" Number of Nodes: "+tree.getNTotal());
    text.append("\n");

    text.append(" Depth: "+tree.maxDepth());
    text.append("\n");

    text.append(" Maximum Number of Children: "+tree.maxBreadth());
    text.append("\n");
    text.append("\n");



    text.append(" Node Information (breadth first enumeration): \n ");

    text.setFont(fuente);
    Enumeration e = tree.getRoot().breadthFirstEnumeration();
     while (e.hasMoreElements()) {
         DefaultMutableTreeNode node = (DefaultMutableTreeNode)e.nextElement();
         String info = nodeInformation(node);
	 text.append(info);
     } // while
     text.setCaretPosition(0);
     text.repaint();
  } // actionTree

 //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  private void actionSelectedNode() {
       // obtaing the selected node
       DefaultMutableTreeNode node = (DefaultMutableTreeNode)
                                  tree.getLastSelectedPathComponent();
       if (node == null) {
           JOptionPane.showMessageDialog(null,"There is no selected node",
	                                                           "DDT Information",JOptionPane.ERROR_MESSAGE);
           return;
        }

	String msg = nodeInformation(node);

   JFrame frame = new JFrame("Information about the selected node");

    // prepare the  panel
    JPanel panel=new JPanel();
    panel.setLayout(new BorderLayout());
    JTextArea text = new JTextArea(msg);
    text.setEditable(false);
    JScrollPane scrollPane = new JScrollPane(text);
    panel.add(scrollPane, BorderLayout.CENTER);

    Font fuente = new Font("Helvetica", Font.PLAIN, 12);
    text.setFont(fuente);


    // button
    JButton button = new JButton("Close");
    button.addActionListener(  new ButtonListener(frame) );
    JPanel panelButton=new JPanel();
    panelButton.setLayout(new FlowLayout());
    panelButton.add(button);
    panel.add(panelButton, BorderLayout.SOUTH);

    // frame
    frame.getContentPane().add(panel);
    frame.setSize(400,200);
    frame.setVisible(true);
    frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

 } // actionSelectedNode

 // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  private void actionExpand() {
      for (int i=0; i<tree.getRowCount(); i++)
            tree.expandRow(i);
  } // actionExpand

  private void actionCollapse() {
     int n = tree.getRowCount();
      for (int i=n-1; i>0; i--)
            tree.collapseRow(i);
  } // actionExpand

  private void   actionRemoveTrusted() {
     tree.removeNodes(true,false);
       oracle.initiateNavigation();
  }

  private void   actionRemoveValid() {
    tree.removeNodes(true,true);
       oracle.initiateNavigation();
  }

   // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  private void actionRoot() {
       // obtaing the selected node
       DefaultMutableTreeNode node = (DefaultMutableTreeNode)
                                  tree.getLastSelectedPathComponent();
       if (node == null) {
           JOptionPane.showMessageDialog(null,"There is no selected node","DDT Information",JOptionPane.ERROR_MESSAGE);
           return;
        }

       tree.setRoot(node);
       oracle.displayQuestion();
  } // actionRoot

  private void actionRemoveSubtree() {
       // obtaing the selected node
       DefaultMutableTreeNode node = (DefaultMutableTreeNode)
                                  tree.getLastSelectedPathComponent();
       if (node == null) {
           JOptionPane.showMessageDialog(null,"There is no selected node","DDT Information",JOptionPane.ERROR_MESSAGE);
           return;
        }

       if (node == tree.getRoot()) {
           JOptionPane.showMessageDialog(null,"The root of the tree cannot be removed","DDT Information",JOptionPane.ERROR_MESSAGE);
           return;
        }

       tree.removeSubtree(node);
       oracle.initiateNavigation();
  } // actionRoot

     // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  private void actionDeepest() {
    tree.deepest();
    oracle.initiateNavigation();
  }

  private void actionSameRule(){
   // obtaing the selected node
  DefaultMutableTreeNode node = (DefaultMutableTreeNode)
                                  tree.getLastSelectedPathComponent();
  if (node == null) {
      JOptionPane.showMessageDialog(null,"There is no selected node","DDT Information",JOptionPane.ERROR_MESSAGE);
       return;
   }
    tree.deepestSameRule(node);
    oracle.initiateNavigation();
  }

  private void actionDivide() {
    tree.divide();
    oracle.initiateNavigation();
   }

  //%%%%%%%%%%%%%%%%%
   private void actionAbout() {
 	String msg = new String("DDT 1.1 \n \n"+
				                "A Declarative Debugger Tool for Toy. \n\n " +
	 					"By Rafael Caballero. 2003.");
        ImageIcon ddtIcon  = new ImageIcon(tree.getPath()+"/ddt.gif");

           JOptionPane.showMessageDialog(null,msg,"About DDT",
	    					JOptionPane.PLAIN_MESSAGE,ddtIcon);
 } // actionAbout

  // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 private void actionNoStrategy() {
      if (oracle.getStrategy() != Oracle.NO_STRATEGY) {
           oracle.setStrategy(Oracle.NO_STRATEGY);
	}
 } // actionNoStrategy

 private void actionTopDown() {
      if (oracle.getStrategy() != Oracle.TOP_DOWN_STRATEGY) {
           oracle.setStrategy(Oracle.TOP_DOWN_STRATEGY);
	}
 } // actionTopDown


 private void actionCombined() {
      if (oracle.getStrategy() != Oracle.COMBINED_STRATEGY) {
           oracle.setStrategy(Oracle.COMBINED_STRATEGY);
	}
 } // actionCombined

  private void actionDivideAndQuery() {
     if (oracle.getStrategy() != Oracle.DIVIDE_AND_QUERY_STRATEGY) {
           oracle.setStrategy(Oracle.DIVIDE_AND_QUERY_STRATEGY);
	}
  } // actionDivideAndQuery

  // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // auxiliar functions
   // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  String nodeInformation(DefaultMutableTreeNode node) {
       // prepare the message
       String msg = new String();
       NodeInfo nodeInfo =(NodeInfo)(node.getUserObject());
       int state = nodeInfo.getNodeState();
       msg += " \n";
       msg += " Basic fact : "+nodeInfo.toString()+".\n";
       msg += " State : ''"+NodeInfo.getStateString(state)+"''.\n";
       msg += " Program rule : "+nodeInfo.getFunctionName()+"."+
							nodeInfo.getFunctionNumber()+".\n";
//       msg += " Rule instance : "+nodeInfo.getFunctionInstance()+".\n";
/*       msg += " Entails : ";

       Vector entails = nodeInfo.getEntails();
       if (entails != null)
       for (int i=0; i<entails.size(); i++) {
            msg += entails.get(i);
	    if ((i+1)<entails.size())
	        msg +=", ";
       }
       msg +=".\n";
*/
  return msg;
 } // nodeInformation

} // MenuListener

class ButtonListener implements ActionListener {
	JFrame frame;

        public ButtonListener(JFrame theFrame) {
	 frame = theFrame;
	 }
        // 8. Interface method that has been implemented.
        //    Ring the bell five times...

        public void actionPerformed(ActionEvent ae) {
	 frame.dispose();
	}
}
