/*
 * Created on 25-sep-2005
 *
 */
package terms;

import java.util.TreeSet;
import java.util.Vector;
import java.util.Iterator;


/**
 * @author Rafael Caballero
 * Represents a program rule condition
 */
public class Condition {
	private Vector conds; // the condition is repr. through a vector o terms
	
    public Condition() { conds = null; }
	public Condition(Vector v) {conds = v; }
   
   public boolean empty(){return conds==null || conds.size()==0; }
 
   public String toString() {
   	String s = "";
   	for (int i=0; i<conds.size(); i++) {
   		Term t = (Term) conds.elementAt(i);
   		s += ""+t;
   		if (i<conds.size()-1) s+=", ";
   	}
   	return s;
   }
   
   public Vector getConds() { return conds; }
   public void setConds(Vector conds) {this.conds = conds; }
   
   /**
    *  Variables in this term.
    */
	public TreeSet variables() {
	   TreeSet t = null;	
	   TreeSet t2;
	  for (int i=0; i<conds.size(); i++) {
	   	 		t2 = ((Term)conds.elementAt(i)).variables();
	   	 		if (t2 != null) {
	   	 			if (t==null) t = new TreeSet();
	   	 			t.addAll(t2);
	   	 		}
	  }
	  return t;
	}

	/**
	 * 
	 * 
	 * Simpliflies the constraint condition 
	 * @param  tree union of vars(lhs) and of vars(rhs)
	 */
	public void simplify(TreeSet tree) {
		int i=0;
		String condS = toString();
		while (i<conds.size()) {
			Term cond = (Term) conds.elementAt(i);
			TreeSet vars = cond.variables();
			
			
			if (firstSimplifyCondition(tree,cond) ||
				secondSimplifyCondition(tree,vars))	
					conds.removeElementAt(i);
			else i++;
			}
	}
	
	// first: if the constraint is of the form
	// (A/=B) an either A or B is not in vars(lhs) U vars(rhs)
	private boolean firstSimplifyCondition(TreeSet tree, Term t) {
		boolean result = false;

		if (t.getClass().getName().equals("terms.RoundBracketExpression")) {
			RoundBracketExpression rbe = (RoundBracketExpression) t;
			Term inside = rbe.getT();
			if (inside.getClass().getName().equals("terms.Application")) {
				Application app = (Application) inside;
				Term e = app.getE();
				Vector args = app.getArgs();
				if (args.size()==2 && 
					args.elementAt(0).getClass().getName().equals("terms.Operator") ){
					
					Operator ope = (Operator) args.elementAt(0);
					if (ope.getOp().equals("/=")) 
						if (args.elementAt(1).getClass().getName().equals("terms.Variable")  ) {
							Variable v = (Variable)args.elementAt(1);
							String name = v.getId();
							if (tree==null || !tree.contains(name)) result = true;
						} 
					    if (!result && e.getClass().getName().equals("terms.Variable")) {
							Variable v = (Variable)e;
							String name = v.getId();
							if (tree==null || !tree.contains(name)) result = true;
						}
					
				}
			 }
		}
			
		return result;
	}

	// second: if no variable of the constraint is in vars(lhs) U vars(rhs)
	private boolean secondSimplifyCondition(TreeSet tree, TreeSet vars) {
		boolean removeIt = true;
		
		if (vars != null) {
			   // if the set of vars in the condition has no element
			   // in common with 'tree', remove it
			   Iterator ite = vars.iterator();
			   
			   while (removeIt && ite.hasNext()) {
			   	 String ele = (String) ite.next();
			   	 if (tree!=null && tree.contains(ele)) // ele occurred in the basic fact
			   	 	removeIt = false;
			   }
		}
		return removeIt;
	}

}
