/* Copyright(C) 1999, Swedish Institute of Computer Science */

#include "fd.h"
#include "dvars.h"

#if MULTI_SP_AWARE
#define scalar_product_init(A1,A2) scalar_product_init(HIDDEN_ARG, A1,A2)
#define scalar_product_le(A1,A2) scalar_product_le(HIDDEN_ARG, A1,A2)
#define scalar_product_ge(A1,A2) scalar_product_ge(HIDDEN_ARG, A1,A2)
#define free_one_dp(A1,A2) free_one_dp(HIDDEN_ARG, A1,A2)
#define free_all_dp(A1) free_all_dp(HIDDEN_ARG, A1)
#define maybe_rehash(A1) maybe_rehash(HIDDEN_ARG, A1)
#define knapsack_filter_noninc(A1,A2,A3,A4) knapsack_filter_noninc(HIDDEN_ARG, A1,A2,A3,A4)
#define knapsack_filter_inc(A1,A2,A3) knapsack_filter_inc(HIDDEN_ARG, A1,A2,A3)
#define knapsack_filter(A1,A2,A3,A4,A5,A6) knapsack_filter(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)
#define daemon_copy_state(A1,A2) daemon_copy_state(HIDDEN_ARG, A1,A2)

#if DynamicMallocBase /* [PM] 3.10.1 MAGIC needed for DynamicMallocBase */
#define add_ztd(A1,A2) add_ztd(HIDDEN_ARG, A1,A2)
#define scalar_product_le_fast(A1,A2) scalar_product_le_fast(HIDDEN_ARG, A1,A2)
#define scalar_product_ge_fast(A1,A2) scalar_product_ge_fast(HIDDEN_ARG, A1,A2)
#define fd_product_inverse_dispatch(A1,A2,A3) fd_product_inverse_dispatch(HIDDEN_ARG, A1,A2,A3)
#endif /* DynamicMallocBase */
#endif /* MULTI_SP_AWARE */

#define NegateLB(LB) ((LB)==Inf ? Sup : Tminus(LB))

#define NegateUB(UB) ((UB)==Sup ? Inf : Tminus(UB))

#define ABS(G) ((G)>=0 ? (G) : -(G))

static INLINE long gcd (long c1,long c2)
{
  if (c1 > 1 && c2 > 1) {
    while (c1 != 0 && c2 != 0) {
      c1 %= c2;
      if (c1 != 0) c2 %= c1;
    }
    return c1+c2;
  } else {
    return 1;
  }
}

#define HALFSHIFT (4<<LogSizeOfWord)
#ifdef __MSVC_RUNTIME_CHECKS
/* [PM] 3.9.1 .NET "/RTCc Reports when a value is assigned to a
        smaller data type that results in a data loss." */
#define LOWER_UHALF(X) ((UHALF)((X) & ((UHALF)~0)))
#else
#define LOWER_UHALF(X) ((UHALF)(X))
#endif

/* returns TRUE on overflow */
/* ripped off from wamfunc.c */
static BOOL long_mult ( /* (x, y, productp) */
     WORD x, WORD y, WORD *productp)
{
  UWORD lower, tmp;
  UHALF x0, x1, y0, y1;
  int negated;
  
  if (x>>HALFSHIFT==x>>(HALFSHIFT-1) &&
      y>>HALFSHIFT==y>>(HALFSHIFT-1)) {
    *productp = x*y;
    return FALSE;
  }

  negated = 0;
  if (x<0)
    x = -x, negated ^= 1;
  if (y<0)
    y = -y, negated ^= 1;

  x0 = LOWER_UHALF(x);
  x1 = (UHALF)(x>>HALFSHIFT);
  y0 = LOWER_UHALF(y);
  y1 = (UHALF)(y>>HALFSHIFT);

  lower = 0L;
  if (x0 && y0) {
    lower = (UWORD)x0*(UWORD)y0;
  }
  if (x1 && y1) {
    return TRUE;
  } else if (x0 && y1) {
    tmp = (UWORD)x0*(UWORD)y1 + (lower>>HALFSHIFT);
    if (tmp>>HALFSHIFT)
      return TRUE;
    else
      lower = (tmp<<HALFSHIFT) + LOWER_UHALF(lower);
  } else if (x1 && y0) {
    tmp = (UWORD)x1*(UWORD)y0 + (lower>>HALFSHIFT);
    if (tmp>>HALFSHIFT)
      return TRUE;
    else
      lower = (tmp<<HALFSHIFT) + LOWER_UHALF(lower);
  }
  if ((WORD)lower<0L && (lower<<1L || !negated))
    return TRUE;
  if (negated) lower = -lower;
  *productp = lower;
  return FALSE;
}

/* precond: t1 and t2 are tagged integers */
static TAGGED safe_mul_val ( /* (t1,t2) */
     TAGGED t1, TAGGED t2)
{
  long st = GetSmall(t1);
  long su = GetSmall(t2);
  long prod;
  
  if (!long_mult(st,su,&prod) && IntIsSmall(prod))
    return MakeSmall(prod);
  else
    return (((st<0)^(su<0)) ? Inf : Sup);
}

/* operations exported to indexical.c */

TAGGED safe_mul ( /* (t1,t2) */
                TAGGED t1, TAGGED t2)
{
  if (TagIsSmall(t1)) {
    if (TagIsSmall(t2))
      t1 = safe_mul_val(t1,t2);
    else if (Teqz(t1))
      return ERRORTAG;	/* 0*inf -> give up */
    else
      t1 = (Tltz(t1)^(t2==Inf) ? Inf : Sup);
  } else if (TagIsSmall(t2)) {
    if (Teqz(t2))
      t1 = ERRORTAG;		/* inf*0 -> give up */
    else
      t1 = (Tltz(t2)^(t1==Inf) ? Inf : Sup);
  } else
    t1 = (t1!=t2 ? Inf : Sup);
  return t1;
}

/* ceiling of quotient */
TAGGED safe_divu ( /* (t1,t2) */
     TAGGED t1, TAGGED t2)
{
  if (TagIsSmall(t1)) {
    if (TagIsSmall(t2)) {
      if (Tnez(t2)) {
	WORD x = t1-TaggedZero;
	WORD y = t2-TaggedZero;
	WORD q = (  x==0 ? 0
		    : (x<0)^(y<0) ? x/y 
		    : x>0 ? (x+y-1)/y
		    : (x+y+1)/y
		    );
		    
	t1 = MakeSmall(q);
      } else if (Tltz(t1))
	t1 = Inf;
      else if (Tgtz(t1))
	t1 = Sup;
      else
	t1 = ERRORTAG;	/* 0/0 -> give up */
    } else if (Tnez(t1))	/* +/sup, -/inf -> 1 */
      t1 = (Tltz(t1)^(t2==Inf) ? TaggedZero : TaggedOne);
  } else if (TagIsSmall(t2))
    t1 = (Tltz(t2)^(t1==Inf) ? Inf : Sup);
  else
    t1 = ERRORTAG;		/* inf/inf -> give up */
  return t1;
}

/* floor of quotient */
TAGGED safe_divd ( /* (t1,t2) */
     TAGGED t1, TAGGED t2)
{
  if (TagIsSmall(t1)) {
    if (TagIsSmall(t2)) {
      if (Tnez(t2)) {
	WORD x = t1-TaggedZero;
	WORD y = t2-TaggedZero;
	WORD q = ( x==0 ? 0
		   : (x<0)^(y>0) ? x/y 
		   : x>0 ? (x-y-1)/y
		   : (x-y+1)/y
		   );
	      
	t1 = MakeSmall(q);
      } else if (Tltz(t1))
	t1 = Inf;
      else if (Tgtz(t1))
	t1 = Sup;
      else
	t1 = ERRORTAG;	/* 0/0 -> give up */
    } else if (Tnez(t1))	/* -/sup, +/inf -> -1 */
      t1 = (Tltz(t1)^(t2==Sup) ? TaggedZero : TaggedMinusOne);
  } else if (TagIsSmall(t2))
    t1 = (Tltz(t2)^(t1==Inf) ? Inf : Sup);
  else
    t1 = ERRORTAG;		/* inf/inf -> give up */
  return t1;
}

TAGGED safe_plus ( /* (t1,t2) */
     TAGGED t1, TAGGED t2)
{
  if (TagIsSmall(t1)) {
    if (TagIsSmall(t2)) {
      TADDCHK(t1,t2);
    } else
      t1 = t2;
  }
  return t1;
}

TAGGED safe_minus ( /* (t1,t2) */
     TAGGED t1, TAGGED t2)
{
  if (TagIsSmall(t1)) {
    if (TagIsSmall(t2)) {
      TSUBCHK(t1,t2);
    } else
      t1 = (t2==Inf ? Sup : Inf);
  }
  return t1;
}


/* operations on double word entities, for scalar_product */

struct dword {
  union {
    UWORD asuword;
    WORD asword;
  } msw;
  union {
    UWORD asuword;
    WORD asword;
  } lsw;
};

typedef struct dword *Dword;

#define ASLONG(DW) ((DW)->lsw.asword)

static BOOL
issmall_d(Dword x)
{
  return
    ((x->msw.asword==0  && x->lsw.asword>=0 && x->lsw.asword<HighInt) ||
     (x->msw.asword==-1 && x->lsw.asword<0  && x->lsw.asword>=-HighInt));
}

#define islong_d(X)					\
    (((X)->msw.asword==0  && (X)->lsw.asword>=0) ||	\
     ((X)->msw.asword==-1 && (X)->lsw.asword<0))

#define eq_dd(X,Y)				\
    (((X)->lsw.asword==(Y)->lsw.asword) &&	\
     ((X)->msw.asword==(Y)->msw.asword))

static int 
cmp_dd(Dword x,Dword y)
{
  if (x->msw.asword==y->msw.asword)
    return
      (x->lsw.asuword < y->lsw.asuword) ? -1 :
      (x->lsw.asuword > y->lsw.asuword) ?  1 : 0;
  else if (x->msw.asword < y->msw.asword)
    return -1;
  else
    return 1;
}

static int 
cmp_dz(Dword x)
{
  if (x->msw.asword==0)
    return
      (x->lsw.asuword > 0) ?  1 : 0;
  else if (x->msw.asword < 0)
    return -1;
  else
    return 1;
}

#define add_zdd(X,Y) (*(Y) = *(X))

static void add_ztd MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                           TAGGED x,Dword y)
{
  if (TagIsSmall(x)) {
    y->lsw.asword = GetSmall(x);
    y->msw.asword = (Tltz(x) ? -1 : 0);
  } else {
    TAGGED *s = TagToSTR(x);
    y->lsw.asuword = s[1];
    if (s[0]==MakeFunctorFix)
      y->msw.asword = ((WORD)s[1]<0 ? -1 : 0);
    else
      y->msw.asuword = s[2];
  }
}

static void 
add_zld(long x,Dword y)
{
  y->lsw.asword = x;
  y->msw.asword = (x<0 ? -1 : 0);
}

static void 
add_ddd(Dword x,Dword y,Dword z)
{
  UWORD lsw = x->lsw.asuword + y->lsw.asuword;
  BOOL carry = (lsw < x->lsw.asuword);

  z->lsw.asuword = lsw;
  z->msw.asuword = x->msw.asuword + y->msw.asuword + carry;
}

static void 
sub_ddd(Dword x,Dword y,Dword z)
{
  UWORD lsw = x->lsw.asuword - y->lsw.asuword;
  BOOL borrow = (lsw > x->lsw.asuword);

  z->lsw.asuword = lsw;
  z->msw.asuword = x->msw.asuword - y->msw.asuword - borrow;
}

static void 
sub_dld(Dword x,long y,Dword z)
{
  UWORD lsw = x->lsw.asuword - (UWORD)y;
  BOOL borrow = (lsw > x->lsw.asuword) - (y<0);

  z->lsw.asuword = lsw;
  z->msw.asuword = x->msw.asuword - borrow;
}

static void 
sub_zdd(Dword x,Dword y)
{
  UWORD lsw;
  
  y->lsw.asuword = lsw = - x->lsw.asuword;
  y->msw.asuword = - x->msw.asuword;
  if (lsw!=0)
    y->msw.asuword--;
}

static void 
add_dtd(Dword x,TAGGED ty,Dword z)
{
  UWORD lsw;
  int carry;

  if (ty==Inf) {
    lsw = x->lsw.asuword;
    carry = -1;
  } else if (ty==Sup) {
    lsw = x->lsw.asuword;
    carry = 1;
  } else {
    lsw = x->lsw.asuword + GetSmall(ty);
    carry = (lsw < x->lsw.asuword) - Tltz(ty);
  }
  z->lsw.asuword = lsw;
  z->msw.asuword = x->msw.asuword + carry;
}

static void 
add_dud(Dword x,UWORD y,Dword z)
{
  UWORD lsw = x->lsw.asuword + y;
  int carry = (lsw < x->lsw.asuword);
  
  z->lsw.asuword = lsw;
  z->msw.asuword = x->msw.asuword + carry;
}

static void 
sub_dud(Dword x,UWORD y,Dword z)
{
  UWORD lsw = x->lsw.asuword - y;
  int carry = (lsw < x->lsw.asuword) - (y>0);
  
  z->lsw.asuword = lsw;
  z->msw.asuword = x->msw.asuword + carry;
}


static void 
mul_uud(UWORD x,UWORD y,Dword z)
{
  int carry = 0;
  UWORD lsw;
  UWORD x0 = LOWER_UHALF(x);
  UWORD x1 = (UHALF)(x>>HALFSHIFT);
  UWORD y0 = LOWER_UHALF(y);
  UWORD y1 = (UHALF)(y>>HALFSHIFT);
  UWORD p01 = (UWORD)x0*(UWORD)y1;
  UWORD p10 = (UWORD)x1*(UWORD)y0;
  
  z->lsw.asuword = (UWORD)x0*(UWORD)y0;
  z->msw.asuword = (UWORD)x1*(UWORD)y1;
  lsw = z->lsw.asuword + (p01<<HALFSHIFT);
  if (lsw<z->lsw.asuword) carry++;
  z->lsw.asuword = lsw;
  lsw = z->lsw.asuword + (p10<<HALFSHIFT);
  if (lsw<z->lsw.asuword) carry++;
  z->lsw.asuword = lsw;
  z->msw.asuword += (p01>>HALFSHIFT) + (p10>>HALFSHIFT) + carry;
}

static void 
mul_ltd(WORD x,TAGGED ty,Dword z)
{
  if (ty==Inf) {
    z->msw.asword = -x;
    z->lsw.asword = 0;
  } else if (ty==Sup) {
    z->msw.asword = x;
    z->lsw.asword = 0;
  } else {
    WORD y = GetSmall(ty);
  
    if (x>>HALFSHIFT==x>>(HALFSHIFT-1) &&
	y>>HALFSHIFT==y>>(HALFSHIFT-1)) {
      z->lsw.asword = x*y;
      z->msw.asword = (z->lsw.asword<0 ? -1 : 0);
    } else {
      int negated = 0;
      
      if (x<0)
	x = -x, negated ^= 1;
      if (y<0)
	y = -y, negated ^= 1;

      mul_uud(x,y,z);
      if (negated)
	sub_zdd(z,z);
    }
  }
}

/* Roll-your-own division and remainder algorithms.
   Assume everything is unsigned.

   Consider X/Y, X = (A<<32)+B, |X|<=63, |Y|<=25.

   Let:    Q1 = (1<<32)/Y       |Q1|<=31
           R1 = (1<<32)%Y       |R1|<=25
	   Q2 = B/Y             |Q2|<=6
	   R2 = B%Y             |R2|<=25

   Case 1 (A=0): X/Y = Q2, X%Y = R2.

   Case 2 (A>0): X/Y = A*Q1 + Q2 + (A*R1 + R2)/Y
                 X%Y = (A*R1 + R2)%Y
*/

/* x is nonnegative! */
static void 
div_dud_internal(Dword x,UWORD y,Dword z)
{
  struct dword dividend = *x;
  struct dword term;
  UWORD q1, r1, q2, r2;

  if (y==1) {
    z->lsw.asuword = x->lsw.asuword;
    z->msw.asuword = x->msw.asuword;
  } else {
    z->lsw.asuword = 0;
    z->msw.asuword = 0;
    /* note: dividend is two times too small */
    q1 = (((UWORD)((-1L)<<(WORDSIZE-1)))/y)<<1;
    r1 = (((UWORD)((-1L)<<(WORDSIZE-1)))%y)<<1;
    if (r1>=y) {
      q1++;
      r1 -= y;
    }
    q2 = dividend.lsw.asuword/y;
    r2 = dividend.lsw.asuword%y;
    while (dividend.msw.asuword>0) {
      mul_uud(q1,dividend.msw.asuword,&term);
      add_ddd(z,&term,z);
      add_dud(z,q2,z);
      mul_uud(r1,dividend.msw.asuword,&dividend);
      add_dud(&dividend,r2,&dividend);
      q2 = dividend.lsw.asuword/y;
      r2 = dividend.lsw.asuword%y;
    }
    add_dud(z,q2,z);
  }
}

static void 
div_dld(Dword x,WORD y,Dword z)
{
  int negx=0, negy=0;

  if (cmp_dz(x)<0) {
    negx=1;
    sub_zdd(x,x);
  }
  if (y<0) {
    negy=1;
    y = -y;
  }
  div_dud_internal(x,y,z);
  if (negx)
    sub_zdd(x,x);
  if (negx^negy)
    sub_zdd(z,z);
}

static WORD 
div_dll(Dword x,WORD y)
{
  struct dword z;

  div_dld(x,y,&z);
  return z.lsw.asword;
}

/* x is nonnegative! */
static UWORD 
mod_duu_internal(Dword x,UWORD y)
{
  struct dword dividend = *x;
  UWORD r1, r2;

  if (y==1)
    return 0;
  /* note: dividend is two times too small */
  r1 = (((UWORD)((-1L)<<(WORDSIZE-1)))%y)<<1;
  if (r1>=y) {
    r1 -= y;
  }
  r2 = dividend.lsw.asuword%y;
  while (dividend.msw.asuword>0) {
    mul_uud(r1,dividend.msw.asuword,&dividend);
    add_dud(&dividend,r2,&dividend);
    r2 = dividend.lsw.asuword%y;
  }
  return r2;
}

/* see adjust_bounds_gcd for x<0 case! */
static UWORD 
mod_duu(Dword x,UWORD y)
{
  UWORD m;

  if (cmp_dz(x)>=0)
    return mod_duu_internal(x,y);
  sub_zdd(x,x);
  m = mod_duu_internal(x,y);
  sub_zdd(x,x);
  return (m==0 ? 0 : y-m);	/* modulo, not remainder */
}

typedef long TERM;

struct linear_data {
  void (SPCDECL *destructor)(void *);
  void (SPCDECL *daemon)(HIDDEN_PROTO
			 void *,SP_globref,TAGGED); /* (handle,attr_ref,global) */
#if MULTI_SP_AWARE
  SPEnv *spenv;
#endif /* MULTI_SP_AWARE */

  SP_globref refbase;
  BOOL fast;
  int op;			/* 1(#>=), 2(#>=), 3(#=), 4(#\=) */
  long stamp;			/* increases up to backtracking */
  int units;			/* maintained incrementally */
  int nonunits;			/* maintained incrementally */
  int gcdall;			/* valid during GCD rules */
  struct dword rhs;		/* RHS - valid up to backtracking */
  int prunings;			/* counter */
  int nvars;			/* #terms */
  int ntargets;			/* #terms that may be targets */
  struct dword bigf;
  struct dword bige;
  TERM *heap;
  int *vheap;
  int heapsize;
  TERM *target;
  Dvar dvar;
  struct {
    struct dword *cmin;		/* min(ai*xi) */
    struct dword *cmax;		/* max(ai*xi) */
    struct dword *interval;	/* cmax-cmin */
    long *coeff;		/* ai */
    long *abscoeff;		/* |ai| */
    long *gcd;			/* valid during GCD rules */
  } term;
};


#define SV(I)   (pdata->target[I])
#define COEFF(t) (pdata->term.coeff[t])
#define ABSCOEFF(t) (pdata->term.abscoeff[t])
#define GCD(t) (pdata->term.gcd[t])
#define DVAR(t) (pdata->dvar+(t))
#define RefAttr(T) (pdata->refbase + 2*(T))
#define RefVar(T) (pdata->refbase + 2*(T) + 1)
#define CMIN(t) (pdata->term.cmin+(t))
#define CMAX(t) (pdata->term.cmax+(t))
#define INTERVAL(t) (pdata->term.interval+(t))

static void SPCDECL linear_destructor(void *pdata_v)
{
  struct linear_data *pdata = (struct linear_data *)pdata_v;
  FD_SETUP_SPENV(pdata->spenv)

  SP_free_globrefs(pdata->refbase,pdata->nvars<<1);
  SP_free(pdata);
}

#define SWAP(I,J)				\
{						\
  TERM vi = pdata->heap[I];			\
  TERM vj = pdata->heap[J];			\
  pdata->heap[I] = vj;				\
  pdata->heap[J] = vi;				\
  pdata->vheap[vi] = (J);			\
  pdata->vheap[vj] = (I);			\
}

static void 
spheapify(struct linear_data *pdata,
	  int i)
{
  TERM *heap = pdata->heap;
  struct dword *key = pdata->term.interval;
  
  for (;;) {
    int l = (i<<1)+1;
    int topmost = i;
    if (l<pdata->heapsize && cmp_dd(key+heap[l],key+heap[topmost])>0)
      topmost = l;
    if (l+1<pdata->heapsize && cmp_dd(key+heap[l+1],key+heap[topmost])>0)
      topmost = l+1;
    if (topmost==i)
      break;
    SWAP(i,topmost);
    i = topmost;
  }
}

static void 
refresh_gcd(struct linear_data *pdata)
{
  int ntargets = pdata->ntargets;
  int units = pdata->units;
  int nonunits = pdata->nonunits;
  long g;
  int i;

  if (nonunits==0)
    pdata->gcdall = 1;
  else if (units>1) {
    pdata->gcdall = 1;
    for (i=0; i<ntargets; i++)
      GCD(SV(i)) = 1;
  } else if (units==1) { /* GCD(i)=gcdall=1 except for one i */
    int ix=0;			/* avoid false alarm */
    pdata->gcdall = 1;
    g = -1;
    for (i=0; i<ntargets; i++) {
      TERM elt = SV(i);
      long ac = ABSCOEFF(elt);

      GCD(elt) = 1;
      if (cmp_dz(INTERVAL(elt))==0)
	;
      else if (ac==1)
	ix = elt;
      else if (g== -1)
	g = ac;
      else
	g = gcd(g,ac);
    }
    GCD(ix) = g;
  } else {	/* for each i, compute GCD of all coefficients except i */
    int i0;
    for (i0=0; i0<ntargets && cmp_dz(INTERVAL(SV(i0)))==0; i0++)
      ;
    g = ABSCOEFF(SV(i0));
    for (i=i0+1; i<ntargets; i++) {
      TERM elt = SV(i);
      
      if (cmp_dz(INTERVAL(elt))>0) {
	GCD(elt) = g;
	g = gcd(g,ABSCOEFF(elt));
      }
    }
    pdata->gcdall = g;
				/* now, SV(i)->gcd = gcd(a[0] ... a[i-1]) */
    for (i=ntargets-1; i>=0 && cmp_dz(INTERVAL(SV(i)))==0; i--)
      ;
    g = ABSCOEFF(SV(i));
    for (i--; i>i0; i--) {
      TERM elt = SV(i);
      
      if (cmp_dz(INTERVAL(elt))>0) {
	GCD(elt) = gcd(g,GCD(elt));
	g = gcd(g,ABSCOEFF(elt));
      }
    }
    GCD(SV(i0)) = g;
				/* now, SV(i)->gcd = gcd({a[j] | i!=j}) */
  }
}

static TAGGED
daemon_copy_state MAGIC (HIDDEN_PROTO
			 Argdecl,
			 TAGGED global)
{
  TAGGED *s, *h, tstate;
  int ar, i;
  LetShadowregs;
  
  tstate = RefMutable(CTagToArg(global,1));
  ar = Arity(TagToHeadfunctor(tstate));
  s = TagToArg(tstate,0);
  if (s >= GLOBAL_UNCOND) {
    s[ar] += IStep(1);		/* increment stamp */
  } else {
    h = w->global_top;
    for (i=0; i<ar+1; i++)
      h[i] = s[i];
    w->global_top = h+ar+1;
    h[ar] += IStep(1);		/* increment stamp */
    tstate = MakeStructure(h);
    update_mutable(w,tstate,CTagToArg(global,1));
  }
  return tstate;
}



static void SPCDECL 
linear_daemon MAGIC (HIDDEN_PROTO
		     void *vdata,
		     SP_globref attr_ref,
		     TAGGED global)
{
  struct linear_data *pdata = (struct linear_data *)vdata;
  TAGGED tstate;
  int ar, state_stamp;
  WAMENV;

  tstate = RefMutable(CTagToArg(global,1));
  ar = Arity(TagToHeadfunctor(tstate));
  state_stamp = GetSmall(CTagToArg(tstate,ar));
  if (pdata->stamp==state_stamp) { /* incremental */
    int j = (attr_ref - pdata->refbase)>>1;
    long c = COEFF(j);
    long ac = ABSCOEFF(j);
    Dvar dv = DVAR(j);
    Dword cminj = CMIN(j);
    Dword cmaxj = CMAX(j);
    Dword intervalj = INTERVAL(j);
    TERM elt;

#if 0
    t1 = RefGlob(attr_ref);
    DerefAttribute(t1,t1);	/* get dom/4 term */
    if (dv->min==DomainMin(t1) && dv->max==DomainMax(t1))
      return;			/* false alarm - self-triggered */
#endif
    tstate = daemon_copy_state(w,global);
    pdata->stamp++;
    add_ddd(&pdata->bigf,cminj,&pdata->bigf);
    sub_ddd(&pdata->bige,cmaxj,&pdata->bige);
    dvar_init(dv, attr_ref, attr_ref+1);
    if (c>0) {
      mul_ltd(c,dvar_min_t(dv),cminj);
      mul_ltd(c,dvar_max_t(dv),cmaxj);
    } else {
      mul_ltd(c,dvar_max_t(dv),cminj);
      mul_ltd(c,dvar_min_t(dv),cmaxj);
    }
    sub_ddd(cmaxj,cminj,intervalj);
    sub_ddd(&pdata->bigf,cminj,&pdata->bigf);
    add_ddd(&pdata->bige,cmaxj,&pdata->bige);
    if (dvar_is_integer(dv)) {
      if (ac==1)
	pdata->units--;
      else
	pdata->nonunits--;
    }
    if (pdata->units+pdata->nonunits==0)
      goto push;
    switch (pdata->op) {
    case 1:
      spheapify(pdata,pdata->vheap[j]);
      elt = pdata->heap[0];
      if (cmp_dd(&pdata->bigf,INTERVAL(elt))<0)
	goto push;
      return;
    case 2:
      spheapify(pdata,pdata->vheap[j]);
      elt = pdata->heap[0];
      if (cmp_dd(&pdata->bige,INTERVAL(elt))<0)
	goto push;
      return;
    case 3:
      spheapify(pdata,pdata->vheap[j]);
      elt = pdata->heap[0];
      if (cmp_dd(&pdata->bigf,INTERVAL(elt))<0 ||
	  cmp_dd(&pdata->bige,INTERVAL(elt))<0)
	goto push;
      return;
    case 4:
      if (pdata->units+pdata->nonunits <= 1)
	goto push;
      return;
    }
  }
 push:
  fd_enqueue_global(w, global, 0x4/* MINMAX, append*/);   
}

static void
scalar_product_init MAGIC (HIDDEN_PROTO
			   struct linear_data *pdata,
			   BOOL incremental)
{
  int op = pdata->op;
  int i;
  /* Phase 1: compute:
     bigf = rhs - sum{min(a_i x_i)}
     bige = sum{max(a_i x_i)} - rhs

     I_i = max(a_i x_i) - min(a_i x_i)
  */
  if (incremental) {
    for (i=0; i<pdata->ntargets; i++) {
      int j = SV(i);
      Dvar dv = DVAR(j);
      
      dvar_init(dv, RefAttr(j), RefVar(j));
      if (dvar_is_integer(dv))
	sub_ddd(&pdata->rhs,CMIN(j),&pdata->rhs);
    }
  } else {
    pdata->fast = TRUE;
    add_zdd(&pdata->rhs,&pdata->bigf);
    sub_zdd(&pdata->bigf,&pdata->bige);
    pdata->units = 0;
    pdata->nonunits = 0;
    for (i=0; i<pdata->ntargets; i++) {
      int j = SV(i);
      long c = COEFF(j);
      long ac = ABSCOEFF(j);
      Dvar dv = DVAR(j);
      Dword cminj = CMIN(j);
      Dword cmaxj = CMAX(j);
      Dword intervalj = INTERVAL(j);

      dvar_init(dv, RefAttr(j), RefVar(j));
      if (c>0) {
	mul_ltd(c,dvar_min_t(dv),cminj);
	mul_ltd(c,dvar_max_t(dv),cmaxj);
      } else {
	mul_ltd(c,dvar_max_t(dv),cminj);
	mul_ltd(c,dvar_min_t(dv),cmaxj);
      }
      sub_ddd(cmaxj,cminj,intervalj);
      if (op<4) {
	pdata->heap[i] = j;
	pdata->vheap[j] = i;
      }
      sub_ddd(&pdata->bigf,cminj,&pdata->bigf);
      add_ddd(&pdata->bige,cmaxj,&pdata->bige);
      if (pdata->fast &&
	  !(islong_d(cminj) && islong_d(cmaxj) && islong_d(intervalj)))
	pdata->fast = FALSE;
      if (dvar_is_integer(dv)) {
	sub_ddd(&pdata->rhs,cminj,&pdata->rhs);
      } else if (ac==1)
	pdata->units++;
      else
	pdata->nonunits++;
    }
    pdata->heapsize = i;
    if (op<4)
      for (i=(pdata->heapsize-2)>>1; i>=0; i--)
	spheapify(pdata,i);
    if (pdata->fast &&
	!(islong_d(&pdata->bige) && islong_d(&pdata->bigf) && islong_d(&pdata->rhs)))
      pdata->fast = FALSE;
  }
}


static BOOL
scalar_product_le MAGIC (HIDDEN_PROTO
			 struct linear_data *pdata,
			 TERM elt)
{
  long c = COEFF(elt);
  Dvar dv = DVAR(elt);
  struct dword decr;
  struct dword cmax0;

  add_zdd(CMAX(elt),&cmax0);

  /*
    Phase 2:

    For <=
    ******
    bigf>=0 is a necessary condition.
    bige<=0 is a sufficient condition.

    enforce:
    bigf >= I_i for all i

    rules:
    x_i <=  floor(F / a_i) + min(x_i)  if a_i>0
    x_i >= -floor(F /-a_i) + max(x_i)  if a_i<0

  */

  if (cmp_dz(&pdata->bigf)<0)
    return FALSE;
  if (c>0) {
    struct dword ub;
    TAGGED tmin = dvar_min_t(dv);

    div_dld(&pdata->bigf,c,&ub);
    if (TagIsSmall(tmin)) {
      add_dtd(&ub,tmin,&ub);
    } else {
      struct dword tmp;
      div_dld(CMIN(elt),c,&tmp);
      add_ddd(&ub,&tmp,&ub);
    }
    if (issmall_d(&ub)) {/* representable ub */
      if (dvar_fix_max_l(dv,ub.lsw.asword)<0)
	return FALSE;
      mul_ltd(c,dvar_max_t(dv),CMAX(elt));
    } else if (cmp_dz(&ub)>0) {	/* ub overflow - can't prune the variable */
      long modc = mod_duu(&pdata->bigf,c);
      
      add_ddd(&pdata->bigf,CMIN(elt),CMAX(elt)); /* ensure bigf >= I_i */
      if (modc>0)
	sub_dud(CMAX(elt),modc,CMAX(elt));
    } else {			/* ub underflow - signal int overflow */
      fd.fd_overflow = TRUE;
      return FALSE;
    }
  } else {
    struct dword lb;
    TAGGED tmax = dvar_max_t(dv);

    div_dld(&pdata->bigf,c,&lb);
    if (TagIsSmall(tmax)) {
      add_dtd(&lb,tmax,&lb);
    } else {
      struct dword tmp;
      div_dld(CMIN(elt),c,&tmp);
      add_ddd(&lb,&tmp,&lb);
    }
    if (issmall_d(&lb)) {/* representable lb */
      if (dvar_fix_min_l(dv,lb.lsw.asword)<0)
	return FALSE;
      mul_ltd(c,dvar_min_t(dv),CMAX(elt));
    } else if (cmp_dz(&lb)<0) {	/* lb underflow - can't prune the variable */
      long modc = mod_duu(&pdata->bigf,-c);
      
      add_ddd(&pdata->bigf,CMIN(elt),CMAX(elt)); /* ensure bigf >= I_i */
      if (modc>0)
	sub_dud(CMAX(elt),modc,CMAX(elt));
    } else {			/* lb overflow - signal int overflow */
      fd.fd_overflow = TRUE;
      return FALSE;
    }
  }
  sub_ddd(&cmax0,CMAX(elt),&decr);
  sub_ddd(INTERVAL(elt),&decr,INTERVAL(elt));
  sub_ddd(&pdata->bige,&decr,&pdata->bige);
  spheapify(pdata,0);
  if (dvar_is_integer(dv)) {
    sub_ddd(&pdata->rhs,CMAX(elt),&pdata->rhs);
    if (ABSCOEFF(elt)==1)
      pdata->units--;
    else
      pdata->nonunits--;
  }
  return TRUE;
}


static BOOL
scalar_product_le_fast MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                              struct linear_data *pdata,
                              TERM elt)
{
  long c = COEFF(elt);
  Dvar dv = DVAR(elt);
  long decr;
  long cmax0 = ASLONG(CMAX(elt));
  long cmax;

  if (ASLONG(&pdata->bigf)<0)
    return FALSE;
  if (c>0) {
    long ub = ASLONG(&pdata->bigf)/c + dvar_min_l(dv);
    if (dvar_fix_max_l(dv,ub)<0)
      return FALSE;
    cmax = c*dvar_max_l(dv);
  } else {
    long lb = -(ASLONG(&pdata->bigf)/(-c)) + dvar_max_l(dv);
    if (dvar_fix_min_l(dv,lb)<0)
      return FALSE;
    cmax = c*dvar_min_l(dv);
  }
  add_zld(cmax,CMAX(elt));
  decr = cmax0-cmax;
  ASLONG(INTERVAL(elt)) -= decr;
  spheapify(pdata,0);
  sub_dld(&pdata->bige,decr,&pdata->bige);
  if (dvar_is_integer(dv)) {
    sub_dld(&pdata->rhs,cmax,&pdata->rhs);
    if (ABSCOEFF(elt)==1)
      pdata->units--;
    else
      pdata->nonunits--;
  }
  return TRUE;
}

static BOOL
scalar_product_ge MAGIC(HIDDEN_PROTO
			struct linear_data *pdata,
			TERM elt)
{
  long c = COEFF(elt);
  Dvar dv = DVAR(elt);
  struct dword decr;
  struct dword cmin0;

  add_zdd(CMIN(elt),&cmin0);
  
  /*
    Phase 2:

    For >=
    ******
    bige>=0 is a necessary condition.
    bigf<=0 is a sufficient condition.

    enforce:
    bige >= I_i for all i

    rules:
    x_i >= -floor(E / a_i) + max(x_i)  if a_i>0
    x_i <=  floor(E /-a_i) + min(x_i)  if a_i<0
  */

  if (cmp_dz(&pdata->bige)<0)
    return FALSE;
  if (c>0) {
    struct dword lb;
    TAGGED tmax = dvar_max_t(dv);

    div_dld(&pdata->bige,-c,&lb);
    if (TagIsSmall(tmax)) {
      add_dtd(&lb,tmax,&lb);
    } else {
      struct dword tmp;
      div_dld(CMAX(elt),c,&tmp);
      add_ddd(&lb,&tmp,&lb);
    }
    if (issmall_d(&lb)) {/* representable lb */
      if (dvar_fix_min_l(dv,lb.lsw.asword)<0)
	return FALSE;
      mul_ltd(c,dvar_min_t(dv),CMIN(elt));
    } else if (cmp_dz(&lb)<0) {	/* lb underflow - can't prune the variable */
      long modc = mod_duu(&pdata->bige,c);
      
      sub_ddd(CMAX(elt),&pdata->bige,CMIN(elt)); /* ensure bige >= I_i */
      if (modc>0)
	add_dud(CMIN(elt),modc,CMIN(elt));
    } else {			/* lb overflow - signal int overflow */
      fd.fd_overflow = TRUE;
      return FALSE;
    }
  } else {
    struct dword ub;
    TAGGED tmin = dvar_min_t(dv);

    div_dld(&pdata->bige,-c,&ub);
    if (TagIsSmall(tmin)) {
      add_dtd(&ub,tmin,&ub);
    } else {
      struct dword tmp;
      div_dld(CMAX(elt),c,&tmp);
      add_ddd(&ub,&tmp,&ub);
    }
    if (issmall_d(&ub)) {/* representable ub */
      if (dvar_fix_max_l(dv,ub.lsw.asword)<0)
	return FALSE;
      mul_ltd(c,dvar_max_t(dv),CMIN(elt));
    } else if (cmp_dz(&ub)>0) {	/* ub overflow - can't prune the variable */
      long modc = mod_duu(&pdata->bige,-c);
      
      sub_ddd(CMAX(elt),&pdata->bige,CMIN(elt)); /* ensure bige >= I_i */
      if (modc>0)
	add_dud(CMIN(elt),modc,CMIN(elt));
    } else {			/* ub underflow - signal int overflow */
      fd.fd_overflow = TRUE;
      return FALSE;
    }
  }
  sub_ddd(CMIN(elt),&cmin0,&decr);
  sub_ddd(INTERVAL(elt),&decr,INTERVAL(elt));
  sub_ddd(&pdata->bigf,&decr,&pdata->bigf);
  spheapify(pdata,0);
  if (dvar_is_integer(dv)) {
    sub_ddd(&pdata->rhs,CMIN(elt),&pdata->rhs);
    if (ABSCOEFF(elt)==1)
      pdata->units--;
    else
      pdata->nonunits--;
  }
  return TRUE;
}

static BOOL
scalar_product_ge_fast MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                              struct linear_data *pdata,
                              TERM elt)
{
  long c = COEFF(elt);
  Dvar dv = DVAR(elt);
  long decr;
  long cmin0 = ASLONG(CMIN(elt));
  long cmin;

  if (ASLONG(&pdata->bige)<0)
    return FALSE;
  if (c>0) {
    long lb = -(ASLONG(&pdata->bige)/c) + dvar_max_l(dv);
    if (dvar_fix_min_l(dv,lb)<0)
      return FALSE;
    cmin = c*dvar_min_l(dv);
  } else {
    long ub = ASLONG(&pdata->bige)/(-c) + dvar_min_l(dv);
    if (dvar_fix_max_l(dv,ub)<0)
      return FALSE;
    cmin = c*dvar_max_l(dv);
  }
  add_zld(cmin,CMIN(elt));
  decr = cmin-cmin0;
  ASLONG(INTERVAL(elt)) -= decr;
  spheapify(pdata,0);
  sub_dld(&pdata->bigf,decr,&pdata->bigf);
  if (dvar_is_integer(dv)) {
    sub_dld(&pdata->rhs,cmin,&pdata->rhs);
    if (ABSCOEFF(elt)==1)
      pdata->units--;
    else
      pdata->nonunits--;
  }
  return TRUE;
}

/* Preconditions:
   0<coeff<mod, 0=<rhs<mod, 0<mod

   Solve min X such that coeff*X = rhs (modulo mod)
*/
static long solve_gcd(long coeff, long rhs, long mod)
{
  if (rhs==0)
    return 0;
  else {
    long rhs1 = rhs%coeff;
    if (rhs1==0)
      return rhs/coeff;
    else
      return (rhs + mod*solve_gcd(mod%coeff, coeff-rhs1, coeff)) / coeff;
  }
}


/* Preconditions: 
   0<mod, gcdall = gcd(coeff,mod)

   Adjusts minx up and maxx down s.t.
   Returns smallest s.t. coeff*minx = coeff*maxx = rhs (modulo mod)
*/
static void adjust_bounds_gcd(long coeff, Dword rhs,
			      long mod, long gcdall,
			      long *minx, long *maxx)
{
  long minx0 = *minx, maxx0 = *maxx;
  long q = mod/gcdall;
  long r, x, s;
  struct dword rhslocal;

  if (coeff>0) {
    add_zdd(rhs,&rhslocal);
  } else {
    sub_zdd(rhs,&rhslocal);
    coeff = -coeff;
  }
  coeff = coeff % mod;
  s = solve_gcd(coeff, mod_duu(&rhslocal,mod), mod);
  r = minx0 % q;
  if (r<0) r += q;
  x = minx0 - r + s;
  if (x<minx0) x += q;
  *minx = x;
  r = maxx0 % q;
  if (r<0) r += q;
  x = maxx0 - r + s;
  if (x>maxx0) x -= q;
  *maxx = x;
}

/*
  '$fd_linear'(+State0, -State, -Actions).
  State = state(CX,Op,RHS,Nground,Handle,Stamp) where CX are all non-ground
  Op in [1,2,3,4] means [#=<,#>=,#=,#\=].
*/
void SPCDECL
prolog_fd_linear MAGIC (HIDDEN_PROTO
			SP_term_ref State0,
			SP_term_ref State,
			SP_term_ref Actions)
{
  WAMENV;
  TAGGED tvec, telt, handle, t1;
  BOOL post, committed, incremental; /* TRUE if state can't be backtracked over */
  int nvars, i;
  long state_stamp;
  int op, ent, total_size;
  int nonground = 0;
  char *ptr;
  struct linear_data *pdata;

  (void)State0;                 /* [PM] 3.9b5 avoid -Wunused */

  w->numstack_end = NULL;
  fd.fd_overflow = FALSE;
/*    X(0) = RefTerm(State0); */
  dvar_export_start(3);
  RefTerm(State) = unify_output_state(w,&handle,&state_stamp,&committed);

  if (!IsVar(handle)) {		/* got [Flag | '$free'(Ptr)] */
    post = FALSE;
    pdata = Pdata(struct linear_data,handle);
    nvars = pdata->nvars;
  } else {			/* build persistent state */
    post = TRUE;
    nvars = 0;
    DerefArg(tvec,X(0),1);	/* get CX0 */
    while (TagIsLST(tvec)) {	/* count terms */
      DerefCdr(tvec,tvec);
      nvars++;
    }
    total_size = 
      nvars*sizeof(struct dvar) +
      3*nvars*sizeof(struct dword) +
      3*nvars*sizeof(long) +
      2*nvars*sizeof(TERM) +
      nvars*sizeof(int);
  
    pdata = Palloc(struct linear_data, total_size, handle);
    ptr = (char *)(pdata+1);
  
    pdata->dvar = (Dvar)ptr;
    ptr += nvars*sizeof(struct dvar);
    pdata->term.cmin = (struct dword *)ptr;
    ptr += nvars*sizeof(struct dword);
    pdata->term.cmax = (struct dword *)ptr;
    ptr += nvars*sizeof(struct dword);
    pdata->term.interval = (struct dword *)ptr;
    ptr += nvars*sizeof(struct dword);
    pdata->term.coeff = (long *)ptr;
    ptr += nvars*sizeof(long);
    pdata->term.abscoeff = (long *)ptr;
    ptr += nvars*sizeof(long);
    pdata->term.gcd = (long *)ptr;
    ptr += nvars*sizeof(long);
    pdata->target = (TERM *)ptr;
    ptr += nvars*sizeof(TERM);
    pdata->heap = (TERM *)ptr;
    ptr += nvars*sizeof(TERM);
    pdata->vheap = (int *)ptr;
    ptr += nvars*sizeof(int);
#if DBG
    if (ptr != (char *)(pdata+1)+total_size)
      printf("SHOULDN'T HAPPEN: expected ptr=%p, got %p\n",
	     (char *)(pdata+1)+total_size, ptr);
#endif
    pdata->destructor = linear_destructor;
    pdata->daemon = linear_daemon;
    FD_STORE_SPENV(pdata->spenv);
    pdata->refbase = SP_alloc_globrefs(nvars<<1);
    pdata->stamp = state_stamp-1; /* don't trust initially */
    DerefArg(telt,X(0),2);
    pdata->op = op = GetSmall(telt);
    pdata->nvars = nvars;
    pdata->ntargets = nvars;
    DerefArg(tvec,X(0),1);	/* get CX0 */
    DerefArg(telt,X(0),3);	/* get RHS */
    add_ztd(telt,&pdata->rhs);
				/* xfer all info to the struct linear_terms */
    for (i=0; i<nvars; i++) {
      TERM elt = i;
      long c;
      
      DerefCar(telt,tvec);
      DerefCdr(tvec,tvec);
      DerefArg(t1,telt,1);
      COEFF(elt) = c = GetSmall(t1);
      ABSCOEFF(elt) = ABS(c);
      DerefArg(t1,telt,2);	/* get domain var */
      RefGlob(RefVar(elt)) = t1;
      DerefArg(t1,telt,3);	/* get attribute */
      RefGlob(RefAttr(elt)) = t1;
      SV(i) = elt;
    }
  }

				/* RESUME HERE */
  fd.gdata = pdata;
  if (state_stamp != pdata->stamp) {
    DerefArg(telt,X(0),3);
    add_ztd(telt,&pdata->rhs);
    DerefArg(telt,X(0),4);
    pdata->ntargets = nvars-GetSmall(telt);
    incremental = FALSE;
  } else
    incremental = TRUE;
  scalar_product_init(pdata,incremental);
  pdata->stamp = state_stamp+1;
  op = pdata->op;
  ent = -1;
				/* (dis)entailment tests based on GCD */
  switch (op) {
  case 3:
    refresh_gcd(pdata);	/* compute gcdall, GCD() */
    if (pdata->gcdall>1 && mod_duu(&pdata->rhs,pdata->gcdall)!=0)
      goto ret;
    break;
  case 4:
    refresh_gcd(pdata);	/* compute gcdall, GCD() */
    if (pdata->gcdall>1 && mod_duu(&pdata->rhs,pdata->gcdall)!=0) {
      ent = 1;
      goto ret;
    }
    break;
  }
 loop:
  switch (op) {
  case 1:			/* #=< */
    if (pdata->heapsize>0) {
      TERM elt = pdata->heap[0]; 
      if (cmp_dd(&pdata->bigf,INTERVAL(elt))<0) {
	if (pdata->fast ?
	    !scalar_product_le_fast(pdata,elt) :
	    !scalar_product_le(pdata,elt))
	  goto ret;
	goto loop;
      }
    }
    break;
  case 2:			/* #>= */
    if (pdata->heapsize>0) {
      TERM elt = pdata->heap[0]; 
      if (cmp_dd(&pdata->bige,INTERVAL(elt))<0) {
	if (pdata->fast ?
	    !scalar_product_ge_fast(pdata,elt) :
	    !scalar_product_ge(pdata,elt))
	  goto ret;
	goto loop;
      }
    }
    break;
  case 3:			/* #= */
    if (pdata->heapsize>0) {
      TERM elt = pdata->heap[0]; 
      if (cmp_dd(&pdata->bigf,INTERVAL(elt))<0) {
	if (pdata->fast ?
	    !scalar_product_le_fast(pdata,elt) :
	    !scalar_product_le(pdata,elt))
	  goto ret;
	goto loop;
      }
      if (cmp_dd(&pdata->bige,INTERVAL(elt))<0) {
	if (pdata->fast ?
	    !scalar_product_ge_fast(pdata,elt) :
	    !scalar_product_ge(pdata,elt))
	  goto ret;
	goto loop;
      }
      refresh_gcd(pdata);	/* compute gcdall, GCD() */
      if (mod_duu(&pdata->rhs,pdata->gcdall) != 0)
	goto ret;
      if (pdata->units<=1)
	for (i=0; i<pdata->ntargets; i++) {
	  TERM elt = SV(i);
	  Dvar dv = DVAR(elt);
	  TAGGED tmin = dvar_min_t(dv);
	  TAGGED tmax = dvar_max_t(dv);

	  if (AreSmall(tmin,tmax) && tmin!=tmax && GCD(elt)>pdata->gcdall) {
	    /* Ensure that:

	       Ai * min(Xi) - RHS = Ai * max(Xi) - RHS = 0 (modulo G)

	       where G is the gcd of all coefficients excepts Ai.
	       Fails if an empty domain is produced.
	    */
	    long imin = GetSmall(tmin);
	    long imax = GetSmall(tmax);
	    int rc;

	    adjust_bounds_gcd(COEFF(elt),
				  &pdata->rhs,
				  GCD(elt),
				  pdata->gcdall,
				  &imin, &imax);
	    rc = dvar_fix_interval_l(dv, imin, imax);
	    if (rc<0)
	      goto ret;
	    if (rc>0) {
	      struct dword cmin0;
	      struct dword cmax0;
	      struct dword decr;
	      long c = COEFF(elt);

	      add_zdd(CMIN(elt),&cmin0);
	      add_zdd(CMAX(elt),&cmax0);

	      if (c>0) {
		mul_ltd(c,dvar_min_t(dv),CMIN(elt));
		mul_ltd(c,dvar_max_t(dv),CMAX(elt));
	      } else {
		mul_ltd(c,dvar_max_t(dv),CMIN(elt));
		mul_ltd(c,dvar_min_t(dv),CMAX(elt));
	      }
	      sub_ddd(CMAX(elt),CMIN(elt),INTERVAL(elt));
	      sub_ddd(&cmax0,CMAX(elt),&decr);
	      sub_ddd(&pdata->bige,&decr,&pdata->bige);
	      sub_ddd(CMIN(elt),&cmin0,&decr);
	      sub_ddd(&pdata->bigf,&decr,&pdata->bigf);
	      spheapify(pdata,pdata->vheap[elt]);
	      if (dvar_is_integer(dv)) {
		sub_ddd(&pdata->rhs,CMIN(elt),&pdata->rhs);
		if (ABSCOEFF(elt)==1)
		  pdata->units--;
		else
		  pdata->nonunits--;
	      }
	      goto loop;
	    }
	  }
	}
    }
    break;
  case 4:			/* #\= */
    nonground = pdata->units+pdata->nonunits;
    if (nonground==0 && cmp_dz(&pdata->rhs)==0)
      goto ret;
    else if (nonground==1) {
      for (i=0; cmp_dz(INTERVAL(SV(i)))==0; i++)
	;
      {
	TERM elt = SV(i);
	Dvar dv = DVAR(elt);

	if (mod_duu(&pdata->rhs,ABSCOEFF(elt))==0) { /* RHS a multiple of coefficient */
	  dvar_prune_value_l(dv,div_dll(&pdata->rhs,COEFF(elt)));
	}
      }
    }
    break;
  }
  switch (op) {
  case 1:
    ent = (cmp_dz(&pdata->bigf)<0 ? -1 : cmp_dz(&pdata->bige)<=0);
    break;
  case 2:
    ent = (cmp_dz(&pdata->bige)<0 ? -1 : cmp_dz(&pdata->bigf)<=0);
    break;
  case 3:
    ent = (cmp_dz(&pdata->bigf)<0 ? -1 :
	   cmp_dz(&pdata->bige)<0 ? -1 : 
	   cmp_dz(&pdata->bigf)>0 ? 0 :
	   cmp_dz(&pdata->bige)>0 ? 0 : 1);
    break;
  case 4:
    ent = (nonground<=1);
    break;
  }

  if (ent>=0) {
    int inf = 0;
    int sup = pdata->ntargets-1;
    TERM held = SV(sup); /* sup is the hole */
    TERM current = SV(inf);
    
    while (inf<=sup) {
      Dvar dv = DVAR(current);
      
      dvar_pruning_done(dv);
      dvar_export(dv, 3);
      if (!dvar_is_integer(dv)) {

	if (post && ent==0) {
	  TAGGED functor;
	  switch (op) {
	  case 1:
	    functor = (COEFF(current)>0 ? fd.functor_min : fd.functor_max);
	    break;
	  case 2:
	    functor = (COEFF(current)<0 ? fd.functor_min : fd.functor_max);
	    break;
	  case 3:
	    functor = fd.functor_minmax;
	    break;
	  case 4:
	    functor = fd.functor_val;
	    break;
          default:              /* [PM] 3.10 tell compiler functor is always initialized  */
	    functor = fd.functor_val;
            #if DBG
            abort();            /* should not happen */
            #endif
	    break;
	  }
	  attach_daemon(pdata, RefAttr(current), RefVar(current),
			X(1), functor, 3);
	}

	SV(inf) = current;
	inf++;
	current = (inf>=sup ? held : SV(inf));
      } else {
	SV(sup) = current;
	sup--;
	current = (inf>=sup ? held : SV(sup));
      }
    }
    pdata->ntargets = inf;
  }
  if (ent==0 && op==3 && pdata->ntargets==2 && cmp_dz(&pdata->rhs)==0) {
    TERM elt0 = SV(0);
    TERM elt1 = SV(1);

    if (COEFF(elt0) == -COEFF(elt1)) {
      dvar_export_equal(DVAR(elt0), DVAR(elt1), 3);
      ent = 1;
    }
  }
  RequireHeap(4,3);
  if (islong_d(&pdata->rhs))
    CTagToArg(X(0),3) = MakeInteger((&pdata->rhs)->lsw.asword);
  else {			/* make a 64-bit bignum */
    TAGGED *h = w->global_top;
    h[0] = TagIndex(TAG_LIN,3);
    h[1] = (&pdata->rhs)->lsw.asuword;
    h[2] = (&pdata->rhs)->msw.asuword;
    h[3] = TagIndex(TAG_LIN,3);
    w->global_top = h+4;
    CTagToArg(X(0),3) = MakeStructure(h);
  }
  CTagToArg(X(0),4) = MakeSmall(nvars-pdata->ntargets);
 ret:
  Pfree(ent==1,handle,pdata);
  dvar_export_done(Actions, ent, 3);
  if (!fd_check_overflow())
    SP_fail();
}

/*** knapsack/3 ***/

/*** support for sparse arrays repr. as hash tables ***/

typedef struct ARITER_struct {
  int next, max;
  struct sw_on_key *sw;
  struct sw_on_key_node *cur;
} ARITER;


#define ariter_empty(it) ((it)->next>=(it)->max)

INLINE static void
ariter_init(ARITER *it, struct sw_on_key *sw)
{
  int i;

  it->sw = sw;
  it->cur = NULL;
  it->next = -1;
  it->max = SwitchSize(sw);
  for (i=0; i<it->max && sw->tab.asnode[i].value.arities==0; i++)
    ;
  it->next = i;
}


INLINE static struct sw_on_key_node *
ariter_next(ARITER *it)
{
  int i;

  it->cur = &it->sw->tab.asnode[it->next];
  for (i=it->next+1; i<it->max && it->sw->tab.asnode[i].value.arities==0; i++)
    ;
  it->next = i;
  return it->cur;
}


INLINE static void
clear_array(struct sw_on_key *sw)
{
  int i = SwitchSize(sw);

  for (--i; i>=0; --i) {
    sw->tab.asnode[i].key = 0;
    sw->tab.asnode[i].value.arities = 0;
  }
}


/* ripped off from expand_sw_on_key */
static void
maybe_rehash MAGIC (HIDDEN_PROTO
		    struct sw_on_key **psw)
{
  struct sw_on_key_node *h1;
  int size = SwitchSize(*psw);
  int newsize = 4;
  int pop = 0;
  struct sw_on_key *newsw;
  int j;
  
  if (size==4)
    return;
  for (j=0; j<size; j++) {
    h1 = &(*psw)->tab.asnode[j];
    if (h1->key && h1->value.try_node)
      pop++;
  }
  for (j=3; pop>j; j<<=1)
    newsize<<=1;
  if (newsize==size)
    return;
  newsw = new_switch_on_key(newsize,NULL);
  newsw->count = pop;
  for (j=0; j<size; j++) {
    TAGGED t;
    /*
    struct try_node *p = NULL;
    struct sw_on_key_node *h2;
    */
    h1 = &(*psw)->tab.asnode[j];
    if ((t=h1->key) && (/*p=*/h1->value.try_node)) {
      *incore_gethash(newsw,t) = *h1;
      /*
	h2 = incore_gethash(newsw,t);
	h2->key = t;
	h2->value.try = p;
      */
    }
  }
  dispose_switch_on_key(*psw);
  *psw = newsw;
}

struct knapsack_data {
  void (SPCDECL *destructor)(void *);
#if MULTI_SP_AWARE
  SPEnv *spenv;
#endif /* MULTI_SP_AWARE */

  SP_globref refbase;
  long stamp;			/* increases up to backtracking */
  long rhs;			/* RHS - valid up to backtracking */
  int nvars;			/* #terms */
  int ntargets;			/* #terms that may be targets */
  Dvar dvar;
  TERM *target;
  struct {
    long offset;
    long lb;
    long ub;
    int nvars;
    int fanout;
    TERM *var;
  } dp;
  struct {
    long *coeff;		/* ai */
    long *cmin;			/* min(ai*xi) */
    long *cmax;			/* max(ai*xi) */
    /* fields for knapsack filtering */
    struct sw_on_key **fanin;
    struct sw_on_key **fanout;
    struct sw_on_key **support;
    long *offset;
  } term;
  /*
  space for arrays
  */
};

#undef  CMIN
#undef  CMAX
#define CMIN(t) (pdata->term.cmin[t])
#define CMAX(t) (pdata->term.cmax[t])
#define FANIN(t) (pdata->term.fanin[t])
#define FANOUT(t) (pdata->term.fanout[t])
#define SUPPORT(t) (pdata->term.support[t])
#define OFFSET(t) (pdata->term.offset[t])

#define AREF(Array,Index) (incore_gethash(Array,MakeSmall(Index))->value.arities)
#define ASET(Array,Index,Value) (dyn_puthash(&(Array),MakeSmall(Index))->value.arities=(Value))
#define AINC(Array,Index,Inc) (dyn_puthash(&(Array),MakeSmall(Index))->value.arities+=(Inc))

static void free_one_dp MAGIC (HIDDEN_PROTO
			       struct knapsack_data *pdata,TERM t)
{
  dispose_switch_on_key(FANIN(t));
  dispose_switch_on_key(FANOUT(t));
  dispose_switch_on_key(SUPPORT(t));
}

static void free_all_dp MAGIC (HIDDEN_PROTO
			       struct knapsack_data *pdata)
{
  int i;
  
  for (i=0; i<pdata->dp.nvars; i++)
    free_one_dp(pdata,pdata->dp.var[i]);
  pdata->dp.nvars = 0;
  pdata->dp.fanout = 0;
}

static void SPCDECL knapsack_destructor(void *pdata_v)
{
  struct knapsack_data *pdata = (struct knapsack_data *)pdata_v;
  FD_SETUP_SPENV(pdata->spenv)

  free_all_dp(pdata);
  SP_free_globrefs(pdata->refbase,(pdata->nvars<<1)+2);
  SP_free(pdata);
}

static void knapsack_filter_noninc MAGIC (HIDDEN_PROTO
					  struct knapsack_data *pdata,
					  long base, long lb, long ub)
{
  long nvars = pdata->ntargets;
  FDITER it;
  ARITER ai;
  long v, cminsum, cmaxsum;
  int i;
  TERM last = SV(nvars-1);

  free_all_dp(pdata);
  for (i=0; i<nvars; i++)
    pdata->dp.var[i] = SV(i);

  for (i=0; i<nvars; i++) {
    TERM elt = pdata->dp.var[i];
    
    FANIN(elt) = new_switch_on_key(4, NULL);
    FANOUT(elt) = new_switch_on_key(4, NULL);
    SUPPORT(elt) = new_switch_on_key(4, NULL);
    OFFSET(elt) = 0;
  }

  /* forward phase */
  cminsum = cmaxsum = 0;
  for (i=0; i<nvars; i++) {
    cminsum += CMIN(SV(i));
    cmaxsum += CMAX(SV(i));
  }
  for (i=0; i<nvars; i++) {
    TERM prev = pdata->dp.var[i-1];
    TERM elt = pdata->dp.var[i];
    Dvar dv = DVAR(elt);
    long c = COEFF(elt);
    long d, cd;
    TAGGED td;
    TAGGED tmin=dvar_min_t(dv);
    TAGGED tmax=dvar_max_t(dv);
    long lb1, ub1, lb2, ub2;

    cminsum -= CMIN(SV(i));
    cmaxsum -= CMAX(SV(i));
    lb1 = lb - cmaxsum;
    ub1 = ub - cminsum;

    if (i==0) {
      lb2 = CEILDIV(lb1-base,c);
      ub2 = FLOORDIV(ub1-base,c);
      fditer_init(&it, dvar_set(dv));
      while (!fditer_empty(&it)) {
	td = fditer_next(&it);
	d = GetSmall(td);
	if (Tlt(td,tmin) || d<lb2) continue;
	if (Tgt(td,tmax) || d>ub2) break;
	cd = c*d;
	pdata->dp.fanout++;
	ASET(FANIN(elt),base+cd,1);
	ASET(SUPPORT(elt),d,1);
      }
    } else {
      ariter_init(&ai, FANIN(prev));
      while (!ariter_empty(&ai)) {
	v = GetSmall(ariter_next(&ai)->key);
	lb2 = CEILDIV(lb1-v,c);
	ub2 = FLOORDIV(ub1-v,c);
	fditer_init(&it, dvar_set(dv));
	while (!fditer_empty(&it)) {
	  td = fditer_next(&it);
	  d = GetSmall(td);
	  if (Tlt(td,tmin) || d<lb2) continue;
	  if (Tgt(td,tmax) || d>ub2) break;
	  cd = c*d;
	  AINC(FANOUT(prev),v,1);
	  AINC(FANIN(elt),v+cd,1);
	  AINC(SUPPORT(elt),d,1);
	}
      }
    }
  }

  /* final fanouts */
  pdata->dp.lb = ub+1;
  pdata->dp.ub = lb-1;
  ariter_init(&ai, FANIN(last));
  while (!ariter_empty(&ai)) {
    v = GetSmall(ariter_next(&ai)->key);
    if (v>=lb && v<=ub) {
      ASET(FANOUT(last),v,1);	/* maintain invariant */
      if (pdata->dp.lb>v) pdata->dp.lb = v;
      if (pdata->dp.ub<v) pdata->dp.ub = v;
    }
  }
  pdata->dp.offset = base;
  pdata->dp.nvars = nvars;
}



static void knapsack_filter_inc MAGIC (HIDDEN_PROTO
				       struct knapsack_data *pdata,
				       long lb, long ub)
{
  long nvars = pdata->dp.nvars;
  long base = pdata->dp.offset;
  FDITER it;
  ARITER ai;
  struct sw_on_key_node *keyval;
  long v;
  int i;
  TERM last = pdata->dp.var[nvars-1];
  long offset = OFFSET(last);

  /* delete arcs for pruned values */
  for (i=0; i<nvars; i++) {
    TERM prev = pdata->dp.var[i-1];
    TERM elt = pdata->dp.var[i];
    Dvar dv = DVAR(elt);
    long c = COEFF(elt);
    long d, cd;
    TAGGED td;
    TAGGED tmin=dvar_min_t(dv);
    TAGGED tmax=dvar_max_t(dv);
    TAGGED told=EmptySet;
    TAGGED tdel;
    
    /* Compute old domain */
    ariter_init(&ai, SUPPORT(elt));
    while (!ariter_empty(&ai)) {
      v = GetSmall(ariter_next(&ai)->key);
      told = fd_insert_into(MakeSmall(v),told);
    }
    tdel = fd_subtract(told,fd_and_interval(dvar_set(dv),tmin,tmax));

    if (i==0) {
      fditer_init(&it, tdel);
      while (!fditer_empty(&it)) {
	td = fditer_next(&it);
	d = GetSmall(td);
	--pdata->dp.fanout;
	ASET(FANIN(elt),base+c*d,0);
	ASET(SUPPORT(elt),d,0);
      }
    } else {
      long offset = OFFSET(prev);
      ariter_init(&ai, FANOUT(prev));
      while (!ariter_empty(&ai)) {
	keyval = ariter_next(&ai);
	v = GetSmall(keyval->key);
	if (AREF(FANIN(prev),v)>0) {
	  if (tdel==EmptySet) continue;
	  fditer_init(&it, tdel);
	} else
	  fditer_init(&it, told);
	while (!fditer_empty(&it)) {
	  td = fditer_next(&it);
	  d = GetSmall(td);
	  cd = c*d+offset;
	  if (v+cd > pdata->dp.ub)
	    break;
	  if (AREF(FANIN(elt),v+cd)>0) {
	    keyval->value.arities--;
	    AINC(FANIN(elt),v+cd,-1);
	    AINC(SUPPORT(elt),d,-1);
	  }
	}
      }
    }
  }

  /* final fanouts */
  clear_array(FANOUT(last));
  pdata->dp.lb = ub+1;
  pdata->dp.ub = lb-1;
  ariter_init(&ai, FANIN(last));
  while (!ariter_empty(&ai)) {
    v = GetSmall(ariter_next(&ai)->key);
    if (v>=lb-offset && v<=ub-offset) {
      ASET(FANOUT(last),v,1);
      if (pdata->dp.lb>v+offset) pdata->dp.lb = v+offset;
      if (pdata->dp.ub<v+offset) pdata->dp.ub = v+offset;
    }
  }
}


/* Dynamic programming based propagator. */
/* -1=disentailed, 0=true, 1=entailed */
static int knapsack_filter MAGIC (HIDDEN_PROTO
				  struct knapsack_data *pdata,
                                  long base, 
                                  BOOL incremental,
                                  long *newlb, long *newub, int *max_prunedp)
{
  int ent = -1, nvars;
  BOOL pruned = FALSE;
  long lb = *newlb;
  long ub = *newub;
  FDITER it;
  ARITER ai;
  struct sw_on_key_node *keyval;
  long v;
  int i;

  if (incremental) {
    knapsack_filter_inc(pdata,lb,ub);
  } else {
    knapsack_filter_noninc(pdata,base,lb,ub);
  }
  *newlb = lb = pdata->dp.lb;
  *newub = ub = pdata->dp.ub;
  nvars = pdata->dp.nvars;
  base = pdata->dp.offset;
  
  /* backward phase */
  for (i=nvars-1; i>=0; i--) {
    TERM prev = pdata->dp.var[i-1];
    TERM elt = pdata->dp.var[i];
    Dvar dv = DVAR(elt);
    long c = COEFF(elt);
    long offset = i==0 ? 0 : OFFSET(prev);
    long d, cd;
    TAGGED td;
    TAGGED tmin=dvar_min_t(dv);
    TAGGED tmax=dvar_max_t(dv);

    ariter_init(&ai, FANIN(elt));
    while (!ariter_empty(&ai)) {
      keyval = ariter_next(&ai);
      v = GetSmall(keyval->key);
      if (AREF(FANOUT(elt),v)==0) {
	keyval->value.arities = 0;
	fditer_init(&it, dvar_set(dv));
	while (!fditer_empty(&it)) {
	  td = fditer_next(&it);
	  if (Tlt(td,tmin)) continue;
	  if (Tgt(td,tmax)) break;
	  d = GetSmall(td);
	  cd = c*d+offset;
	  if (v-cd<base) continue;
	  if (i==0) {
	    if (v-cd == base) {
	      ASET(SUPPORT(elt),d,0);
	      --pdata->dp.fanout;
	      break;
	    }
	  } else
	    if (AREF(FANOUT(prev),v-cd)>0) {
	      AINC(SUPPORT(elt),d,-1);
	      AINC(FANOUT(prev),v-cd,-1);
	    }
	}
      }
    }
  }
  if (pdata->dp.fanout==0)
    goto ret;

  /* pruning */
  for (i=0, nvars=0; i<pdata->dp.nvars; i++) {
    TERM prev = pdata->dp.var[nvars-1];
    TERM elt = pdata->dp.var[i];
    Dvar dv = DVAR(elt);
    long offset = nvars==0 ? base : OFFSET(prev);
    long d, cd=0;
    TAGGED td;
    TAGGED tmin=dvar_min_t(dv);
    TAGGED tmax=dvar_max_t(dv);
    int size=0;

    fditer_init(&it, dvar_set(dv));
    while (!fditer_empty(&it)) {
      td = fditer_next(&it);
      if (Tlt(td,tmin)) continue;
      if (Tgt(td,tmax)) break;
      d = GetSmall(td);
      if (AREF(SUPPORT(elt),d)==0) {
	pruned = TRUE;
	dvar_prune_value_t(dv,td);
	if (*max_prunedp<pdata->ntargets)
	  *max_prunedp = pdata->ntargets;
      } else {
	size++;
	cd = d*COEFF(elt) + offset;
      }
    }
    if (size==1) {		/* shunt elt */
      if (nvars==0) {
	pdata->dp.offset = base = cd+OFFSET(elt);
	pdata->dp.fanout = AREF(FANOUT(elt),cd);
      } else {
	OFFSET(prev) = cd+OFFSET(elt);
	/* not needed - old entries nonzero iff new entries nonzero */
	/* clear_array(FANOUT(prev)); */
	ariter_init(&ai, FANOUT(elt));
	while (!ariter_empty(&ai)) {
	  keyval = ariter_next(&ai);
	  v = GetSmall(keyval->key);
	  ASET(FANOUT(prev),v-cd,keyval->value.arities);
	}
      }
      free_one_dp(pdata,elt);
    } else {
      pdata->dp.var[nvars++] = elt;
      maybe_rehash(&FANIN(elt));
      maybe_rehash(&FANOUT(elt));
      maybe_rehash(&SUPPORT(elt));
    }
  }
  pdata->dp.nvars = nvars;
  ent = (nvars==0);

#if DBG > 1
  /* invariant 1: preserving flow */
  for (i=0; i<nvars; i++) {
    TERM prev = pdata->dp.var[i-1];
    TERM elt = pdata->dp.var[i];
    long fanout=0, fanin=0, support=0;

    if (i==0)
      fanout = pdata->dp.fanout;
    else
      for (v=base; v<=ub; v++)
	fanout += AREF(FANOUT(prev),v);
    for (v=base; v<=ub; v++)
      fanin += AREF(FANIN(elt),v);
    for (v=0; v<=ub-base; v++)
      support += AREF(SUPPORT(elt),v);

    if (fanin!=fanout || fanin!=support) {
      printf("* FANOUT(%d)=%d, FANIN(%d)=%d, SUPPORT(%d)=%d\n",
	     (int)prev,(int)fanout,(int)elt,(int)fanin,(int)elt,(int)support);
    }
  }

  /* invariant 2: zero in iff zero out */
  for (i=0; i<nvars; i++) {
    TERM elt = pdata->dp.var[i];
    
    for (v=base; v<=ub; v++)
      if (((AREF(FANIN(elt),v)==0)^(AREF(FANOUT(elt),v)==0))
	  /*|| AREF(FANIN(elt),v)<0*/)
	printf("* FANIN(%d,%d)=%d, FANOUT(%d,%d)=%d\n",
	       (int)elt,(int)v,(int)AREF(FANIN(elt),v),(int)elt,(int)v,(int)AREF(FANOUT(elt),v));
  }

  /* invariant 3: correct fanin counts */
  for (i=0; i<nvars; i++) {
    TERM prev = pdata->dp.var[i-1];
    TERM elt = pdata->dp.var[i];
    TAGGED told = EmptySet;
    long c = COEFF(elt);
    long offset = i==0 ? 0 : OFFSET(prev);
    long d, cd;
    TAGGED td;
    
    for (v=0; v<=ub-base; v++)
      if (AREF(SUPPORT(elt),v)>0)
	told = fd_insert_into(MakeSmall(v),told);

    for (v=base; v<=ub; v++) {
      int fanin = 0;
      
      if (AREF(FANIN(elt),v)==0) continue;
      if (i==0) {
	if ((v-base)%c==0 && AREF(SUPPORT(elt),(v-base)/c)>0)
	  fanin++;
      } else {
	fditer_init(&it, told);
	while (!fditer_empty(&it)) {
	  td = fditer_next(&it);
	  d = GetSmall(td);
	  cd = c*d+offset;
	  if (v-cd>=base && AREF(FANOUT(prev),v-cd)>0)
	    fanin++;
	}
      }
      if (AREF(FANIN(elt),v)!=fanin)
	printf("* FANIN(%d,%d)=%d, #predecessors=%d\n",
	       (int)elt,(int)v,(int)AREF(FANIN(elt),v),fanin);
    }
  }
#endif

  if (ent==0) {
    int inf = 0;
    int sup = pdata->ntargets-1;
    TERM held = SV(sup); /* sup is the hole */
    TERM current = SV(inf);
    
    while (inf<=sup) {
      Dvar dv = DVAR(current);
      
      if (!dvar_is_integer(dv)) {
	SV(inf) = current;
	inf++;
	current = (inf>=sup ? held : SV(inf));
      } else {
	pdata->rhs -= COEFF(current)*dvar_min_l(dv);
	SV(sup) = current;
	sup--;
	current = (inf>=sup ? held : SV(sup));
      }
    }
    pdata->ntargets = inf;
  }
 ret:
  return ent;
}


/*
  '$fd_knapsack'(+State0, -State, -Actions).
  State = state(CX,RHS,Con,Nground,Handle,Stamp) where CX are all non-ground
  The constraint to be maintained is: CX+Con in [min(RHS),max(RHS)]
*/
void SPCDECL
prolog_fd_knapsack MAGIC (HIDDEN_PROTO
                          SP_term_ref State0,
                          SP_term_ref State,
                          SP_term_ref Actions)
{
  WAMENV;
  TAGGED tvec, telt, handle, t1;
  int max_pruned = 0;		/* pdata->ntargets at last pruning */
  BOOL committed;		/* TRUE if state can't be backtracked over */
  BOOL incremental;		/* TRUE if dom, min, max, cmin, cmax need refreshing */
  int nvars=0, i;
  long newlb, newub, state_stamp;
  int ent, total_size;
  TERM elt;
  struct knapsack_data *pdata;
  char *ptr;

  (void)State0;                 /* [PM] 3.9b5 avoid -Wunused */

  w->numstack_end = NULL;
  fd.fd_overflow = FALSE;

/*    X(0) = RefTerm(State0); */
  dvar_export_start(3);
  RefTerm(State) = unify_output_state(w,&handle,&state_stamp,&committed);

  if (!IsVar(handle)) {		/* got [Flag | '$free'(Ptr)] */
    pdata = Pdata(struct knapsack_data,handle);
    nvars = pdata->nvars;
  } else {			/* build persistent state */
    DerefArg(tvec,X(0),1);	/* get CX0 */
    while (TagIsLST(tvec)) {	/* count terms */
      DerefCdr(tvec,tvec);
      nvars++;
    }
    DerefArg(tvec,X(0),1);	/* get CX0 */
    DerefArg(telt,X(0),3);	/* get Con */
    total_size = 
      (nvars+1)*sizeof(struct dvar) +
      4*nvars*sizeof(long) +
      3*nvars*sizeof(int *) +
      2*nvars*sizeof(TERM);
    pdata = Palloc(struct knapsack_data, total_size, handle);
    ptr = (char *)(pdata+1);
  
    pdata->dvar = (Dvar)ptr;
    ptr += (nvars+1)*sizeof(struct dvar);
    pdata->target = (TERM *)ptr;
    ptr = (char *)(pdata->target+nvars);
    pdata->dp.var = (TERM *)ptr;
    ptr = (char *)(pdata->dp.var+nvars);
    /* arrays for term variables */
    pdata->term.coeff = (long *)ptr;
    ptr = (char *)(pdata->term.coeff+nvars);
    pdata->term.cmin = (long *)ptr;
    ptr = (char *)(pdata->term.cmin+nvars);
    pdata->term.cmax = (long *)ptr;
    ptr = (char *)(pdata->term.cmax+nvars);
    pdata->term.fanin = (struct sw_on_key **)ptr;
    ptr = (char *)(pdata->term.fanin+nvars);
    pdata->term.fanout = (struct sw_on_key **)ptr;
    ptr = (char *)(pdata->term.fanout+nvars);
    pdata->term.support = (struct sw_on_key **)ptr;
    ptr = (char *)(pdata->term.support+nvars);
    pdata->term.offset = (long *)ptr;
    ptr = (char *)(pdata->term.offset+nvars);
#if DBG
    if (ptr != (char *)(pdata+1)+total_size)
      printf("SHOULDN'T HAPPEN: expected ptr=%p, got %p\n",
	     (char *)(pdata+1)+total_size, ptr);
#endif
    pdata->destructor = knapsack_destructor;
    FD_STORE_SPENV(pdata->spenv);
    pdata->refbase = SP_alloc_globrefs((nvars<<1)+2);
    pdata->stamp = state_stamp-1; /* don't trust initially */
    pdata->nvars = nvars;
    pdata->ntargets = nvars;
    pdata->rhs = GetSmall(telt);
    pdata->dp.nvars = 0;
    pdata->dp.fanout = 0;
				/* xfer all info to the struct */
    for (i=0; i<nvars; i++) {
      elt = i;
      DerefCar(telt,tvec);
      DerefCdr(tvec,tvec);
      DerefArg(t1,telt,1);
      COEFF(elt) = GetSmall(t1);
      DerefArg(t1,telt,2);	/* get domain var */
      RefGlob(RefVar(elt)) = t1;
      DerefArg(t1,telt,3);	/* get attribute */
      RefGlob(RefAttr(elt)) = t1;
      SV(i) = elt;
    }
    DerefArg(telt,X(0),2);	/* get RHS */
    DerefArg(t1,telt,1);	/* get domain var */
    RefGlob(RefVar(nvars)) = t1;
    DerefArg(telt,telt,2);	/* get attribute */
    RefGlob(RefAttr(nvars)) = telt;
  }
				/* RESUME HERE */
  fd.gdata = pdata;
  incremental = (state_stamp == pdata->stamp);
  if (!incremental) {
    DerefArg(t1,X(0),3);
    pdata->rhs = GetSmall(t1);
    DerefArg(t1,X(0),4);
    pdata->ntargets = nvars-GetSmall(t1);
  }
  pdata->stamp = state_stamp+1;

  for (i=0; i<pdata->ntargets; i++) {
    long c;
    TERM elt = SV(i);
    Dvar dv = DVAR(elt);

    c = COEFF(elt);
    dvar_init(dv, RefAttr(elt), RefVar(elt));
    if (c>0) {
      CMIN(elt) = c*dvar_min_l(dv);
      CMAX(elt) = c*dvar_max_l(dv);
    } else {
      CMAX(elt) = c*dvar_min_l(dv);
      CMIN(elt) = c*dvar_max_l(dv);
    }
  }

  dvar_init(DVAR(nvars), RefAttr(nvars), RefVar(nvars));
  if (pdata->ntargets>0) {
    newlb = dvar_min_l(DVAR(nvars));
    newub = dvar_max_l(DVAR(nvars));
    ent = knapsack_filter(pdata,
			  -pdata->rhs,
			  incremental,
			  &newlb, &newub, &max_pruned);
  } else {
    ent = 1;
    newlb = newub = -pdata->rhs;
  }
  if (ent < 0)
    goto ret;
  for (i=0; i<max_pruned; i++) {
    elt = SV(i);
    dvar_pruning_done(DVAR(elt));
  }

  dvar_fix_interval_l(DVAR(nvars), newlb, newub);
  for (i=0; i<max_pruned; i++) {
    elt = SV(i);
    dvar_export(DVAR(elt), 3);
  }
  dvar_export(DVAR(nvars), 3);
  CTagToArg(X(0),3) = MakeSmall(pdata->rhs);
  CTagToArg(X(0),4) = MakeSmall(nvars-pdata->ntargets);
 ret:
  Pfree(ent==1,handle,pdata);
  dvar_export_done(Actions, ent, 3);
  if (!fd_check_overflow())
    SP_fail();
}

#define xmin dvar_min_t(dvx)
#define xmax dvar_max_t(dvx)
#define ymin dvar_min_t(dvy)
#define ymax dvar_max_t(dvy)
#define zmin dvar_min_t(dvz)
#define zmax dvar_max_t(dvz)

#define CEILSQRT \
  { \
    long m = GetSmall(ymin); \
    if (m>=3) \
      {			/* m = ceiling(sqrt(m)) */ \
	long n1 = m>>1, n2 = m; \
	while (n1<m) \
	  m = n1, \
	  n1 = (n1 + n2/n1)>>1; \
	if (m<n1 || m*m < n2) \
	  m++; \
      } \
    sqrt_ymin = MakeSmall(m); \
    ymin_align = MakeSmall(m*m); \
  }

#define FLOORSQRT \
    { \
      long n = GetSmall(ymax); \
      if (n>=2) \
	{			/* n = floor(sqrt(n)) */ \
	  long n1 = n>>1, n2 = n; \
 \
	  while (n1<n) \
	    n = n1, \
	    n1 = (n1 + n2/n1)>>1; \
	} \
      sqrt_ymax = MakeSmall(n); \
    }

#define XMutR (r+0)
#define XR (r+1)
#define YMutR (r+2)
#define YR (r+3)
#define ZMutR (r+4)
#define ZR (r+5)

/*
  '$fd_square'(+X, +XMut, +Y, +YMut, -Actions).
*/
void SPCDECL
prolog_fd_square MAGIC (HIDDEN_PROTO
			SP_term_ref State,
			SP_term_ref NewState,
			SP_term_ref Actions)
{
  WAMENV;
  TAGGED x, xmut, y, ymut;
  TAGGED newmin=0, newmax=0;
  TAGGED sqrt_ymin, sqrt_ymax, ymin_align;
  SP_globref r = SP_alloc_globrefs(4);
  unsigned int xflag=0, yflag=0; /* 1=consistent */
  struct dvar localvar[2];
  Dvar dvx = &localvar[0];
  Dvar dvy = &localvar[1];
  int ent = -1;			/* disentailed unless otherwise */

  w->numstack_end = NULL;
  fd.fd_overflow = FALSE;
  RefTerm(NewState) = RefTerm(State);
  dvar_export_start(3);
  DerefNonvar(X(0));
  DerefArg(x,X(0),1);
  DerefArg(xmut,X(0),2);
  DerefArg(y,X(0),3);
  DerefArg(ymut,X(0),4);
  RefGlob(XR) = x;
  RefGlob(XMutR) = xmut;
  RefGlob(YR) = y;
  RefGlob(YMutR) = ymut;
  dvar_init(dvx, XMutR, XR);
  dvar_init(dvy, YMutR, YR);
  sqrt_ymin = ymin;
  sqrt_ymax = ymax;
    
  CEILSQRT;
  if (TagIsSmall(ymax))
    FLOORSQRT;

  while (!(xflag&yflag&1))
    {
      if (!(yflag&1))
	{			/* y /\= x^2 */
	  yflag++;
	  switch ((!TagIsSmall(xmin) ? 0x0 : Tltz(xmin) ? 0x4 : 0x8)+
		  (!TagIsSmall(xmax) ? 0x0 : Tltz(xmax) ? 0x1 : 0x2))
	    {
	    case 0x0:		/* inf..sup */
	    case 0x2:		/* inf..+b */
	    case 0x4:		/*  -a..sup */
	      newmin = ymin;
	      newmax = ymax;
	      break;
	    case 0x1:		/* inf..-b => b^2..sup */
	      newmin = safe_mul_val(xmax,xmax);
	      newmax = ymax;
	      break;
	    case 0x5:		/*  -a..-b => b^2..a^2 */
	      newmin = safe_mul_val(xmax,xmax);
	      newmax = safe_mul_val(xmin,xmin);
	      break;
	    case 0x6:		/*  -a..b => 0..max(a^2,b^2) */
	      if (Tgt(Tminus(xmin),xmax))
		newmax = safe_mul_val(xmin,xmin);
	      else
		newmax = safe_mul_val(xmax,xmax);
	      break;
	    case 0x8:		/* a..sup => a^2..sup */
	      newmin = safe_mul_val(xmin,xmin);
	      newmax = ymax;
	      break;
	    case 0xa:		/*  a..b => a^2..b^2 */
	      newmin = safe_mul_val(xmin,xmin);
	      newmax = safe_mul_val(xmax,xmax);
	      break;
	    }
	  if (EmptyInterval(ymin_align,newmin))
	    newmin = ymin_align;
	  if (EmptyInterval(newmax,ymax))
	    newmax = ymax;
	  while (newmin!=ymin || newmax!=ymax)
	    {
	      xflag &= ~1;
	      if (dvar_fix_interval_t(dvy, newmin, newmax)<0)
		goto fail_or_oflo;
	      if (newmin!=ymin) {
		CEILSQRT;
		newmin = ymin_align;
	      }
	      if (newmax!=ymax) {
		FLOORSQRT;
		newmax = ymax;
	      }
	    }
	}
      else if (!(xflag&1))
	{			/* x /\= sqrt(y) */
	  xflag++;
	  if (Tlez(xmax)) /* x must be negative */ 
	    newmin = NegateUB(sqrt_ymax),
	    newmax = NegateLB(sqrt_ymin);
	  else if (xmin!=Inf && Tgez(xmin)) /* x must be positive */
	    newmin = sqrt_ymin, newmax = sqrt_ymax;
	  else
	    newmin = NegateUB(sqrt_ymax), newmax = sqrt_ymax;
	  
	  /* dom(y) may not be consistent after pruning dom(x) */

	  if (EmptyInterval(xmin,newmin))
	    newmin = xmin;
	  else if (newmin!=xmin)
	    yflag &= ~1;
	  if (EmptyInterval(newmax,xmax))
	    newmax = xmax;
	  else if (newmax!=xmax)
	    yflag &= ~1;
	  if (newmin!=xmin || newmax!=xmax) {
	    if (dvar_fix_interval_t(dvx, newmin, newmax)<0)
	      goto fail_or_oflo;
	  }
	}
    }

  dvar_export(dvx, 3);
  dvar_export(dvy, 3);
  ent = (xmin==xmax);
 ret:
  dvar_export_done(Actions, ent, 3);
  if (!fd_check_overflow())
    SP_fail();
  SP_free_globrefs(r,4);
  return;
  
fail_or_oflo:
  if (newmin==Sup || newmax==Inf) /* can't represent bounds */
    fd.fd_overflow = TRUE;
  goto ret;
}



/* -1=overflow, 0=fail, 2=no pruning, 1=pruning */
static int fd_product_inverse_dispatch MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
     Dvar dvx,
     Dvar dvy,
     Dvar dvz)
{
  TAGGED newmin, newmax;
  int rc = 2;
  int x_is_mon=0;
  long xl=0, xu=0, xsize=0;

  
     switch ((!TagIsSmall(xmin) ? 0x00 :
	      Tltz(xmin) ? 0x40 : Teqz(xmin) ? 0x80 : 0xc0)+
	     (!TagIsSmall(xmax) ? 0x30 :
	      Tltz(xmax) ? 0x00 : Teqz(xmax) ? 0x10 : 0x20)+
	     (!TagIsSmall(ymin) ? 0x00 :
	      Tltz(ymin) ? 0x04 : Teqz(ymin) ? 0x08 : 0x0c)+
	     (!TagIsSmall(ymax) ? 0x03 :
	      Tltz(ymax) ? 0x00 : Teqz(ymax) ? 0x01 : 0x02))
       {			/* automatically generated table! */
       case 0x90: /* 0-0 / inf- -(d) => 0-0 */
       case 0x94: /* 0-0 / -(c)- -(d) => 0-0 */
       case 0x9e: /* 0-0 / +(c)- +(d) => 0-0 */
       case 0x9f: /* 0-0 / +(c)-sup => 0-0 */
	   newmin = TaggedZero;
	   newmax = TaggedZero;
	   break;
       case 0xeb: /* +(a)- +(b) / 0-sup => 1-b */
	   x_is_mon = 1;
	   newmin = TaggedOne;
	   newmax = xmax;
	   break;
       case 0x0: /* inf- -(b) / inf- -(d) => 1-sup */
       case 0x1: /* inf- -(b) / inf-0 => 1-sup */
       case 0xfb: /* +(a)-sup / 0-sup => 1-sup */
       case 0xff: /* +(a)-sup / +(c)-sup => 1-sup */
	   newmin = TaggedOne;
	   newmax = zmax;
	   break;
       case 0x10: /* inf-0 / inf- -(d) => 0-sup */
       case 0x14: /* inf-0 / -(c)- -(d) => 0-sup */
       case 0xbe: /* 0-sup / +(c)- +(d) => 0-sup */
       case 0xbf: /* 0-sup / +(c)-sup => 0-sup */
	   newmin = TaggedZero;
	   newmax = zmax;
	   break;
       case 0x41: /* -(a)- -(b) / inf-0 => 1- -(a) */
	   x_is_mon = 1;
	   newmin = TaggedOne;
	   newmax = Tminus(xmin);
	   break;
       case 0x40: /* -(a)- -(b) / inf- -(d) => 1-a/d */
	   x_is_mon = 1;
	   newmin = TaggedOne;
	   newmax = MakeSmall(-CEILDIV(GetSmall(xmin),-GetSmall(ymax)));
	   break;
       case 0x50: /* -(a)-0 / inf- -(d) => 0-a/d */
       case 0x54: /* -(a)-0 / -(c)- -(d) => 0-a/d */
	   newmin = TaggedZero;
	   newmax = MakeSmall(-CEILDIV(GetSmall(xmin),-GetSmall(ymax)));
	   break;
       case 0xae: /* 0- +(b) / +(c)- +(d) => 0-b/c */
       case 0xaf: /* 0- +(b) / +(c)-sup => 0-b/c */
	   newmin = TaggedZero;
	   newmax = MakeSmall(FLOORDIV(GetSmall(xmax),GetSmall(ymin)));
	   break;
       case 0xef: /* +(a)- +(b) / +(c)-sup => 1-b/c */
	   x_is_mon = 1;
	   newmin = TaggedOne;
	   newmax = MakeSmall(FLOORDIV(GetSmall(xmax),GetSmall(ymin)));
	   break;
       case 0x9: /* inf- -(b) / 0-0 => 1-0 */
       case 0x49: /* -(a)- -(b) / 0-0 => 1-0 */
       case 0xe9: /* +(a)- +(b) / 0-0 => 1-0 */
       case 0xf9: /* +(a)-sup / 0-0 => 1-0 */
	   return 0;
       case 0x4b: /* -(a)- -(b) / 0-sup => a- -1 */
	   x_is_mon = 1;
	   newmin = xmin;
	   newmax = TaggedMinusOne;
	   break;
       case 0x42: /* -(a)- -(b) / inf- +(d) => a- -(a) */
       case 0x43: /* -(a)- -(b) / inf-sup => a- -(a) */
       case 0x46: /* -(a)- -(b) / -(c)- +(d) => a- -(a) */
       case 0x47: /* -(a)- -(b) / -(c)-sup => a- -(a) */
	   x_is_mon = 1;
	   newmin = xmin;
	   newmax = Tminus(xmin);
	   break;
       case 0x4a: /* -(a)- -(b) / 0- +(d) => a-b/d */
	   x_is_mon = 1;
	   newmin = xmin;
	   newmax = MakeSmall(FLOORDIV(GetSmall(xmax),GetSmall(ymax)));
	   break;
       case 0xf: /* inf- -(b) / +(c)-sup => inf- -1 */
       case 0xf0: /* +(a)-sup / inf- -(d) => inf- -1 */
	   newmin = zmin;
	   newmax = TaggedMinusOne;
	   break;
       case 0xb: /* inf- -(b) / 0-sup => inf- -1 */
       case 0xf1: /* +(a)-sup / inf-0 => inf- -1 */
	   newmin = zmin;
	   newmax = TaggedMinusOne;
	   break;
       case 0x1e: /* inf-0 / +(c)- +(d) => inf-0 */
       case 0x1f: /* inf-0 / +(c)-sup => inf-0 */
       case 0xb0: /* 0-sup / inf- -(d) => inf-0 */
       case 0xb4: /* 0-sup / -(c)- -(d) => inf-0 */
	   newmin = zmin;
	   newmax = TaggedZero;
	   break;
       case 0xf4: /* +(a)-sup / -(c)- -(d) => inf-a/c */
       case 0xf5: /* +(a)-sup / -(c)-0 => inf-a/c */
	   newmin = zmin;
	   newmax = MakeSmall(-CEILDIV(GetSmall(xmin),-GetSmall(ymin)));
	   break;
       case 0x70: /* -(a)-sup / inf- -(d) => inf-a/d */
       case 0x74: /* -(a)-sup / -(c)- -(d) => inf-a/d */
	   newmin = zmin;
	   newmax = MakeSmall(-CEILDIV(GetSmall(xmin),-GetSmall(ymax)));
	   break;
       case 0x2e: /* inf- +(b) / +(c)- +(d) => inf-b/c */
       case 0x2f: /* inf- +(b) / +(c)-sup => inf-b/c */
	   newmin = zmin;
	   newmax = MakeSmall(FLOORDIV(GetSmall(xmax),GetSmall(ymin)));
	   break;
       case 0xa: /* inf- -(b) / 0- +(d) => inf-b/d */
       case 0xe: /* inf- -(b) / +(c)- +(d) => inf-b/d */
	   newmin = zmin;
	   newmax = MakeSmall(FLOORDIV(GetSmall(xmax),GetSmall(ymax)));
	   break;
       case 0xe1: /* +(a)- +(b) / inf-0 => -(b)- -1 */
	   x_is_mon = 1;
	   newmin = Tminus(xmax);
	   newmax = TaggedMinusOne;
	   break;
       case 0xe2: /* +(a)- +(b) / inf- +(d) => -(b)-b */
       case 0xe3: /* +(a)- +(b) / inf-sup => -(b)-b */
       case 0xe6: /* +(a)- +(b) / -(c)- +(d) => -(b)-b */
       case 0xe7: /* +(a)- +(b) / -(c)-sup => -(b)-b */
	   x_is_mon = 1;
	   newmin = Tminus(xmax);
	   newmax = xmax;
	   break;
       case 0xe5: /* +(a)- +(b) / -(c)-0 => -(b)-a/c */
	   x_is_mon = 1;
	   newmin = Tminus(xmax);
	   newmax = MakeSmall(-CEILDIV(GetSmall(xmin),-GetSmall(ymin)));
	   break;
       case 0x4f: /* -(a)- -(b) / +(c)-sup => a/c- -1 */
	   x_is_mon = 1;
	   newmin = MakeSmall(CEILDIV(GetSmall(xmin),GetSmall(ymin)));
	   newmax = TaggedMinusOne;
	   break;
       case 0x5e: /* -(a)-0 / +(c)- +(d) => a/c-0 */
       case 0x5f: /* -(a)-0 / +(c)-sup => a/c-0 */
	   newmin = MakeSmall(CEILDIV(GetSmall(xmin),GetSmall(ymin)));
	   newmax = TaggedZero;
	   break;
       case 0x7e: /* -(a)-sup / +(c)- +(d) => a/c-sup */
       case 0x7f: /* -(a)-sup / +(c)-sup => a/c-sup */
	   newmin = MakeSmall(CEILDIV(GetSmall(xmin),GetSmall(ymin)));
	   newmax = zmax;
	   break;
       case 0x6e: /* -(a)- +(b) / +(c)- +(d) => a/c-b/c */
       case 0x6f: /* -(a)- +(b) / +(c)-sup => a/c-b/c */
	   newmin = MakeSmall(CEILDIV(GetSmall(xmin),GetSmall(ymin)));
	   newmax = MakeSmall(FLOORDIV(GetSmall(xmax),GetSmall(ymin)));
	   break;
       case 0x4e: /* -(a)- -(b) / +(c)- +(d) => a/c-b/d */
	   x_is_mon = 1;
	   newmin = MakeSmall(CEILDIV(GetSmall(xmin),GetSmall(ymin)));
	   newmax = MakeSmall(FLOORDIV(GetSmall(xmax),GetSmall(ymax)));
	   break;
       case 0xea: /* +(a)- +(b) / 0- +(d) => a/d-b */
	   x_is_mon = 1;
	   newmin = MakeSmall(CEILDIV(GetSmall(xmin),GetSmall(ymax)));
	   newmax = xmax;
	   break;
       case 0xfa: /* +(a)-sup / 0- +(d) => a/d-sup */
       case 0xfe: /* +(a)-sup / +(c)- +(d) => a/d-sup */
	   newmin = MakeSmall(CEILDIV(GetSmall(xmin),GetSmall(ymax)));
	   newmax = zmax;
	   break;
       case 0xee: /* +(a)- +(b) / +(c)- +(d) => a/d-b/c */
	   x_is_mon = 1;
	   newmin = MakeSmall(CEILDIV(GetSmall(xmin),GetSmall(ymax)));
	   newmax = MakeSmall(FLOORDIV(GetSmall(xmax),GetSmall(ymin)));
	   break;
       case 0x4: /* inf- -(b) / -(c)- -(d) => b/c-sup */
       case 0x5: /* inf- -(b) / -(c)-0 => b/c-sup */
	   newmin = MakeSmall(-FLOORDIV(GetSmall(xmax),-GetSmall(ymin)));
	   newmax = zmax;
	   break;
       case 0x45: /* -(a)- -(b) / -(c)-0 => b/c- -(a) */
	   x_is_mon = 1;
	   newmin = MakeSmall(-FLOORDIV(GetSmall(xmax),-GetSmall(ymin)));
	   newmax = Tminus(xmin);
	   break;
       case 0x44: /* -(a)- -(b) / -(c)- -(d) => b/c-a/d */
	   x_is_mon = 1;
	   newmin = MakeSmall(-FLOORDIV(GetSmall(xmax),-GetSmall(ymin)));
	   newmax = MakeSmall(-CEILDIV(GetSmall(xmin),-GetSmall(ymax)));
	   break;
       case 0xa0: /* 0- +(b) / inf- -(d) => b/d-0 */
       case 0xa4: /* 0- +(b) / -(c)- -(d) => b/d-0 */
	   newmin = MakeSmall(-FLOORDIV(GetSmall(xmax),-GetSmall(ymax)));
	   newmax = TaggedZero;
	   break;
       case 0xe0: /* +(a)- +(b) / inf- -(d) => b/d- -1 */
	   x_is_mon = 1;
	   newmin = MakeSmall(-FLOORDIV(GetSmall(xmax),-GetSmall(ymax)));
	   newmax = TaggedMinusOne;
	   break;
       case 0x20: /* inf- +(b) / inf- -(d) => b/d-sup */
       case 0x24: /* inf- +(b) / -(c)- -(d) => b/d-sup */
	   newmin = MakeSmall(-FLOORDIV(GetSmall(xmax),-GetSmall(ymax)));
	   newmax = zmax;
	   break;
       case 0xe4: /* +(a)- +(b) / -(c)- -(d) => b/d-a/c */
	   x_is_mon = 1;
	   newmin = MakeSmall(-FLOORDIV(GetSmall(xmax),-GetSmall(ymax)));
	   newmax = MakeSmall(-CEILDIV(GetSmall(xmin),-GetSmall(ymin)));
	   break;
       case 0x60: /* -(a)- +(b) / inf- -(d) => b/d-a/d */
       case 0x64: /* -(a)- +(b) / -(c)- -(d) => b/d-a/d */
	   newmin = MakeSmall(-FLOORDIV(GetSmall(xmax),-GetSmall(ymax)));
	   newmax = MakeSmall(-CEILDIV(GetSmall(xmin),-GetSmall(ymax)));
	   break;
       default:
	 return 2;
       }
    if (EmptyInterval(zmin,newmin))
      newmin = zmin;
    if (EmptyInterval(newmax,zmax))
      newmax = zmax;
    if (x_is_mon)
      {
	long m = GetSmall(newmin);
	long n = GetSmall(newmax);
	
	xl = GetSmall(xmin)-1;
	xu = GetSmall(xmax);
	xsize = xu-xl;
	if (xl<0)
	  xl++, xu++;

	while (m<=n && xsize<(m<0 ? -m : m) && xl/m == xu/m)
	  m++;
	while (m<=n && xsize<(n<0 ? -n : n) && xl/n == xu/n)
	  n--;
	
	newmin = MakeSmall(m);
	newmax = MakeSmall(n);
      }

    while (newmin!=zmin || newmax!=zmax)
      {
	rc = 1;
	if (dvar_fix_interval_t(dvz, newmin, newmax)<0)
	  return (newmin==Sup || newmax==Inf) ? -1 : 0;
	newmin = zmin; newmax = zmax;
	if (x_is_mon)
	  {
	    long m = GetSmall(newmin);
	    long n = GetSmall(newmax);
	    
	    while (m<=n && xsize<(m<0 ? -m : m) && xl/m == xu/m)
	      m++;
	    while (m<=n && xsize<(n<0 ? -n : n) && xl/n == xu/n)
	      n--;

	    newmin = MakeSmall(m);
	    newmax = MakeSmall(n);
	  }
      }
    return rc;
}




/*
  '$fd_product'(+X, +XMut, +Y, +YMut, +Z, +ZMut, -Actions).
*/
void SPCDECL
prolog_fd_product MAGIC (HIDDEN_PROTO
			 SP_term_ref State,
			 SP_term_ref NewState,
			 SP_term_ref Actions)
{
  WAMENV;
  TAGGED x, xmut, y, ymut, z, zmut;
  TAGGED newmin, newmax;
  SP_globref r = SP_alloc_globrefs(6);
  unsigned int xflag=0, yflag=0, zflag=0; /* 1=consistent */
  struct dvar localvar[3];
  Dvar dvx = &localvar[0];
  Dvar dvy = &localvar[1];
  Dvar dvz = &localvar[2];
  int ent = -1;			/* disentailed unless otherwise */

  w->numstack_end = NULL;
  fd.fd_overflow = FALSE;
  RefTerm(NewState) = RefTerm(State);
  dvar_export_start(3);
  DerefNonvar(X(0));
  DerefArg(x,X(0),1);
  DerefArg(xmut,X(0),2);
  DerefArg(y,X(0),3);
  DerefArg(ymut,X(0),4);
  DerefArg(z,X(0),5);
  DerefArg(zmut,X(0),6);
  RefGlob(XR) = x;
  RefGlob(XMutR) = xmut;
  RefGlob(YR) = y;
  RefGlob(YMutR) = ymut;
  RefGlob(ZR) = z;
  RefGlob(ZMutR) = zmut;
  dvar_init(dvx, XMutR, XR);
  dvar_init(dvy, YMutR, YR);
  dvar_init(dvz, ZMutR, ZR);

  while (!(xflag&yflag&zflag&1))
    {
      if (!(zflag&1))
	{			/* z /\= x*y */
	  zflag++;
	  switch ((!TagIsSmall(xmin) ? 0x00 :
		   Tltz(xmin) ? 0x40 : Teqz(xmin) ? 0x80 : 0xc0)+
		  (!TagIsSmall(xmax) ? 0x30 :
		   Tltz(xmax) ? 0x00 : Teqz(xmax) ? 0x10 : 0x20)+
		  (!TagIsSmall(ymin) ? 0x00 :
		   Tltz(ymin) ? 0x04 : Teqz(ymin) ? 0x08 : 0x0c)+
		  (!TagIsSmall(ymax) ? 0x03 :
		   Tltz(ymax) ? 0x00 : Teqz(ymax) ? 0x01 : 0x02))

	    {			/* automatically generated table! */
	    case 0x9: /* inf- -(b) * 0-0 => 0-0 */
	    case 0x19: /* inf-0 * 0-0 => 0-0 */
	    case 0x29: /* inf- +(b) * 0-0 => 0-0 */
	    case 0x39: /* inf-sup * 0-0 => 0-0 */
	    case 0x49: /* -(a)- -(b) * 0-0 => 0-0 */
	    case 0x59: /* -(a)-0 * 0-0 => 0-0 */
	    case 0x69: /* -(a)- +(b) * 0-0 => 0-0 */
	    case 0x79: /* -(a)-sup * 0-0 => 0-0 */
	    case 0x90: /* 0-0 * inf- -(d) => 0-0 */
	    case 0x91: /* 0-0 * inf-0 => 0-0 */
	    case 0x92: /* 0-0 * inf- +(d) => 0-0 */
	    case 0x93: /* 0-0 * inf-sup => 0-0 */
	    case 0x94: /* 0-0 * -(c)- -(d) => 0-0 */
	    case 0x95: /* 0-0 * -(c)-0 => 0-0 */
	    case 0x96: /* 0-0 * -(c)- +(d) => 0-0 */
	    case 0x97: /* 0-0 * -(c)-sup => 0-0 */
	    case 0x99: /* 0-0 * 0-0 => 0-0 */
	    case 0x9a: /* 0-0 * 0- +(d) => 0-0 */
	    case 0x9b: /* 0-0 * 0-sup => 0-0 */
	    case 0x9e: /* 0-0 * +(c)- +(d) => 0-0 */
	    case 0x9f: /* 0-0 * +(c)-sup => 0-0 */
	    case 0xa9: /* 0- +(b) * 0-0 => 0-0 */
	    case 0xb9: /* 0-sup * 0-0 => 0-0 */
	    case 0xe9: /* +(a)- +(b) * 0-0 => 0-0 */
	    case 0xf9: /* +(a)-sup * 0-0 => 0-0 */
	      newmin = TaggedZero;
	      newmax = TaggedZero;
	      break;
	    case 0x1: /* inf- -(b) * inf-0 => 0-sup */
	    case 0x5: /* inf- -(b) * -(c)-0 => 0-sup */
	    case 0x10: /* inf-0 * inf- -(d) => 0-sup */
	    case 0x11: /* inf-0 * inf-0 => 0-sup */
	    case 0x14: /* inf-0 * -(c)- -(d) => 0-sup */
	    case 0x15: /* inf-0 * -(c)-0 => 0-sup */
	    case 0x41: /* -(a)- -(b) * inf-0 => 0-sup */
	    case 0x50: /* -(a)-0 * inf- -(d) => 0-sup */
	    case 0x51: /* -(a)-0 * inf-0 => 0-sup */
	    case 0xab: /* 0- +(b) * 0-sup => 0-sup */
	    case 0xaf: /* 0- +(b) * +(c)-sup => 0-sup */
	    case 0xba: /* 0-sup * 0- +(d) => 0-sup */
	    case 0xbb: /* 0-sup * 0-sup => 0-sup */
	    case 0xbe: /* 0-sup * +(c)- +(d) => 0-sup */
	    case 0xbf: /* 0-sup * +(c)-sup => 0-sup */
	    case 0xeb: /* +(a)- +(b) * 0-sup => 0-sup */
	    case 0xfa: /* +(a)-sup * 0- +(d) => 0-sup */
	    case 0xfb: /* +(a)-sup * 0-sup => 0-sup */
	      newmin = TaggedZero;
	      newmax = zmax;
	      break;
	    case 0xaa: /* 0- +(b) * 0- +(d) => 0- +(b)* +(d) */
	    case 0xae: /* 0- +(b) * +(c)- +(d) => 0- +(b)* +(d) */
	    case 0xea: /* +(a)- +(b) * 0- +(d) => 0- +(b)* +(d) */
	      newmin = TaggedZero;
	      newmax = safe_mul_val(xmax,ymax);
	      break;
	    case 0x45: /* -(a)- -(b) * -(c)-0 => 0- -(a)* -(c) */
	    case 0x54: /* -(a)-0 * -(c)- -(d) => 0- -(a)* -(c) */
	    case 0x55: /* -(a)-0 * -(c)-0 => 0- -(a)* -(c) */
	      newmin = TaggedZero;
	      newmax = safe_mul_val(xmin,ymin);
	      break;
	    case 0xa: /* inf- -(b) * 0- +(d) => inf-0 */
	    case 0xb: /* inf- -(b) * 0-sup => inf-0 */
	    case 0x1a: /* inf-0 * 0- +(d) => inf-0 */
	    case 0x1b: /* inf-0 * 0-sup => inf-0 */
	    case 0x1e: /* inf-0 * +(c)- +(d) => inf-0 */
	    case 0x1f: /* inf-0 * +(c)-sup => inf-0 */
	    case 0x4b: /* -(a)- -(b) * 0-sup => inf-0 */
	    case 0x5b: /* -(a)-0 * 0-sup => inf-0 */
	    case 0x5f: /* -(a)-0 * +(c)-sup => inf-0 */
	    case 0xa0: /* 0- +(b) * inf- -(d) => inf-0 */
	    case 0xa1: /* 0- +(b) * inf-0 => inf-0 */
	    case 0xb0: /* 0-sup * inf- -(d) => inf-0 */
	    case 0xb1: /* 0-sup * inf-0 => inf-0 */
	    case 0xb4: /* 0-sup * -(c)- -(d) => inf-0 */
	    case 0xb5: /* 0-sup * -(c)-0 => inf-0 */
	    case 0xe1: /* +(a)- +(b) * inf-0 => inf-0 */
	    case 0xf1: /* +(a)-sup * inf-0 => inf-0 */
	    case 0xf5: /* +(a)-sup * -(c)-0 => inf-0 */
	      newmin = zmin;
	      newmax = TaggedZero;
	      break;
	    case 0xe0: /* +(a)- +(b) * inf- -(d) => inf- +(a)* -(d) */
	    case 0xf0: /* +(a)-sup * inf- -(d) => inf- +(a)* -(d) */
	    case 0xf4: /* +(a)-sup * -(c)- -(d) => inf- +(a)* -(d) */
	      newmin = zmin;
	      newmax = safe_mul_val(xmin,ymax);
	      break;
	    case 0x2a: /* inf- +(b) * 0- +(d) => inf- +(b)* +(d) */
	    case 0x2e: /* inf- +(b) * +(c)- +(d) => inf- +(b)* +(d) */
	    case 0xa2: /* 0- +(b) * inf- +(d) => inf- +(b)* +(d) */
	    case 0xe2: /* +(a)- +(b) * inf- +(d) => inf- +(b)* +(d) */
	      newmin = zmin;
	      newmax = safe_mul_val(xmax,ymax);
	      break;
	    case 0x47: /* -(a)- -(b) * -(c)-sup => inf- -(a)* -(c) */
	    case 0x57: /* -(a)-0 * -(c)-sup => inf- -(a)* -(c) */
	    case 0x74: /* -(a)-sup * -(c)- -(d) => inf- -(a)* -(c) */
	    case 0x75: /* -(a)-sup * -(c)-0 => inf- -(a)* -(c) */
	      newmin = zmin;
	      newmax = safe_mul_val(xmin,ymin);
	      break;
	    case 0xe: /* inf- -(b) * +(c)- +(d) => inf- -(b)* +(c) */
	    case 0xf: /* inf- -(b) * +(c)-sup => inf- -(b)* +(c) */
	    case 0x4f: /* -(a)- -(b) * +(c)-sup => inf- -(b)* +(c) */
	      newmin = zmin;
	      newmax = safe_mul_val(xmax,ymin);
	      break;
	    case 0xef: /* +(a)- +(b) * +(c)-sup => +(a)* +(c)-sup */
	    case 0xfe: /* +(a)-sup * +(c)- +(d) => +(a)* +(c)-sup */
	    case 0xff: /* +(a)-sup * +(c)-sup => +(a)* +(c)-sup */
	      newmin = safe_mul_val(xmin,ymin);
	      newmax = zmax;
	      break;
	    case 0xee: /* +(a)- +(b) * +(c)- +(d) => +(a)* +(c)- +(b)* +(d) */
	      newmin = safe_mul_val(xmin,ymin);
	      newmax = safe_mul_val(xmax,ymax);
	      break;
	    case 0xa4: /* 0- +(b) * -(c)- -(d) => +(b)* -(c)-0 */
	    case 0xa5: /* 0- +(b) * -(c)-0 => +(b)* -(c)-0 */
	    case 0xe5: /* +(a)- +(b) * -(c)-0 => +(b)* -(c)-0 */
	      newmin = safe_mul_val(xmax,ymin);
	      newmax = TaggedZero;
	      break;
	    case 0x24: /* inf- +(b) * -(c)- -(d) => +(b)* -(c)-sup */
	    case 0x25: /* inf- +(b) * -(c)-0 => +(b)* -(c)-sup */
	    case 0xa7: /* 0- +(b) * -(c)-sup => +(b)* -(c)-sup */
	    case 0xe7: /* +(a)- +(b) * -(c)-sup => +(b)* -(c)-sup */
	      newmin = safe_mul_val(xmax,ymin);
	      newmax = zmax;
	      break;
	    case 0xe4: /* +(a)- +(b) * -(c)- -(d) => +(b)* -(c)- +(a)* -(d) */
	      newmin = safe_mul_val(xmax,ymin);
	      newmax = safe_mul_val(xmin,ymax);
	      break;
	    case 0xa6: /* 0- +(b) * -(c)- +(d) => +(b)* -(c)- +(b)* +(d) */
	    case 0xe6: /* +(a)- +(b) * -(c)- +(d) => +(b)* -(c)- +(b)* +(d) */
	      newmin = safe_mul_val(xmax,ymin);
	      newmax = safe_mul_val(xmax,ymax);
	      break;
	    case 0x64: /* -(a)- +(b) * -(c)- -(d) => +(b)* -(c)- -(a)* -(c) */
	    case 0x65: /* -(a)- +(b) * -(c)-0 => +(b)* -(c)- -(a)* -(c) */
	      newmin = safe_mul_val(xmax,ymin);
	      newmax = safe_mul_val(xmin,ymin);
	      break;
	    case 0x4a: /* -(a)- -(b) * 0- +(d) => -(a)* +(d)-0 */
	    case 0x5a: /* -(a)-0 * 0- +(d) => -(a)* +(d)-0 */
	    case 0x5e: /* -(a)-0 * +(c)- +(d) => -(a)* +(d)-0 */
	      newmin = safe_mul_val(xmin,ymax);
	      newmax = TaggedZero;
	      break;
	    case 0x42: /* -(a)- -(b) * inf- +(d) => -(a)* +(d)-sup */
	    case 0x52: /* -(a)-0 * inf- +(d) => -(a)* +(d)-sup */
	    case 0x7a: /* -(a)-sup * 0- +(d) => -(a)* +(d)-sup */
	    case 0x7e: /* -(a)-sup * +(c)- +(d) => -(a)* +(d)-sup */
	      newmin = safe_mul_val(xmin,ymax);
	      newmax = zmax;
	      break;
	    case 0x6a: /* -(a)- +(b) * 0- +(d) => -(a)* +(d)- +(b)* +(d) */
	    case 0x6e: /* -(a)- +(b) * +(c)- +(d) => -(a)* +(d)- +(b)* +(d) */
	      newmin = safe_mul_val(xmin,ymax);
	      newmax = safe_mul_val(xmax,ymax);
	      break;
	    case 0x46: /* -(a)- -(b) * -(c)- +(d) => -(a)* +(d)- -(a)* -(c) */
	    case 0x56: /* -(a)-0 * -(c)- +(d) => -(a)* +(d)- -(a)* -(c) */
	      newmin = safe_mul_val(xmin,ymax);
	      newmax = safe_mul_val(xmin,ymin);
	      break;
	    case 0x4e: /* -(a)- -(b) * +(c)- +(d) => -(a)* +(d)- -(b)* +(c) */
	      newmin = safe_mul_val(xmin,ymax);
	      newmax = safe_mul_val(xmax,ymin);
	      break;
	    case 0x0: /* inf- -(b) * inf- -(d) => -(b)* -(d)-sup */
	    case 0x4: /* inf- -(b) * -(c)- -(d) => -(b)* -(d)-sup */
	    case 0x40: /* -(a)- -(b) * inf- -(d) => -(b)* -(d)-sup */
	      newmin = safe_mul_val(xmax,ymax);
	      newmax = zmax;
	      break;
	    case 0x44: /* -(a)- -(b) * -(c)- -(d) => -(b)* -(d)- -(a)* -(c) */
	      newmin = safe_mul_val(xmax,ymax);
	      newmax = safe_mul_val(xmin,ymin);
	      break;
	    case 0x66: /* -(a)- +(b) * -(c)- +(d) => min(-(a)* +(d),+(b)* -(c))-max(-(a)* -(c),+(b)* +(d)) */
	      {
		TAGGED minmax = safe_mul_val(xmin,ymax);
		TAGGED maxmin = safe_mul_val(xmax,ymin);
		TAGGED minmin = safe_mul_val(xmin,ymin);
		TAGGED maxmax = safe_mul_val(xmax,ymax);

		newmin = EmptyInterval(minmax,maxmin) ? maxmin : minmax;
		newmax = EmptyInterval(minmin,maxmax) ? minmin : maxmax;
	      }
	      break;
	    default:
	      continue;
	    }
	  if (EmptyInterval(zmin,newmin))
	    newmin = zmin;
	  if (EmptyInterval(newmax,zmax))
	    newmax = zmax;
	  if (newmin!=zmin || newmax!=zmax)
	    {
	      if (dvar_fix_interval_t(dvz, newmin, newmax)<0)
		goto fail_or_oflo;
	      if (newmin!=zmin || newmax!=zmax)
		xflag &= ~1, yflag &= ~1;
	    }
	}
      else if (!(xflag&1))
	{			/* x /\= z/y */
	  xflag++;
	  switch (fd_product_inverse_dispatch(dvz, dvy, dvx)) {
	  case -1: goto oflo;
	  case 0: goto ret;
	  case 1: yflag &= ~1, zflag &= ~1;
	    /* case 2: continue; */
	  }
	}
      else if (!(yflag&1))
	{			/* y /\= z/x */
	  yflag++;
	  switch (fd_product_inverse_dispatch(dvz, dvx, dvy)) {
	  case -1: goto oflo;
	  case 0: goto ret;
	  case 1: xflag &= ~1, zflag &= ~1;
	    /* case 2: continue; */
	  }
	}
    }

  dvar_export(dvx, 3);
  dvar_export(dvy, 3);
  dvar_export(dvz, 3);
  ent = ((xmin==xmax)+(ymin==ymax)+(zmin==zmax) >= 2);
 ret:
  dvar_export_done(Actions, ent, 3);
  if (!fd_check_overflow())
    SP_fail();
  SP_free_globrefs(r,6);
  return;
  
fail_or_oflo:
  if (newmin==Sup || newmax==Inf) /* can't represent bounds */
    {
    oflo:
      fd.fd_overflow = TRUE;
    }
  goto ret;
}



/*
  '$fd_quotient'(+X, +XMut, +K, +Y, +YMut, -Actions).
*/
void SPCDECL
prolog_fd_quotient MAGIC (HIDDEN_PROTO
			  SP_term_ref State,
			  SP_term_ref NewState,
			  SP_term_ref Actions)
{
  WAMENV;
  TAGGED x, xmut, tk, y, ymut;
  TAGGED newmin, newmax;
  SP_globref r = SP_alloc_globrefs(4);
  unsigned int xflag=0, yflag=0; /* 1=consistent */
  long k;
  struct dvar localvar[2];
  Dvar dvx = &localvar[0];
  Dvar dvy = &localvar[1];
  int ent = -1;			/* disentailed unless otherwise */

  w->numstack_end = NULL;
  fd.fd_overflow = FALSE;
  RefTerm(NewState) = RefTerm(State);
  dvar_export_start(3);
  DerefNonvar(X(0));
  DerefArg(x,X(0),1);
  DerefArg(xmut,X(0),2);
  DerefArg(tk,X(0),3);
  DerefArg(y,X(0),4);
  DerefArg(ymut,X(0),5);
  RefGlob(XR) = x;
  RefGlob(XMutR) = xmut;
  RefGlob(YR) = y;
  RefGlob(YMutR) = ymut;
  dvar_init(dvx, XMutR, XR);
  dvar_init(dvy, YMutR, YR);
  DerefNonvar(tk);
  k = GetSmall(tk);

  while (!(xflag&yflag&1))
    {
      if (!(yflag&1))
	{			/* y /\= x/k */
	  yflag++;
	  switch ((!TagIsSmall(xmin) ? 0x0 :
		   Tltz(xmin) ? 0x4 : Teqz(xmin) ? 0x8 : 0xc)+
		  (!TagIsSmall(xmax) ? 0x3 :
		   Tltz(xmax) ? 0x0 : Teqz(xmax) ? 0x1 : 0x2)+
		  (k>0 ? 0x0 : 0x10))
	    {
	    case 0x1:		/* k>0 inf..0 */
	    case 0x1b:		/* k<0   0..sup */
	      newmin = ymin;
	      newmax = TaggedZero;
	      break;
	    case 0x0:		/* k>0 inf..-b */
	    case 0x2:		/* k>0 inf..+b */
	      newmin = ymin;
	      newmax = MakeSmall(GetSmall(xmax)/k);
	      break;
	    case 0x5:		/* k>0  -a..0 */
	      newmin = MakeSmall(GetSmall(xmin)/k);
	      newmax = TaggedZero;
	      break;
	    case 0x9:		/* k>0   0..0 */
	    case 0x19:		/* k<0   0..0 */
	      newmin = TaggedZero;
	      newmax = TaggedZero;
	      break;
	    case 0xa:		/* k>0   0..+b */
	      newmin = TaggedZero;
	      newmax = MakeSmall(GetSmall(xmax)/k);
	      break;
	    case 0xb:		/* k>0   0..sup */
	    case 0x11:		/* k<0 inf..0 */
	      newmin = TaggedZero;
	      newmax = ymax;
	      break;
	    case 0x4:		/* k>0  -a..-b */
	    case 0x6:		/* k>0  -a..+b */
	    case 0xe:		/* k>0  +a..+b */
	      newmin = MakeSmall(GetSmall(xmin)/k);
	      newmax = MakeSmall(GetSmall(xmax)/k);
	      break;
	    case 0x7:		/* k>0  -a..sup */
	    case 0xf:		/* k>0  +a..sup */
	      newmin = MakeSmall(GetSmall(xmin)/k);
	      newmax = ymax;
	      break;
	    case 0x10:		/* k<0 inf..-b */
	    case 0x12:		/* k<0 inf..+b */
	      newmin = MakeSmall(GetSmall(xmax)/k);
	      newmax = ymax;
	      break;
	    case 0x15:		/* k<0  -a..0 */
	      newmin = TaggedZero;
	      newmax = MakeSmall(GetSmall(xmin)/k);
	      break;
	    case 0x1a:		/* k<0   0..+b */
	      newmin = MakeSmall(GetSmall(xmax)/k);
	      newmax = TaggedZero;
	      break;
	    case 0x14:		/* k<0  -a..-b */
	    case 0x16:		/* k<0  -a..+b */
	    case 0x1e:		/* k<0  +a..+b */
	      newmin = MakeSmall(GetSmall(xmax)/k);
	      newmax = MakeSmall(GetSmall(xmin)/k);
	      break;
	    case 0x17:		/* k<0  -a..sup */
	    case 0x1f:		/* k<0  +a..sup */
	      newmin = ymin;
	      newmax = MakeSmall(GetSmall(xmin)/k);
	      break;
	    default:		/* inf..sup */
	      continue;
	    }
	  
	  /* dom(x) may not be consistent after pruning dom(y) */
	  
	  if (EmptyInterval(ymin,newmin))
	    newmin = ymin;
	  else if (newmin!=ymin)
	    xflag &= ~1;
	  if (EmptyInterval(newmax,ymax))
	    newmax = ymax;
	  else if (newmax!=ymax)
	    xflag &= ~1;
	  if (newmin!=ymin || newmax!=ymax)
	    {
	      if (dvar_fix_interval_t(dvy, newmin, newmax)<0)
		goto fail_or_oflo;
	    }
	}
      else if (!(xflag&1))
	{			/* x /\= y*k (+ some slack) */
	  xflag++;
	  switch ((!TagIsSmall(ymin) ? 0x0 :
		   Tltz(ymin) ? 0x4 : Teqz(ymin) ? 0x8 : 0xc)+
		  (!TagIsSmall(ymax) ? 0x3 :
		   Tltz(ymax) ? 0x0 : Teqz(ymax) ? 0x1 : 0x2)+
		  (k>0 ? 0x0 : 0x10))
	    {
	    case 0x1:		/* k>0 inf..0 => inf..k-1 */
	      newmin = xmin;
	      newmax = MakeSmall(k-1);
	      break;
	    case 0x1b:		/* k<0   0..sup => inf..-k-1*/
	      newmin = xmin;
	      newmax = MakeSmall(-k-1);
	      break;
	    case 0x0:		/* k>0 inf..-b => inf..-bk */
	      newmin = xmin;
	      newmax = safe_mul_val(ymax,tk);
	      break;
	    case 0x2:		/* k>0 inf..+b => inf..+bk+k-1 */
	      newmin = xmin;
	      newmax = safe_mul_val(ymax+IStep(1),tk);
	      if (TagIsSmall(newmax))
		newmax -= IStep(1);
	      break;
	    case 0x5:		/* k>0  -a..0 => -ak-k+1..k-1 */
	      newmin = safe_mul_val(ymin-IStep(1),tk);
	      if (TagIsSmall(newmin))
		newmin += IStep(1);
	      newmax = MakeSmall(k-1);
	      break;
	    case 0x9:		/* k>0   0..0 => -k+1..k-1 */
	      newmin = MakeSmall(1-k);
	      newmax = MakeSmall(k-1);
	      break;
	    case 0x19:		/* k<0   0..0 =>  k+1..-k-1*/
	      newmin = MakeSmall(k+1);
	      newmax = MakeSmall(-k-1);
	      break;
	    case 0xa:		/* k>0   0..+b => -k+1..+bk+k-1*/
	      newmin = MakeSmall(1-k);
	      newmax = safe_mul_val(ymax+IStep(1),tk);
	      if (TagIsSmall(newmax))
		newmax -= IStep(1);
	      break;
	    case 0xb:		/* k>0   0..sup => -k+1..sup */
	      newmin = MakeSmall(1-k);
	      newmax = xmax;
	      break;
	    case 0x11:		/* k<0 inf..0   =>  k+1..sup */
	      newmin = MakeSmall(k+1);
	      newmax = xmax;
	      break;
	    case 0x4:		/* k>0  -a..-b => -ak-k+1..-bk */
	      newmin = safe_mul_val(ymin-IStep(1),tk);
	      if (TagIsSmall(newmin))
		newmin += IStep(1);
	      newmax = safe_mul_val(ymax,tk);
	      break;
	    case 0x6:		/* k>0  -a..+b => -ak-k+1..+bk+k-1 */
	      newmin = safe_mul_val(ymin-IStep(1),tk);
	      if (TagIsSmall(newmin))
		newmin += IStep(1);
	      newmax = safe_mul_val(ymax+IStep(1),tk);
	      if (TagIsSmall(newmax))
		newmax -= IStep(1);
	      break;
	    case 0xe:		/* k>0  +a..+b => +ak..+bk+k-1 */
	      newmin = safe_mul_val(ymin,tk);
	      newmax = safe_mul_val(ymax+IStep(1),tk);
	      if (TagIsSmall(newmax))
		newmax -= IStep(1);
	      break;
	    case 0x7:		/* k>0  -a..sup => -ak-k+1..sup */
	      newmin = safe_mul_val(ymin-IStep(1),tk);
	      if (TagIsSmall(newmin))
		newmin += IStep(1);
	      newmax = xmax;
	      break;
	    case 0xf:		/* k>0  +a..sup => +ak..sup */
	      newmin = safe_mul_val(ymin,tk);
	      newmax = xmax;
	      break;
	    case 0x10:		/* k<0 inf..-b => -bk..sup */
	      newmin = safe_mul_val(ymax,tk);
	      newmax = xmax;
	      break;
	    case 0x12:		/* k<0 inf..+b => +bk+k+1..sup */
	      newmin = safe_mul_val(ymax+IStep(1),tk);
	      if (TagIsSmall(newmin))
		newmin += IStep(1);
	      newmax = xmax;
	      break;
	    case 0x15:		/* k<0  -a..0 => k+1..-ak-k-1 */
	      newmin = MakeSmall(k+1);
	      newmax = safe_mul_val(ymin-IStep(1),tk);
	      if (TagIsSmall(newmax))
		newmax -= IStep(1);
	      break;
	    case 0x1a:		/* k<0   0..+b => +bk+k+1..-k-1 */
	      newmin = safe_mul_val(ymax+IStep(1),tk);
	      if (TagIsSmall(newmin))
		newmin += IStep(1);
	      newmax = MakeSmall(-k-1);
	      break;
	    case 0x14:		/* k<0  -a..-b => -bk..-ak-k-1 */
	      newmin = safe_mul_val(ymax,tk);
	      newmax = safe_mul_val(ymin-IStep(1),tk);
	      if (TagIsSmall(newmax))
		newmax -= IStep(1);
	      break;
	    case 0x16:		/* k<0  -a..+b => +bk+k+1..-ak-k-1*/
	      newmin = safe_mul_val(ymax+IStep(1),tk);
	      if (TagIsSmall(newmin))
		newmin += IStep(1);
	      newmax = safe_mul_val(ymin-IStep(1),tk);
	      if (TagIsSmall(newmax))
		newmax -= IStep(1);
	      break;
	    case 0x1e:		/* k<0  +a..+b => +bk+k+1..+ak*/
	      newmin = safe_mul_val(ymax+IStep(1),tk);
	      if (TagIsSmall(newmin))
		newmin += IStep(1);
	      newmax = safe_mul_val(ymin,tk);
	      break;
	    case 0x17:		/* k<0  -a..sup => inf..-ak-k-1*/
	      newmin = xmin;
	      newmax = safe_mul_val(ymin-IStep(1),tk);
	      if (TagIsSmall(newmax))
		newmax -= IStep(1);
	      break;
	    case 0x1f:		/* k<0  +a..sup => inf..+ak*/
	      newmin = xmin;
	      newmax = safe_mul_val(ymin,tk);
	      break;
	    default:		/* inf..sup */
	      continue;
	    }
	  
	  /* dom(y) is consistent after pruning dom(x) */

	  if (EmptyInterval(xmin,newmin))
	    newmin = xmin;
	  if (EmptyInterval(newmax,xmax))
	    newmax = xmax;
	  if (newmin!=xmin || newmax!=xmax)
	    {
	      if (dvar_fix_interval_t(dvx, newmin, newmax)<0)
		goto fail_or_oflo;
	      if (newmin!=xmin || newmax!=xmax)
		yflag &= ~1;
	    }
	}
    }

  dvar_export(dvx, 3);
  dvar_export(dvy, 3);
  ent = (xmin==xmax);
 ret:
  dvar_export_done(Actions, ent, 3);
  if (!fd_check_overflow())
    SP_fail();
  SP_free_globrefs(r,4);
  return;
  
fail_or_oflo:
  if (newmin==Sup || newmax==Inf) /* can't represent bounds */
    fd.fd_overflow = TRUE;
  goto ret;
}



/*
  '$fd_modulo'(+X, +XMut, +K, +Y, +YMut, -Actions).
*/
void SPCDECL
prolog_fd_modulo MAGIC (HIDDEN_PROTO
			SP_term_ref State,
			SP_term_ref NewState,
			SP_term_ref Actions)
{
  WAMENV;
  TAGGED x, xmut, tk, y, ymut;
  TAGGED newmin, newmax;
  unsigned int xflag=0, yflag=0; /* 1=consistent */
  SP_globref r = SP_alloc_globrefs(4);
  long k, amk, bmk, kmin, kmax;
  struct dvar localvar[2];
  Dvar dvx = &localvar[0];
  Dvar dvy = &localvar[1];
  int ent = -1;			/* disentailed unless otherwise */

  w->numstack_end = NULL;
  fd.fd_overflow = FALSE;
  RefTerm(NewState) = RefTerm(State);
  dvar_export_start(3);
  DerefNonvar(X(0));
  DerefArg(x,X(0),1);
  DerefArg(xmut,X(0),2);
  DerefArg(tk,X(0),3);
  DerefArg(y,X(0),4);
  DerefArg(ymut,X(0),5);
  RefGlob(XR) = x;
  RefGlob(XMutR) = xmut;
  RefGlob(YR) = y;
  RefGlob(YMutR) = ymut;
  dvar_init(dvx, XMutR, XR);
  dvar_init(dvy, YMutR, YR);
  DerefNonvar(tk);
  k = GetSmall(tk);
  if (k<0) k = -k;
  kmin = MakeSmall(1-k);
  kmax = MakeSmall(k-1);

  while (!(xflag&yflag&1))
    {
      if (!(yflag&1))
	{			/* y /\= x mod k */
	  yflag++;
	  amk = MakeSmall(GetSmall(xmin) % k);
	  bmk = MakeSmall(GetSmall(xmax) % k);
	  switch ((!TagIsSmall(xmin) ? 0x0 :
		   Tltz(xmin) ? 0x4 : Teqz(xmin) ? 0x8 : 0xc)+
		  (!TagIsSmall(xmax) ? 0x3 :
		   Tltz(xmax) ? 0x0 : Teqz(xmax) ? 0x1 : 0x2))
	    {
	    case 0x0:		/* inf..-b */
	    case 0x1:		/* inf..0 */
	      newmin = kmin;
	      newmax = TaggedZero;
	      break;
	    case 0x2:		/* inf..+b */
	      newmin = kmin;
	      newmax = (xmax >= kmax ? kmax : bmk);
	      break;
	    case 0x4:		/*  -a..-b */
	      if (Tge(xmax-xmin+TaggedZero,kmax) || Tgt(amk,bmk))
		newmin = kmin, newmax = TaggedZero;
	      else
		newmin = amk, newmax = bmk;
	      break;
	    case 0x5:		/*  -a..0 */
	      newmin = (Tle(xmin,kmin) ? kmin : amk);
	      newmax = TaggedZero;
	      break;
	    case 0x6:		/*  -a..+b */
	      newmin = (Tle(xmin,kmin) ? kmin : amk);
	      newmax = (Tge(xmax,kmax) ? kmax : bmk);
	      break;
	    case 0x7:		/*  -a..sup */
	      newmin = (Tle(xmin,kmin) ? kmin : amk);
	      newmax = kmax;
	      break;
	    case 0x9:		/*   0..0 */
	      newmin = newmax = TaggedZero;
	      break;
	    case 0xa:		/*   0..+b */
	      newmin = TaggedZero;
	      newmax = (Tge(xmax,kmax) ? kmax : bmk);
	      break;
	    case 0xb:		/*   0..sup */
	      newmin = TaggedZero;
	      newmax = kmax;
	      break;
	    case 0xe:		/*  +a..+b */
	      if (Tge(xmax-xmin+TaggedZero,kmax) || Tgt(amk,bmk))
		newmin = TaggedZero, newmax = kmax;
	      else
		newmin = amk, newmax = bmk;
	      break;
	    case 0xf:		/*  +a..sup */
	      newmin = TaggedZero;
	      newmax = kmax;
	      break;
	    default:
	      continue;
	    }
	  
	  /* dom(x) may not be consistent after pruning dom(y) */
	  
	  if (EmptyInterval(ymin,newmin))
	    newmin = ymin;
	  else if (newmin!=ymin)
	    xflag &= ~1;
	  if (EmptyInterval(newmax,ymax))
	    newmax = ymax;
	  else if (newmax!=ymax)
	    xflag &= ~1;
	  if (newmin!=ymin || newmax!=ymax)
	    {
	      if (dvar_fix_interval_t(dvy, newmin, newmax)<0)
		goto fail_or_oflo;
	    }
	}
      else if (!(xflag&1))
	{			/* x /\= {x | x mod k = y} */
	  xflag++;
	  amk = MakeSmall(GetSmall(xmin) % k);
	  bmk = MakeSmall(GetSmall(xmax) % k);
	  if ((TagIsSmall(xmin) && (Tlt(amk,ymin) || Tgt(amk,ymax))) ||
	      (TagIsSmall(xmax) && (Tlt(bmk,ymin) || Tgt(bmk,ymax))))
	    {
	      newmin = Inf;
	      newmax = Sup;
	      while (newmin!=xmin || newmax!=xmax)
		{
		  yflag &= ~1;
		  if (dvar_fix_interval_t(dvx, newmin, newmax)<0)
		    goto fail_or_oflo;
		  amk = MakeSmall(GetSmall(xmin) % k);
		  bmk = MakeSmall(GetSmall(xmax) % k);
		  newmin = ( !TagIsSmall(xmin) ? xmin
			     : Tlt(amk,ymin) ? xmin-amk+ymin
			     : Tgt(amk,ymax) ? xmin-amk+ymin+IStep(k)
			     : xmin);
		  newmax = ( !TagIsSmall(xmax) ? xmax
			     : Tgt(bmk,ymax) ? xmax-bmk+ymax
			     : Tlt(bmk,ymin) ? xmax-bmk+ymax-IStep(k)
			     : xmax);
		}
	    }
	}
    }
  
  dvar_export(dvx, 3);
  dvar_export(dvy, 3);
  ent = (xmin==xmax || k==1);
 ret:
  dvar_export_done(Actions, ent, 3);
  if (!fd_check_overflow())
    SP_fail();
  SP_free_globrefs(r,4);
  return;
  
fail_or_oflo:
  if (newmin==Sup || newmax==Inf) /* can't represent bounds */
    fd.fd_overflow = TRUE;
  goto ret;
}

