/* Copyright(C) 1999, Swedish Institute of Computer Science */

#include "fd.h"
#include "dvars.h"

#if MULTI_SP_AWARE
#define request_tell(A1,A2,A3,A4,A5,A6) request_tell(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)
#define request_tell_interval(A1,A2,A3,A4,A5,A6,A7) request_tell_interval(HIDDEN_ARG, A1,A2,A3,A4,A5,A6,A7)
#define request_tell_value(A1,A2,A3,A4,A5,A6) request_tell_value(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)
#define request_rewrite_eq(A1,A2,A3,A4,A5) request_rewrite_eq(HIDDEN_ARG, A1,A2,A3,A4,A5)
#define request_rewrite_leqc(A1,A2,A3,A4,A5,A6) request_rewrite_leqc(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)

#if DynamicMallocBase           /* [PM] 3.10.1 */
#define dvar_adjust_bounds(A1,A2,A3,A4,A5) dvar_adjust_bounds(HIDDEN_ARG, A1,A2,A3,A4,A5)
#endif /* DynamicMallocBase */

#endif /* MULTI_SP_AWARE */

/* initialize dvar */

void
dvar_init MAGIC (HIDDEN_PROTO
		 Dvar dvar,
		 SP_globref attr_ref,
		 SP_globref var_ref)
{
  TAGGED tmp = RefGlob(attr_ref);

  dvar->attr_ref = attr_ref;
  dvar->var_ref = var_ref;
  DerefAttribute(tmp,tmp); /* get dom/4 term */
  dvar->min = DomainMin(tmp);
  dvar->max = DomainMax(tmp);
  dvar->set = DomainSet(tmp);
  dvar->flags = DV_SET_OK;
  if (DomainSize(tmp)==TaggedOne)
    dvar->flags |= DV_PRUNED_VAL;
  if (CTagToCdr(dvar->set)==EmptySet)
    dvar->flags |= DV_INTERVAL;
}

void
dvar_assign MAGIC (HIDDEN_PROTO
		   Dvar dest,
		   Dvar source)
{
  dest->min = source->min;
  dest->max = source->max;
  dest->set = dvar_set(source);
  dest->flags = (source->flags&(DV_PRUNED_VAL|DV_INTERVAL))|DV_SET_OK;
}

/* access primitives */

BOOL
dvar_is_alias(Dvar dv1,Dvar dv2)
{
  TAGGED v1 = RefGlob(dv1->var_ref);
  TAGGED v2 = RefGlob(dv2->var_ref);

  DerefSwitch(v1,;);
  DerefSwitch(v2,;);
  return (v1==v2);
}

TAGGED
dvar_set MAGIC (HIDDEN_PROTO
		Dvar dvar)
{
  if (!(dvar->flags & DV_SET_OK)) {
    dvar->flags |= DV_SET_OK;
    if (dvar->flags & DV_INTERVAL)
      dvar->set = fd_interval(dvar->min, dvar->max);
    else
      dvar->set = fd_and_interval(dvar->set, dvar->min, dvar->max);
  }
  return dvar->set;
}

/* set is assumed to be finite */
long dvar_value_count MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                             Dvar dvar)
{
  long count;
  TAGGED set, range, tail, a, b;

  if (dvar->flags & DV_INTERVAL)
    return GetSmall(dvar->max - dvar->min + TaggedOne);

  count = 0;
  for (set=dvar->set; set!=EmptySet; set=tail) {
    range = CTagToCar(set);
    tail = CTagToCdr(set);
    a = RangeMin(range);
    b = RangeMax(range);
    if (!(dvar->flags & DV_SET_OK)) {
      switch (val_vs_range(dvar->min,range)) {
      case CMP_INSIDE:
	a = dvar->min;
      case CMP_BEFORE:
	break;
      default:
	continue;
      }
      switch (val_vs_range(dvar->max,range)) {
      case CMP_INSIDE:
	b = dvar->max;
	tail = EmptySet;
      case CMP_AFTER:
	break;
      default:
	return count;
      }
    }
    count += GetSmall(b-a+TaggedOne);
  }
  return count;
}

long
dvar_interval_count MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                           Dvar dvar)
{
  long count;
  TAGGED set, range, tail;

  if (dvar->flags & DV_INTERVAL)
    return 1;

  count = 0;
  for (set=dvar->set; set!=EmptySet; set=tail) {
    range = CTagToCar(set);
    tail = CTagToCdr(set);
    if (!(dvar->flags & DV_SET_OK)) {
      switch (val_vs_range(dvar->min,range)) {
      case CMP_INSIDE:
      case CMP_BEFORE:
	break;
      default:
	continue;
      }
      switch (val_vs_range(dvar->max,range)) {
      case CMP_INSIDE:
	tail = EmptySet;
      case CMP_AFTER:
	break;
      default:
	return count;
      }
    }
    count++;
  }
  return count;
}


/* comparisons */

int
dvar_compare_set MAGIC (HIDDEN_PROTO
			Dvar dvar,
			TAGGED set)
{
  BOOL ok = (dvar->flags & DV_SET_OK);
  int rc;

  if (dvar->flags & DV_INTERVAL) {
    switch ((rc=fd_compare_interval(set,dvar->min,dvar->max))) {
    case FDI_SUPERSET:
      return FDI_SUBSET;
    case FDI_SUBSET:
      return FDI_SUPERSET;
    default:
      return rc;
    }
  }

  switch (fd_compare(dvar->set,set)) {
  case FDI_EQUAL:
    if (ok)
      return FDI_EQUAL;
    switch (fd_compare_interval(set,dvar->min,dvar->max)) {
      /* "wrong" order */
    case FDI_SUPERSET:
    case FDI_INTERSECT:
      return FDI_SUBSET;
    case FDI_EQUAL:
    case FDI_SUBSET:
    case FDI_DISJOINT:
    default:
      return FDI_EQUAL;
    }
    
  case FDI_SUBSET:
    return FDI_SUBSET;
    
  case FDI_DISJOINT:
    return FDI_DISJOINT;

  case FDI_SUPERSET:
    if (ok)
      return FDI_SUPERSET;
    switch (fd_compare_interval(set,dvar->min,dvar->max)) {
      /* "wrong" order */
    case FDI_EQUAL:
      return FDI_EQUAL;
    case FDI_SUPERSET:
      return FDI_SUBSET;
    case FDI_DISJOINT:
      return FDI_DISJOINT;
    case FDI_SUBSET:
    case FDI_INTERSECT:
    default:
      goto general;
    }
    
  case FDI_INTERSECT:
    if (ok)
      return FDI_INTERSECT;
    switch (fd_compare_interval(set,dvar->min,dvar->max)) {
      /* "wrong" order */
    case FDI_EQUAL:
    case FDI_SUPERSET:
      return FDI_SUBSET;
    case FDI_DISJOINT:
      return FDI_DISJOINT;
    case FDI_SUBSET:
    case FDI_INTERSECT:
    default:
      goto general;
    }
  }
 general:
  dvar->flags |= DV_SET_OK;
  dvar->set = fd_and_interval(dvar->set, dvar->min, dvar->max);
  return fd_compare(dvar->set, set);
}

int
dvar_compare_interval_t MAGIC (HIDDEN_PROTO
			       Dvar dvar,
			       TAGGED tmin,
			       TAGGED tmax)
{
  BOOL ok = (dvar->flags & DV_SET_OK);

  if (dvar->flags & DV_INTERVAL)
    return fd_compare_intervals(dvar->min,dvar->max,tmin,tmax);

  switch (fd_compare_interval(dvar->set,tmin,tmax)) {
  case FDI_EQUAL:
    if (ok)
      return FDI_EQUAL;
    switch (fd_compare_intervals(tmin,tmax,dvar->min,dvar->max)) {
      /* "wrong" order */
    case FDI_SUPERSET:
    case FDI_INTERSECT:
      return FDI_SUBSET;
    case FDI_EQUAL:
    case FDI_SUBSET:
    case FDI_DISJOINT:
    default:
      return FDI_EQUAL;
    }
    
  case FDI_SUBSET:
    return FDI_SUBSET;
    
  case FDI_DISJOINT:
    return FDI_DISJOINT;

  case FDI_SUPERSET:
    if (ok)
      return FDI_SUPERSET;
    switch (fd_compare_intervals(tmin,tmax,dvar->min,dvar->max)) {
      /* "wrong" order */
    case FDI_EQUAL:
      return FDI_EQUAL;
    case FDI_SUPERSET:
      return FDI_SUBSET;
    case FDI_DISJOINT:
      return FDI_DISJOINT;
    case FDI_SUBSET:
    case FDI_INTERSECT:
    default:
      goto general;
    }
    
  case FDI_INTERSECT:
    if (ok)
      return FDI_INTERSECT;
    switch (fd_compare_intervals(tmin,tmax,dvar->min,dvar->max)) {
      /* "wrong" order */
    case FDI_EQUAL:
    case FDI_SUPERSET:
      return FDI_SUBSET;
    case FDI_DISJOINT:
      return FDI_DISJOINT;
    case FDI_SUBSET:
    case FDI_INTERSECT:
    default:
      goto general;
    }
  }
 general:
  dvar->flags |= DV_SET_OK;
  dvar->set = fd_and_interval(dvar->set, dvar->min, dvar->max);
  return fd_compare_interval(dvar->set, tmin, tmax);
}

BOOL
dvar_contains_value_t MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                             Dvar dvar, TAGGED tvalue)
{
  return
    InInterval(tvalue,dvar->min,dvar->max) &&
    ((dvar->flags & DV_INTERVAL) || fd_member(tvalue,dvar->set));
}

/* iterators etc. */

/* -CLPFD_MAXINT means no predecessor */
/* if no successor, CLPFD_MAXINT is returned */
long
dvar_successor_l MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                        Dvar dvar,long val)
{
  TAGGED tval =
    dvar_successor_t(dvar, val==-CLPFD_MAXINT ? Inf : MakeSmall(val));

  return (tval==Sup ? CLPFD_MAXINT : GetSmall(tval));
}

/* Inf means no predecessor */
/* if no successor, Sup is returned */
TAGGED
dvar_successor_t MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                        Dvar dvar,TAGGED tval)
{
  if (FDlt(tval,dvar->min))
    return dvar->min;
  if (FDgt(tval,dvar->max))
    return Sup;
  tval += IStep(1);
  if (!(dvar->flags & DV_INTERVAL)) {
    TAGGED set = dvar->set;
    TAGGED range;

    while (set!=EmptySet) {
      range = CTagToCar(set); 
      set = CTagToCdr(set);   
      switch (val_vs_range(tval,range)) {
      case CMP_BEFORE:
	tval = RangeMin(range);
      case CMP_INSIDE:
	set = EmptySet;
      }
    }
  }
  return tval;
}


/* CLPFD_MAXINT means no successor */
/* if no predecessor, -CLPFD_MAXINT is returned */
long
dvar_predecessor_l MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                          Dvar dvar,long val)
{
  TAGGED tval =
    dvar_predecessor_t(dvar, val==CLPFD_MAXINT ? Sup : MakeSmall(val));

  return (tval==Inf ? -CLPFD_MAXINT : GetSmall(tval));
}

/* Inf means no successor */
/* if no predecessor, Inf is returned */
TAGGED
dvar_predecessor_t MAGIC(DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                         Dvar dvar,TAGGED tval)
{
  if (FDgt(tval,dvar->max))
    return dvar->max;
  if (FDlt(tval,dvar->min))
    return Inf;
  tval -= IStep(1);
  if (!(dvar->flags & DV_INTERVAL)) {
    TAGGED set = dvar->set;
    TAGGED range;
    TAGGED prevmax = Inf;

    while (set!=EmptySet) {
      range = CTagToCar(set); 
      set = CTagToCdr(set);   
      switch (val_vs_range(tval,range)) {
      case CMP_BEFORE:
	tval = prevmax;
      case CMP_INSIDE:
	set = EmptySet;
      }
      prevmax = RangeMax(range);
    }
  }
  return tval;
}


void 
dviter_init MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                   DVITER *it, Dvar dvar)
{
  TAGGED min = dvar->min;
  TAGGED max = dvar->max;
  TAGGED fdset = 0;
  
  if (!(dvar->flags & DV_INTERVAL)) {
    fdset = dvar->set;
    /* skip intervals that are < min */
    while (TagIsLST(fdset)) {
      TAGGED r = CTagToCar(fdset);
      if (!FDlt(RangeMax(r),min)) {
	if (FDlt(min,RangeMin(r)))
	  min = RangeMin(r);
	break;
      }
      fdset = CTagToCdr(fdset);
    }
  }
  it->min = min;
  it->max = max;
  it->fdset = fdset;
}

void
dviter_next_interval_l MAGIC(DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                             DVITER *it, long *min, long *max)
{
  TAGGED tmin, tmax;

  dviter_next_interval_t(it,&tmin,&tmax);
  *min = GetSmall(tmin);
  *max = GetSmall(tmax);
}

void
dviter_next_interval_t MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                              DVITER *it, TAGGED *tmin, TAGGED *tmax)
{
  TAGGED fdset = it->fdset;
    
  if (!fdset) {			/* (dvar->flags & DV_INTERVAL) */
    *tmin = it->min;
    *tmax = it->max;
    fdset = EmptySet;
  } else {
    TAGGED r = CTagToCar(fdset);
    
    *tmin = it->min;
    *tmax = FDlt(RangeMax(r),it->max) ? RangeMax(r) : it->max;
    fdset = CTagToCdr(fdset);
    if (TagIsLST(fdset)) {
      r = CTagToCar(fdset);
      it->min = RangeMin(r);
      if (FDgt(it->min,it->max)) /* detect end */
	fdset = EmptySet;
    }
  }
  it->fdset = fdset;
}

TAGGED
dviter_next_value_t MAGIC(DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                          DVITER *it)
{
  TAGGED fdset = it->fdset;
  TAGGED tnext = it->min;
    
  it->min = tnext+IStep(1);
  if (it->max==tnext)
    it->fdset = EmptySet;
  else if (fdset) {		/* !(dvar->flags & DV_INTERVAL) */
    TAGGED r = CTagToCar(fdset);
    
    if (RangeMax(r)==tnext) {
      fdset = CTagToCdr(fdset);
      if (TagIsLST(fdset)) {
	r = CTagToCar(fdset);
	it->min = RangeMin(r);
	if (FDgt(it->min,it->max)) /* detect end */
	  fdset = EmptySet;
      }
      it->fdset = fdset;
    }
  }
  return tnext;
}

/* ensure next elt will be > t */
void
dviter_skip_t MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                     DVITER *it, TAGGED t)
{
  if (FDlt(it->min,t+IStep(1)))
    it->min = t+IStep(1);
  if (it->fdset) {		/* !(dvar->flags & DV_INTERVAL) */
    TAGGED r = CTagToCar(it->fdset);
    
    while (FDgt(it->min,RangeMax(r))) {
      it->fdset = CTagToCdr(it->fdset);
      if (dviter_empty(it))
	return;
      r = CTagToCar(it->fdset);
    }
    if (FDlt(it->min,RangeMin(r)))
      it->min = RangeMin(r);
  }
  if (FDgt(it->min,it->max))
    it->fdset = EmptySet;
}

/* Precondition:
     Domain not empty.
     tmin, tmax, tset to become new dvar fields, if successful.
   Postconditions:
     dvar->min, dvar->max adjusted to existing domain elements.
     dvar->flags & DV_INTERVAL set to its correct value.
     dvar->flags & DV_SET_OK   set to its correct value.
     Value, i.e. rc and inferred values, or:ed into dvar->flags.
     On failure, dvar fields are not written.
*/
static int
dvar_adjust_bounds MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                          Dvar dvar,int rc,TAGGED tmin,TAGGED tmax,TAGGED tset)
{
  TAGGED d, r, rprev=0;
  int cmp;
  BOOL ok = TRUE;
  
  r = CTagToCar(tset);
  d = CTagToCdr(tset);
  cmp = point_vs_range(tmin,r);
  while (d!=EmptySet && cmp==CMP_AFTER) {
    tset = d;			/* optimization */
    r = CTagToCar(d);
    d = CTagToCdr(d);
    cmp = point_vs_range(tmin,r);
  }
  if (cmp==CMP_BEFORE) {
    rc |= DV_PRUNED_MIN;
    tmin = RangeMin(r);
  }
  if (tmin!=RangeMin(r))
    ok = FALSE;
  if (FDgt(tmin,tmax))		/* e.g. min=0, max=2, set=3..3 */
    return -1;
  cmp = point_vs_range(tmax,r);
  while (d!=EmptySet && cmp==CMP_AFTER) {
    rprev = r;
    r = CTagToCar(d);
    d = CTagToCdr(d);
    cmp = point_vs_range(tmax,r);
  }
  switch (cmp) {
  case CMP_BEFORE:
    r = rprev;
    ok = FALSE;
  case CMP_AFTER:
    rc |= DV_PRUNED_MAX;
    tmax = RangeMax(r);
  }
  if (d!=EmptySet || tmax!=RangeMax(r))
    ok = FALSE;
				/* succeess - store dvar fields */
  dvar->flags &= ~(DV_SET_OK|DV_INTERVAL);
  if (point_vs_range(tmin,r)==CMP_INSIDE)
    dvar->flags |= DV_INTERVAL;
  if (ok)
    dvar->flags |= DV_SET_OK;
  if (tmin==tmax)
    rc |= DV_PRUNED_VAL;
  dvar->flags |= (rc & ~DV_PRUNED_DOM);
  dvar->min = tmin;
  dvar->max = tmax;
  dvar->set = tset;
  return rc;
}

/* pruning: remove the arg. from the domain */

int
dvar_prune_interval_t MAGIC (HIDDEN_PROTO
			     Dvar dvar,
			     TAGGED lbt,
			     TAGGED ubt)
{
  int rc = 0;
  
  if (FDgt(lbt,ubt))
    return 0;

  switch (fd_interval_cmp(lbt,ubt,dvar->min,dvar->max)) {
  case FD_BEFORE:
  case FD_MEETS:
  case FD_MET_BY:
  case FD_AFTER:
    return 0;
  case FD_OVERLAPS:
  case FD_STARTS:
    rc = DV_PRUNED_DOM|DV_PRUNED_MIN;
    if (dvar->flags & DV_INTERVAL) {
      dvar->flags &= ~DV_SET_OK;
      dvar->min = ubt+IStep(1);
      break;
    } else
      return dvar_adjust_bounds(dvar,rc,ubt+IStep(1),dvar->max,dvar->set);
  case FD_DURING:
    if ((dvar->flags & DV_INTERVAL) ||
	fd_compare_interval(dvar->set,lbt,ubt)!=FDI_DISJOINT) {
      dvar->flags |= DV_PRUNED_DOM;
      dvar->flags &= ~DV_INTERVAL;
      dvar->set = fd_subtract_interval(dvar->set,lbt,ubt);
      rc = DV_PRUNED_DOM;
    }
    break;
  case FD_FINISHES:
  case FD_OVERLAPPED_BY:
    rc = DV_PRUNED_DOM|DV_PRUNED_MAX;
    if (dvar->flags & DV_INTERVAL) {
      dvar->flags &= ~DV_SET_OK;
      dvar->max = lbt-IStep(1);
      break;
    } else
      return dvar_adjust_bounds(dvar,rc,dvar->min,lbt-IStep(1),dvar->set);
  case FD_FINISHED_BY:
  case FD_CONTAINS:
  case FD_EQUALS:
  case FD_STARTED_BY:
  default:
    return -1;
  }
  if (dvar->min==dvar->max)
    rc |= DV_PRUNED_VAL;
  dvar->flags |= (rc & ~DV_PRUNED_DOM);
  return rc;
}

int
dvar_prune_set MAGIC (HIDDEN_PROTO
		      Dvar dvar,
		      TAGGED set)
{
  if (set==EmptySet)
    return 0;
  switch (fd_compare_interval(set,dvar->min,dvar->max)) {
  case FDI_EQUAL:
  case FDI_SUPERSET:
    return -1;
  case FDI_DISJOINT:
    return 0;
  case FDI_SUBSET:
  case FDI_INTERSECT:
  default:
    switch ((dvar->flags & DV_INTERVAL)
	    ? FDI_SUBSET
	    : fd_compare(set,dvar->set)) {
    case FDI_EQUAL:
    case FDI_SUPERSET:
      return -1;
    case FDI_DISJOINT:
      return 0;
    case FDI_SUBSET:
    case FDI_INTERSECT:
    default:
      dvar->flags |= DV_PRUNED_DOM;
      return dvar_adjust_bounds(dvar,DV_PRUNED_DOM,
				dvar->min,dvar->max,fd_subtract(dvar->set,set));
    }
  }
}

/* fixing: assign the arg. to the domain */

int
dvar_fix_interval_t MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                           Dvar dvar,TAGGED lbt,TAGGED ubt)
{
  int rc = 0;
  TAGGED tmin = dvar->min;
  TAGGED tmax = dvar->max;
  
  if (FDgt(lbt,tmin))
    rc |= DV_PRUNED_MIN|DV_PRUNED_DOM;
  if (FDlt(ubt,tmax))
    rc |= DV_PRUNED_MAX|DV_PRUNED_DOM;
  if (rc) {
    if (rc & DV_PRUNED_MIN)
      tmin = lbt;
    if (rc & DV_PRUNED_MAX)
      tmax = ubt;
    if (FDgt(tmin,tmax))
      return -1;
    if (dvar->flags & DV_INTERVAL)
      dvar->flags &= ~DV_SET_OK;
    else
      return dvar_adjust_bounds(dvar,rc,tmin,tmax,dvar->set);
    if (tmin==tmax)
      rc |= DV_PRUNED_VAL;
    dvar->flags |= (rc & ~DV_PRUNED_DOM);
    dvar->min = tmin;
    dvar->max = tmax;
  }
  return rc;
}

int
dvar_fix_set MAGIC (HIDDEN_PROTO
		    Dvar dvar,
		    TAGGED set)
{
  int cmp;
  
  if (set==EmptySet)
    return -1;
  switch ((cmp=fd_compare_interval(set,dvar->min,dvar->max))) {
  case FDI_EQUAL:
  case FDI_SUPERSET:
    return 0;
  case FDI_DISJOINT:
    return -1;
  case FDI_SUBSET:
  case FDI_INTERSECT:
  default:
    switch ((dvar->flags & DV_INTERVAL)
	    ? cmp
	    : fd_compare(set,dvar->set)) {
    case FDI_EQUAL:
    case FDI_SUPERSET:
      return 0;
    case FDI_DISJOINT:
      return -1;
    case FDI_SUBSET:
      dvar->flags |= DV_PRUNED_DOM;
      return dvar_adjust_bounds(dvar,DV_PRUNED_DOM,dvar->min,dvar->max,set);
    case FDI_INTERSECT:
    default:
      dvar->flags |= DV_PRUNED_DOM;
      return dvar_adjust_bounds(dvar,DV_PRUNED_DOM,dvar->min,dvar->max,
				(dvar->flags & DV_INTERVAL) ? set :
				fd_and(dvar->set,set));
    }
  }
}

/* pruning done */

void
dvar_pruning_done MAGIC (HIDDEN_PROTO
			 Dvar dvar)
{
  WAMENV;
  
  if ((dvar->flags & (DV_PRUNED_DOM|DV_INTERVAL))==DV_PRUNED_DOM) {
    if (!(dvar->flags & DV_SET_OK)) {
      dvar->flags |= DV_SET_OK;
      dvar->set = fd_and_interval(dvar->set, dvar->min, dvar->max);
    }
    dvar->set = fd_localize(w,dvar->set);
  }
}

/* exporting bindings */

/* Assuming fdset is localized or interval. */
/* Assumption that fdset neither contains old domain nor that they are disjoint
   is unsafe in the context of co-references.
*/

static void 
request_tell_value MAGIC (HIDDEN_PROTO
			  Argdecl,
			  TAGGED dest_attribute, TAGGED dest_var,
			  TAGGED value,
			  int outarg, int nlive)
{
  TAGGED *h, functor=0, old, dom;
  int why = 0;

  X(nlive) = dest_attribute;		/* preserve over GC */
  X(nlive+1) = dest_var;
  DerefAttribute(dom,dest_attribute);
  if (fd.debugging) {
    RequireHeap(5,nlive+2);	/* GC */
    h = w->global_top;
    h[0] = fd.functor_eq;
    h[1] = X(nlive+1);
    h[2] = value;
    h[3] = MakeStructure(h);
    h[4] = X(outarg);
    w->global_top = h+5;
    X(outarg) = MakeList(h+3);
  } else {
    old = DomainSet(dom);
    if (!fd_member(value,old)) {
#if DBG>1
      fprintf(stderr,"* FD: failing request_tell_value, co-reference?\n");
#endif
      why = -1;
    } else if (DomainSize(dom)==MakeSmall(1)) {
#if DBG>1
      fprintf(stderr,"* FD: useless request_tell_value, co-reference?\n");
#endif
    } else {
      why = fd_tell_value(w,old,value,nlive+2); /* GC */
    }
    if (why<0) {
      RequireHeap(2,nlive+2);	/* GC */
      h = w->global_top;
      h[0] = atom_fail;
      h[1] = X(outarg);
      w->global_top = h+2;
      X(outarg) = MakeList(h);
      return;
    } else if (why==0) {
      return;
    } else if (1) {
      DerefSwitch(X(nlive+1),fd_told(w,why&7,nlive+2););
      /* otherwise, co-reference already dealt with */
    } else {
      switch (why & 7) {
      case 1:			/* impossible */
	functor = fd.functor_dom;
	break;
      case 2:
      case 3:
	functor = fd.functor_min;
	break;
      case 4:
      case 5:
	functor = fd.functor_max;
	break;
      case 6:
      case 7:
	functor = fd.functor_minmax;
      }
      RequireHeap(4,nlive+2);	/* GC */
      h = w->global_top;
      h[0] = functor;
      h[1] = X(nlive+1);
      h[2] = MakeStructure(h);
      h[3] = X(outarg);
      w->global_top = h+4;
      X(outarg) = MakeList(h+2);
    }
  }
}

static void 
request_tell_interval MAGIC (HIDDEN_PROTO
			     Argdecl,
			     TAGGED dest_attribute, TAGGED dest_var,
			     TAGGED min, TAGGED max,
			     int outarg, int nlive)
{
  TAGGED *h, functor, old, min1, max1, dset;
  int why;

  if (min==max) {
    request_tell_value(Arg, dest_attribute, dest_var, min, outarg, nlive);
    return;
  }
  functor = 0;
  why = 0;
  X(nlive) = dest_attribute;		/* preserve over GC */
  X(nlive+1) = dest_var;
  DerefAttribute(old,dest_attribute);
  dset = DomainSet(old);
  if (fd.debugging) {
    dset = fd_globalize(w,fd_interval(min,max),5,nlive+2);
    h = w->global_top;
    h[0] = fd.functor_in_set2;
    h[1] = X(nlive+1);
    h[2] = dset;
    h[3] = MakeStructure(h);
    h[4] = X(outarg);
    w->global_top = h+5;
    X(outarg) = MakeList(h+3);
  } else {
    switch (fd_compare_interval(dset,min,max)) {
    case FDI_SUBSET:
    case FDI_EQUAL:
#if DBG>1
      fprintf(stderr,"* FD: useless request_tell_interval, co-reference?\n");
#endif
      break;
    case FDI_DISJOINT:
#if DBG>1
      fprintf(stderr,"* FD: failing request_tell_interval, co-reference?\n");
#endif
      why = -1;
      break;
    case FDI_INTERSECT:
      if (CTagToCdr(dset)!=EmptySet) {
	why = fd_tell(w,dset,
		      fd_localize_if_holes(w,fd_and_interval(dset,min,max)),
		      nlive+2); /* GC */
	break;
      }
    case FDI_SUPERSET:
      min1 = DomainMin(old);
      max1 = DomainMax(old);
      if (min1!=min) {
	if (!EmptyInterval(min1,min))
	  why |= MASK_MIN+MASK_MINMAX+MASK_DOM;
	else
	  min = min1;
      }
      if (max!=max1) {
	if (!EmptyInterval(max,max1))
	  why |= MASK_MAX+MASK_MINMAX+MASK_DOM;
	else
	  max = max1;
      }
      why = fd_tell_interval(w,dset,min,max,why,nlive+2); /* GC */
    }
    if (why<0) {
      RequireHeap(2,nlive+2);	/* GC */
      h = w->global_top;
      h[0] = atom_fail;
      h[1] = X(outarg);
      w->global_top = h+2;
      X(outarg) = MakeList(h);
      return;
    } else if (why==0) {
      return;
    } else if (1) {
      DerefSwitch(X(nlive+1),fd_told(w,why&7,nlive+2););
      /* otherwise, co-reference already dealt with */
    } else {
      switch (why & 7) {
      case 1:			/* impossible */
	functor = fd.functor_dom;
	break;
      case 2:
      case 3:
	functor = fd.functor_min;
	break;
      case 4:
      case 5:
	functor = fd.functor_max;
	break;
      case 6:
      case 7:
	functor = fd.functor_minmax;
      }
      RequireHeap(4,nlive+2);	/* GC */
      h = w->global_top;
      h[0] = functor;
      h[1] = X(nlive+1);
      h[2] = MakeStructure(h);
      h[3] = X(outarg);
      w->global_top = h+4;
      X(outarg) = MakeList(h+2);
    }
  }
}

static void 
request_tell MAGIC (HIDDEN_PROTO
		    Argdecl,
		    TAGGED dest_attribute, TAGGED dest_var, TAGGED fdset,
		    int outarg, int nlive)
{
  TAGGED *h, functor, old;
  int why;

  if (CTagToCdr(fdset)==EmptySet) { /* FDset is an interval */
    TAGGED fdsetrange = CTagToCar(fdset);
    TAGGED min = RangeMin(fdsetrange);
    TAGGED max = RangeMax(fdsetrange);
      
    request_tell_interval(Arg, dest_attribute, dest_var,
			  min, max, outarg, nlive);
    return;
  }
  functor = 0;
  why = 0;
  X(nlive) = dest_attribute;		/* preserve over GC */
  X(nlive+1) = dest_var;
  if (fd.debugging) {
    fdset = fd_globalize(w,fdset,5,nlive+2);
    h = w->global_top;
    h[0] = fd.functor_in_set2;
    h[1] = X(nlive+1);
    h[2] = fdset;
    h[3] = MakeStructure(h);
    h[4] = X(outarg);
    w->global_top = h+5;
    X(outarg) = MakeList(h+3);
  } else {
    DerefAttribute(old,X(nlive));
    old = DomainSet(old);
    switch (fd_compare(old,fdset)) {
    case FDI_SUBSET:
    case FDI_EQUAL:
#if DBG>1
      fprintf(stderr,"* FD: useless request_tell, co-reference?\n");
#endif
      break;
    case FDI_DISJOINT:
#if DBG>1
      fprintf(stderr,"* FD: failing request_tell, co-reference?\n");
#endif
      why = -1;
      break;
    case FDI_SUPERSET:
      why = fd_tell(w,old,fdset,nlive+2); /* GC */
      break;
    case FDI_INTERSECT:
      if (CTagToCdr(old)==EmptySet) {
	TAGGED range = CTagToCar(old);
	TAGGED min = RangeMin(range);
	TAGGED max = RangeMax(range);

	why = fd_tell(w,old,fd_and_interval(fdset,min,max),nlive+2); /* GC */
      } else
	why = fd_tell(w,old,fd_localize_if_holes(w,fd_and(old,fdset)),
		      nlive+2); /* GC */
    }
    if (why<0) {
      RequireHeap(2,nlive+2);	/* GC */
      h = w->global_top;
      h[0] = atom_fail;
      h[1] = X(outarg);
      w->global_top = h+2;
      X(outarg) = MakeList(h);
      return;
    } else if (why==0) {
      return;
    } else if (1) {
      DerefSwitch(X(nlive+1),fd_told(w,why&7,nlive+2););
      /* otherwise, co-reference already dealt with */
    } else {
      switch (why & 7) {
      case 1:
	functor = fd.functor_dom;
	break;
      case 2:
      case 3:
	functor = fd.functor_min;
	break;
      case 4:
      case 5:
	functor = fd.functor_max;
	break;
      case 6:
      case 7:
	functor = fd.functor_minmax;
      }
      RequireHeap(4,nlive+2);	/* GC */
      h = w->global_top;
      h[0] = functor;
      h[1] = X(nlive+1);
      h[2] = MakeStructure(h);
      h[3] = X(outarg);
      w->global_top = h+4;
      X(outarg) = MakeList(h+2);
    }
  }
}

static void 
request_rewrite_eq MAGIC (HIDDEN_PROTO
			  Argdecl, TAGGED x_var, TAGGED y_var,
			  int outarg, int nlive)
{
  TAGGED *h;
  
  if (x_var==y_var) {
  } else if (!fd.debugging) {
    (void)cunify(w,x_var,y_var);
  } else {
    X(nlive) = x_var;
    X(nlive+1) = y_var;
    RequireHeap(7,nlive+2);	/* GC */
    h = w->global_top;
    h[0] = fd.functor_eq;
    h[1] = X(nlive);
    h[2] = X(nlive+1);
    h[3] = fd.functor_call;
    h[4] = MakeStructure(h);
    h[5] = MakeStructure(h+3);
    h[6] = X(outarg);
    w->global_top = h+7;
    X(outarg) = MakeList(h+5);
  }
}

static void 
request_rewrite_leqc MAGIC (HIDDEN_PROTO
			    Argdecl, TAGGED x_var, TAGGED y_var, int c,
			    int outarg, int nlive)
{
  TAGGED *h;
  
  X(nlive) = x_var;
  X(nlive+1) = y_var;
  RequireHeap(11,nlive+2);
  h = w->global_top;
  h[0] = fd.functor_leqc;
  h[1] = X(nlive);
  h[2] = X(nlive+1);
  h[3] = MakeSmall(c);
  h[4] = fd.functor_colon;
  h[5] = fd.fd_module->name;
  h[6] = MakeStructure(h);
  h[7] = fd.functor_call;
  h[8] = MakeStructure(h+4);
  h[9] = MakeStructure(h+7);
  h[10] = X(outarg);
  w->global_top = h+11;
  X(outarg) = MakeList(h+9);
}

void
dvar_export_start MAGIC (HIDDEN_PROTO
			 int arity)
{
  WAMENV;
  
  X(arity-1) = atom_nil;		/* actions list */
}


void
dvar_export MAGIC (HIDDEN_PROTO
		   Dvar dvar,
		   int arity)
{
  WAMENV;
  
  if ((dvar->flags & (DV_PRUNED_DOM|DV_INTERVAL))==DV_PRUNED_DOM)
    request_tell(w, RefGlob(dvar->attr_ref), RefGlob(dvar->var_ref),
		    dvar->set, arity-1, arity);
  else if (dvar->flags & (DV_PRUNED_MIN|DV_PRUNED_MAX))
    request_tell_interval(w, RefGlob(dvar->attr_ref), RefGlob(dvar->var_ref),
			     dvar->min, dvar->max, arity-1, arity);
}

void
dvar_export_equal MAGIC (HIDDEN_PROTO
			 Dvar dv1, Dvar dv2,
			 int arity)
{
  WAMENV;
  
  request_rewrite_eq(w, RefGlob(dv1->var_ref), RefGlob(dv2->var_ref),
		     arity-1, arity);
}

void
dvar_export_leqc MAGIC (HIDDEN_PROTO
			Dvar dv1, Dvar dv2, int c,
			int arity)
{
  WAMENV;
  
  request_rewrite_leqc(w, RefGlob(dv1->var_ref), RefGlob(dv2->var_ref), c,
		       arity-1, arity);
}

/* Support functions for new global API */
/* Preconditions:
   - All new domains localized, or protected in term refs.
   - ar X regs live; X(ar-1) is the tail of the action list 
   - CONTPAD heap words guaranteed
   Postconditions:
   - CONTPAD heap words guaranteed
*/
/* ent: -1 = fail, 0 = suspend, 1 = exit */
void
dvar_export_done MAGIC (HIDDEN_PROTO
			SP_term_ref Actions,
			int ent,
			int arity)
{
  WAMENV;
  TAGGED *h;
  TAGGED action = atom_exit;
  
  switch (ent) {
  case -1:
    action = atom_fail;
  case 1:
    RequireHeap(2,arity);	/* GC */
    h = w->global_top;
    h[0] = action;
    h[1] = X(arity-1);
    w->global_top = h+2;
    X(arity-1) = MakeList(h);
  }
  RefTerm(Actions) = X(arity-1);
}

#if DBG
/* Dump a dvar */
void
dvar_dump MAGIC (DYMA_HIDDEN_PROTO     /* [PM] 3.10.1 DynamicMallocBase */
                 Dvar dv)
{
  char mins[16], maxs[16], flags[100];

  strcpy(flags, "");
  if (dv->flags&DV_PRUNED_DOM)
    strcat(flags, "+PRUNED_DOM");
  if (dv->flags&DV_PRUNED_MIN)
    strcat(flags, "+PRUNED_MIN");
  if (dv->flags&DV_PRUNED_MAX)
    strcat(flags, "+PRUNED_MAX");
  if (dv->flags&DV_PRUNED_VAL)
    strcat(flags, "+PRUNED_VAL");
  if (dv->flags&DV_SET_OK)
    strcat(flags, "+SET_OK");
  if (dv->flags&DV_INTERVAL)
    strcat(flags, "+INTERVAL");

  if (TagIsSmall(dv->min))
    sprintf(mins, "%ld", GetSmall(dv->min));
  else
    strcpy(mins, "inf");
  if (TagIsSmall(dv->max))
    sprintf(maxs, "%ld", GetSmall(dv->max));
  else
    strcpy(maxs, "sup");
  
  printf("DVAR address=%p attr_ref=%p var_ref=%p\n",
	 dv, dv->attr_ref, dv->var_ref);
  printf("     flags=%s\n", flags+1);
  printf("     min=%s max=%s\n", mins, maxs);
  if (!(dv->flags&DV_INTERVAL)) {
    char *sep = "{";
    DVITER it;
    
    printf("     dom=");
    dviter_init(&it,dv);
    while (!dviter_empty(&it)) {
      TAGGED tmin, tmax;
      dviter_next_interval_t(&it, &tmin, &tmax);
      if (TagIsSmall(tmin))
	sprintf(mins, "%ld", GetSmall(tmin));
      else
	strcpy(mins, "inf");
      if (TagIsSmall(tmax))
	sprintf(maxs, "%ld", GetSmall(tmax));
      else
	strcpy(maxs, "sup");
      printf("%s%s..%s", sep, mins, maxs);
      sep = ",";
    }
    printf("}\n");
  }
}

/* test all cases of dvar_fix_set */
void
dvar_validate MAGIC (HIDDEN_PROTO_VOID)
{
  WAMENV;
  unsigned int oldmask, newmask, intersection;
  int oldarray[5], newarray[5], gotarray[5];
  int i, lb, ub, gotmin, gotmax, gothole, chk_set_ok, chk_interval;
  struct dvar dv;
  TAGGED oldset, newset, expected, got;
  FDCONS cons;
  FDITER iter;

  for (oldmask=1; oldmask<31; oldmask++)
    for (newmask=1; newmask<31; newmask++) {
      intersection = (oldmask&newmask);
      if (intersection!=oldmask) {
	w->numstack_end = NULL;
	for (i=0; i<5; i++) {
	  oldarray[i] = ((1<<i)&oldmask) ? 1 : 0;
	  newarray[i] = ((1<<i)&newmask) ? 1 : 0;
	}
	for (lb=0; lb<5; lb++)
	  if (oldarray[lb])
	    for (ub=lb; ub<5; ub++)
	      if (oldarray[ub]) {
		/* test instance now determined by oldarray,lb,ub, newarray */
		w->numstack_end = NULL;
		fdcons_init(&cons);
		for (i=0; i<5; i++)
		  if (oldarray[i])
		    fdcons_add(&cons,MakeSmall(i));
		oldset = fdcons_set(&cons);
		dv.set = oldset;
		dv.min = fd_min(dv.set);
		dv.max = fd_max(dv.set);
		dv.flags = DV_SET_OK;
		if (dv.min==dv.max)
		  dv.flags |= DV_PRUNED_VAL;
		if (CTagToCdr(dv.set)==EmptySet)
		  dv.flags |= DV_INTERVAL;
		fdcons_init(&cons);
		for (i=0; i<5; i++)
		  if (newarray[i])
		    fdcons_add(&cons,MakeSmall(i));
		newset = fdcons_set(&cons);
		
		expected = fd_and(fd_and_interval(dv.set,MakeSmall(lb),MakeSmall(ub)),
				  newset);
		for (i=0; i<5; i++)
		  gotarray[i] = 0;
		dvar_fix_interval_l(&dv, lb, ub);
		gotmin = 5;
		gotmax = 0;
		gothole = 0;
		if (dvar_fix_set(&dv, newset)<0)
		  got = EmptySet;
		else {
		  got = fd_and_interval(dv.set,dv.min,dv.max);
		  fditer_init(&iter, dv.set);
		  while (!fditer_empty(&iter))
		    gotarray[GetSmall(fditer_next(&iter))] = 1;
		  for (i=0; i<5; i++)
		    if (gotarray[i]) {
		      if (gotmin>i) gotmin = i;
		      gotmax = i;
		    }
		  for (i=dvar_min_l(&dv); i<=dvar_max_l(&dv); i++)
		    if (!gotarray[i])
		      gothole = 1;
		}
		chk_set_ok =
		  (got==EmptySet ||
		   !(dv.flags & DV_SET_OK) ==
		   (dvar_min_l(&dv)!=gotmin || dvar_max_l(&dv)!=gotmax));
		chk_interval =
		  (got==EmptySet || !(dv.flags & DV_INTERVAL) == gothole);
		if (fd_compare(expected,got)!=FDI_EQUAL ||
		    !chk_set_ok ||
		    !chk_interval) {
		  printf("! wrong result in dvar_fix_set\n");
		  printf("OLD SET = [%d %d %d %d %d]\n",
			 oldarray[0], oldarray[1], oldarray[2], oldarray[3], oldarray[4]);
		  printf("NEW SET = [%d %d %d %d %d]\n",
			 newarray[0], newarray[1], newarray[2], newarray[3], newarray[4]);
		  printf("GOT SET = [%d %d %d %d %d]\n",
			 gotarray[0], gotarray[1], gotarray[2], gotarray[3], gotarray[4]);
		  printf("oldlb=%d oldub=%d newlb=%d newub=%d flags=0x%x\n",
			 lb, ub, (int)dvar_min_l(&dv), (int)dvar_max_l(&dv), dv.flags);
		  return;
		}
	      }
      }
    }
}
#endif

#if 0

/* Prolog code to compute case analysis. */

:- use_module(library(ordsets)).

equal(p,r) --> opt(pr), opt(pqr), opt(q).
subset(p,r) --> opt(pr), opt(pqr), opt(q), must(r,qr).
superset(p,r) --> opt(pr), opt(pqr), opt(q), must(p,pq).
disjoint(p,r) --> must(p,pq), must(r,qr), opt(q).
intersect(p,r) --> must(p,pq), must(r,qr), must(pr,pqr), opt(q).

equal(q,r) --> opt(qr), opt(pqr), opt(p).
subset(q,r) --> opt(qr), opt(pqr), opt(p), must(r,pr).
superset(q,r) --> opt(qr), opt(pqr), opt(p), must(q,pq).
disjoint(q,r) --> must(q,pq), must(r,pr), opt(p).
intersect(q,r) --> must(q,pq), must(r,pr), must(qr,pqr), opt(p).

opt(X) --> []; [X].

must(X,Y) --> [X]; [Y]; [X,Y].

main(C1, C2, Relations) :-
	case(p, C1),
	case(q, C2),
	combine(C1, C2, Set),
	patterns_to_relations(Set, Relations).

combine(Rel1, Rel2, Set) :-
	findall(P, pattern(Rel1,P), Bag1),
	findall(Q, pattern(Rel2,Q), Bag2),
	sort(Bag1, Set1),
	sort(Bag2, Set2),
	ord_intersection(Set1, Set2, Set).

pattern(Rule, Set) :-
	phrase(Rule, Bag),
	sort(Bag, Set),
	ord_intersect(Set, [pq,pqr]).
	
case(X, equal(X,r)).
case(X, subset(X,r)).
case(X, superset(X,r)).
case(X, disjoint(X,r)).
case(X, intersect(X,r)).

patterns_to_relations([], []).
patterns_to_relations([S|L1], [R|L2]) :-
	pattern_to_relation(S, R),
	patterns_to_relations(L1, L2).

pattern_to_relation(Pat, equal) :-
	ord_disjoint([pq,pr,qr,r], Pat).
pattern_to_relation(Pat, subset) :-
	ord_disjoint([pq], Pat),
	ord_intersect([pr,qr,r], Pat).
pattern_to_relation(Pat, superset) :-
	ord_intersect([pq], Pat),
	ord_disjoint([pr,qr,r], Pat).
pattern_to_relation(Pat, disjoint) :-
	ord_intersect([pq], Pat),
	ord_intersect([pr,qr,r], Pat),
	ord_disjoint([pqr], Pat).
pattern_to_relation(Pat, intersect) :-
	ord_intersect([pq], Pat),
	ord_intersect([pr,qr,r], Pat),
	ord_intersect([pqr], Pat).

#endif

