/* Copyright(C) 2002, Swedish Institute of Computer Science */

/* Domain variable abstraction */
/*******************************/

#define DV_PRUNED_DOM 0x1	/* some domain element was pruned */
#define DV_PRUNED_MIN 0x2	/* min was adjusted */
#define DV_PRUNED_MAX 0x4	/* max was adjusted */
#define DV_PRUNED_VAL 0x8	/* domain is/became a singleton */
#define DV_SET_OK     0x10	/* means set \subseteq min..max */
#define DV_INTERVAL   0x20	/* domain is an interval */

struct dvar {
  TAGGED min;			/* kept up to date */
  TAGGED max;			/* kept up to date */
  TAGGED set;
  SP_globref attr_ref;
  SP_globref var_ref;
  unsigned int flags;
};

typedef struct dvar *Dvar;

typedef struct {
  TAGGED min, max, fdset;
} DVITER;

#if MULTI_SP_AWARE
#define dvar_init(A1,A2,A3) dvar_init(HIDDEN_ARG,A1,A2,A3)
#define dvar_assign(A1,A2) dvar_assign(HIDDEN_ARG,A1,A2)
#define dvar_set(A1) dvar_set(HIDDEN_ARG,A1)
#define dvar_compare_interval_t(A1,A2,A3) dvar_compare_interval_t(HIDDEN_ARG,A1,A2,A3)
#define dvar_compare_set(A1,A2) dvar_compare_set(HIDDEN_ARG,A1,A2)
#define dvar_prune_interval_t(A1,A2,A3) dvar_prune_interval_t(HIDDEN_ARG,A1,A2,A3)
#define dvar_prune_set(A1,A2) dvar_prune_set(HIDDEN_ARG,A1,A2)
#define dvar_fix_set(A1,A2) dvar_fix_set(HIDDEN_ARG,A1,A2)
#define dvar_pruning_done(A1) dvar_pruning_done(HIDDEN_ARG,A1)
#define dvar_export_start(A1) dvar_export_start(HIDDEN_ARG,A1)
#define dvar_export(A1,A2) dvar_export(HIDDEN_ARG,A1,A2)
#define dvar_export_equal(A1,A2,A3) dvar_export_equal(HIDDEN_ARG,A1,A2,A3)
#define dvar_export_leqc(A1,A2,A3,A4) dvar_export_leqc(HIDDEN_ARG,A1,A2,A3,A4)
#define dvar_export_done(A1,A2,A3) dvar_export_done(HIDDEN_ARG,A1,A2,A3)
#define dvar_validate() dvar_validate(HIDDEN_ARG)

#if DynamicMallocBase           /* [PM] 3.10.1 MAGIC needed for DynamicMallocBase */
#define dvar_value_count(A1) dvar_value_count(HIDDEN_ARG,A1)
#define dvar_successor_t(A1,A2) dvar_successor_t(HIDDEN_ARG,A1,A2)
#define dvar_successor_l(A1,A2) dvar_successor_l(HIDDEN_ARG,A1,A2)
#define dvar_predecessor_l(A1,A2) dvar_predecessor_l(HIDDEN_ARG,A1,A2)
#define dvar_predecessor_t(A1,A2) dvar_predecessor_t(HIDDEN_ARG,A1,A2)
#define dvar_interval_count(A1) dvar_interval_count(HIDDEN_ARG,A1)
#define dvar_contains_value_t(A1,A2) dvar_contains_value_t(HIDDEN_ARG, A1,A2)
#define dviter_init(A1,A2) dviter_init(HIDDEN_ARG,A1,A2)
#define dviter_next_value_t(A1) dviter_next_value_t(HIDDEN_ARG,A1)
#define dviter_next_interval_l(A1,A2,A3) dviter_next_interval_l(HIDDEN_ARG, A1,A2,A3)
#define dviter_next_interval_t(A1,A2,A3) dviter_next_interval_t(HIDDEN_ARG, A1,A2,A3)
#define dviter_skip_t(A1,A2) dviter_skip_t(HIDDEN_ARG, A1,A2)
#define dvar_fix_interval_t(A1,A2,A3) dvar_fix_interval_t(HIDDEN_ARG, A1,A2,A3)
#endif /* DynamicMallocBase */

#endif /* MULTI_SP_AWARE */

/* Purpose: 
     (Re)Initialize dvar when entering filtering alg.
   Synopsis:
     void dvar_init(dvar, attr_ref, var_ref);
   Arguments:
     Dvar dvar            - the dvar
     SP_globref attr_ref - SP_globref holding Prolog FD attribute
     SP_globref var_ref  - SP_globref holding Prolog FD varable
*/
extern void
dvar_init MAGIC (HIDDEN_PROTO
		 Dvar dvar,
		 SP_globref attr_ref,
		 SP_globref var_ref);

/* Purpose: 
     Clone a dvar, e.g. for constructive disjunction.
   Synopsis:
     void dvar_assign(dest, source);
   Arguments:
     Dvar dest            - destination dvar
     Dvar source          - source dvar
*/
extern void
dvar_assign MAGIC (HIDDEN_PROTO
		   Dvar dest,
		   Dvar source);

/* Purpose:
     Access various dvar attributes.
   Synopsis:
     BOOL   dvar_is_alias(dvar,dvar);
     BOOL   dvar_is_integer(dvar);
     BOOL   dvar_is_interval(dvar);
     long   dvar_min_l(dvar);
     long   dvar_max_l(dvar);
     TAGGED dvar_min_t(dvar);
     TAGGED dvar_max_t(dvar);
     TAGGED dvar_set(dvar);
     long   dvar_interval_count(dvar);
     long   dvar_value_count(dvar);
   Arguments:
     Dvar dvar          - the dvar
*/
#define dvar_is_integer(DVAR)  ((DVAR)->flags & DV_PRUNED_VAL)
#define dvar_is_integer_first(DVAR)  			\
((DVAR)->flags==(DV_PRUNED_VAL|DV_SET_OK|DV_INTERVAL))
#define dvar_is_interval(DVAR) ((DVAR)->flags & DV_INTERVAL)
#define dvar_min_t(DVAR) ((DVAR)->min)
#define dvar_min_l(DVAR) GetSmall((DVAR)->min)
#define dvar_max_t(DVAR) ((DVAR)->max)
#define dvar_max_l(DVAR) GetSmall((DVAR)->max)

extern BOOL
dvar_is_alias(Dvar dv1,Dvar dv2);

extern TAGGED
dvar_set MAGIC (HIDDEN_PROTO
		Dvar dvar);


extern long
dvar_interval_count MAGIC (DYMA_HIDDEN_PROTO Dvar dvar);


extern long
dvar_value_count MAGIC (DYMA_HIDDEN_PROTO Dvar dvar);


/* Purpose:
     Compare a dvar with a set/interval/value.
   Synopsis:
     int    dvar_compare_set(dvar, set);
     int    dvar_compare_interval_l(dvar, min, max);
     int    dvar_compare_interval_t(dvar, tmin, tmax);
     BOOL   dvar_contains_value_l(dvar, value);
     BOOL   dvar_contains_value_t(dvar, tvalue);
   Arguments:
     Dvar dvar          - the dvar
     TAGGED set         - the set to compare with
     long min           - min. of the interval to compare with
     long max           - max. of the interval to compare with
     long value         - the value to test for membership
     TAGGED tmin        - min. of the interval to compare with
     TAGGED tmax        - max. of the interval to compare with
     TAGGED tvalue      - the value to test for membership
   int  values:
     FDI_EQUAL          - domain is equal to set/interval
     FDI_SUBSET         - domain is a struct subset of set/interval
     FDI_SUPERSET       - domain is a strict superset of set/interval
     FDI_DISJOINT       - domain does not intersect set/interval,
                          which is non-empty
     FDI_INTERSECT      - domain is neither subset nor superset,
                          but intersects set/interval
   BOOL values:
     TRUE               - the value is in the domain
     FALSE              - the value is not in the domain
*/

extern int
dvar_compare_set MAGIC (HIDDEN_PROTO
			Dvar dvar, TAGGED set);

#define dvar_compare_interval_l(DVAR,MIN,MAX) \
        dvar_compare_interval_t(DVAR,MakeSmall(MIN),MakeSmall(MAX))

extern int
dvar_compare_interval_t MAGIC (HIDDEN_PROTO
			       Dvar dvar, TAGGED tmin, TAGGED tmax);

#define dvar_contains_value_l(DVAR,VALUE) \
        dvar_contains_value_t(DVAR,MakeSmall(VALUE))

extern BOOL
dvar_contains_value_t MAGIC (DYMA_HIDDEN_PROTO Dvar dvar, TAGGED tvalue);

/* Purpose:
     Get successor domain element.
   Synopsis:
     long   dvar_successor_l(dvar, val);
     TAGGED dvar_successor_t(dvar, tval);
   Arguments:
     Dvar dvar          - the dvar
     long val           - the predecessor element
                          -CLPFD_MAXINT means no predecessor
     TAGGED tval        - the predecessor element
                          Inf means no predecessor
   long   values:
     CLPFD_MAXINT       - no successor exists
     anything else      - successor
   TAGGED values:
     Sup                - no successor exists
     anything else      - successor
*/

extern long
dvar_successor_l MAGIC (DYMA_HIDDEN_PROTO
                        Dvar dvar,long val);

extern TAGGED
dvar_successor_t MAGIC (DYMA_HIDDEN_PROTO
                        Dvar dvar,TAGGED tval);

/* Purpose:
     Get predecessor domain element.
   Synopsis:
     long   dvar_predecessor_l(dvar, val);
     TAGGED dvar_predecessor_t(dvar, tval);
   Arguments:
     Dvar dvar          - the dvar
     long val           - the successor element
                          CLPFD_MAXINT means no successor
     TAGGED tval        - the successor element
                          Sup means no successor
   long   values:
     -CLPFD_MAXINT      - no predecessor exists
     anything else      - predecessor
   TAGGED values:
     Inf                - no predecessor exists
     anything else      - predecessor
*/

extern long
dvar_predecessor_l MAGIC (DYMA_HIDDEN_PROTO
                          Dvar dvar,long val);

extern TAGGED
dvar_predecessor_t MAGIC (DYMA_HIDDEN_PROTO
                          Dvar dvar,TAGGED tval);

/* Purpose:
     Iterate over domain intervals or values.
   Synopsis:
     void dviter_init(&it, dvar);
     BOOL dviter_empty(&it);
     void dviter_next_interval_l(&it, &min, &max);
     void dviter_next_interval_t(&it, &tmin, &tmax);
     long dviter_next_value_l(&it);
     TAGGED dviter_next_value_t(&it);
     void dviter_skip_l(&it, value);
     void dviter_skip_t(&it, tvalue);
   Arguments:
     Dvar dvar          - the dvar
     DVITER it          - the iterator
     long   min         - min. of the next interval
     long   max         - max. of the next interval
     long   value       - the next value
     TAGGED tmin        - min. of the next interval
     TAGGED tmax        - max. of the next interval
     TAGGED tvalue      - the next value
   BOOL values:
     TRUE               - no more intervals or values
     FALSE              - there are more intervals and values
*/

extern void 
dviter_init MAGIC (DYMA_HIDDEN_PROTO
                   DVITER *it, Dvar dvar);

#define dviter_empty(it) ((it)->fdset==EmptySet)

extern void
dviter_next_interval_l MAGIC (DYMA_HIDDEN_PROTO DVITER *it, long *min, long *max);

extern void
dviter_next_interval_t MAGIC (DYMA_HIDDEN_PROTO DVITER *it, TAGGED *tmin, TAGGED *tmax);

#define dviter_next_value_l(it) \
        GetSmall(dviter_next_value_t(it))

extern TAGGED
dviter_next_value_t MAGIC (DYMA_HIDDEN_PROTO DVITER *it);

#define dviter_skip_l(it,l) dviter_skip_t(it,MakeSmall(l))

extern void
dviter_skip_t MAGIC (DYMA_HIDDEN_PROTO DVITER *it, TAGGED t);

/* Purpose:
     Remove a set/interval/value from a dvar.
   Synopsis:
     int    dvar_prune_set(dvar, set);
     int    dvar_prune_interval_l(dvar, min, max);
     int    dvar_prune_interval_t(dvar, tmin, tmax);
     int    dvar_prune_value_l(dvar, value);
     int    dvar_prune_value_t(dvar, tvalue);
   Arguments:
     Dvar dvar          - the dvar to prune
     TAGGED set         - the set to remove
     long min           - min. of the interval to remove
     long max           - max. of the interval to remove
     long value         - the value to remove
     TAGGED tmin        - min. of the interval to remove
     TAGGED tmax        - max. of the interval to remove
     TAGGED tvalue      - the value to remove
   int  values:
     -1                 - failure (empty domain)
     >=0                - success, value is sum of DV_PRUNED_{DOM,MIN,MAX,VAL}
*/

#define dvar_prune_value_l(DVAR,L) dvar_prune_value_t(DVAR,MakeSmall(L))

#define dvar_prune_value_t(DVAR,T) dvar_prune_interval_t(DVAR,T,T)

#define dvar_prune_interval_l(DVAR,L,U) \
        dvar_prune_interval_t(DVAR,MakeSmall(L),MakeSmall(U))

extern int 
dvar_prune_interval_t MAGIC (HIDDEN_PROTO
			     Dvar dvar,TAGGED lbt,TAGGED ubt);

int 
dvar_prune_set MAGIC (HIDDEN_PROTO
		      Dvar dvar,TAGGED set);

/* Purpose:
     Fix a dvar to a set/interval/value.
   Synopsis:
     int    dvar_fix_set(dvar, set);
     int    dvar_fix_interval_l(dvar, min, max);
     int    dvar_fix_interval_t(dvar, tmin, tmax);
     int    dvar_fix_min_l(dvar, min);
     int    dvar_fix_min_t(dvar, tmin);
     int    dvar_fix_max_l(dvar, max);
     int    dvar_fix_max_t(dvar, tmax);
     int    dvar_fix_value_l(dvar, value);
     int    dvar_fix_value_t(dvar, tvalue);
   Arguments:
     Dvar dvar          - the dvar to prune
     TAGGED set         - the set to impose
     long min           - min. of the interval to impose
     long max           - max. of the interval to impose
     long value         - the value to impose
     TAGGED tmin        - min. of the interval to impose
     TAGGED tmax        - max. of the interval to impose
     TAGGED tvalue      - the value to impose
   int  values:
     -1                 - failure (empty domain)
     >=0                - success, value is sum of DV_PRUNED_{DOM,MIN,MAX,VAL}
*/

#define dvar_fix_value_l(DVAR,L) \
        dvar_fix_value_t(DVAR,MakeSmall(L))

#define dvar_fix_value_t(DVAR,T) \
        dvar_fix_interval_t(DVAR,T,T)

#define dvar_fix_interval_l(DVAR,L,U) \
        dvar_fix_interval_t(DVAR,MakeSmall(L),MakeSmall(U))

#define dvar_fix_min_l(DVAR,L) \
        dvar_fix_min_t(DVAR,MakeSmall(L))

#define dvar_fix_min_t(DVAR,TL) \
        dvar_fix_interval_t(DVAR,TL,Sup)

#define dvar_fix_max_l(DVAR,U) \
        dvar_fix_interval_t(DVAR,Inf,MakeSmall(U))

#define dvar_fix_max_t(DVAR,UL) \
        dvar_fix_interval_t(DVAR,Inf,UL)

extern int 
dvar_fix_interval_t MAGIC (DYMA_HIDDEN_PROTO Dvar dvar,TAGGED lbt,TAGGED ubt);

extern int 
dvar_fix_set MAGIC (HIDDEN_PROTO
		    Dvar dvar,TAGGED set);

/* Purpose: 
     Export dvars to Prolog environment.
     Signal success/failure/entailment.
     1. For each dvar: dvar_pruning_done(dvar)
     2. dvar_export_start(arity);
     3. For each dvar: dvar_export(dvar, arity);
     4. dvar_export_done(actions, ent, arity);
   Synopsis:
     void dvar_pruning_done(dvar);
     void dvar_export_start(arity);
     void dvar_export(dvar, arity);
     void dvar_done(actions, ent, arity);
   Arguments:
     Dvar dvar            - the dvar
     int  arity           - the arity of the filtering algo. (usually 3)
     int  ent             - the dvar
     SP_term_ref actions  - SP_term_ref holding actions list
*/

extern void
dvar_pruning_done MAGIC (HIDDEN_PROTO
			 Dvar dvar);

extern void
dvar_export_start MAGIC (HIDDEN_PROTO
			 int arity);

extern void
dvar_export MAGIC (HIDDEN_PROTO
		   Dvar dvar, int arity);

extern void
dvar_export_equal MAGIC (HIDDEN_PROTO
			 Dvar dv1, Dvar dv2, int arity);

extern void
dvar_export_leqc MAGIC (HIDDEN_PROTO
			Dvar dv1, Dvar dv2, int c, int arity);

extern void
dvar_export_done MAGIC (HIDDEN_PROTO
			SP_term_ref Actions, int ent, int arity);


#if DBG

#if DynamicMallocBase
#define dvar_dump(A1) dvar_dump(HIDDEN_ARG, A1)
#endif /* DynamicMallocBase */

extern void dvar_dump MAGIC (DYMA_HIDDEN_PROTO Dvar dv);
#endif /* DBG */

#if DBG
extern void dvar_validate MAGIC (HIDDEN_PROTO_VOID);
#endif


/* TODO:
   - handle inf/sup consistently for long typed ops
   - handle unification
*/
