/* Copyright(C) 2002, Swedish Institute of Computer Science */

#include "fd.h"
#include "dvars.h"

#undef alpha                    /* [PM] 3.10 used as variable name in this file but a pre-processor macro on ALPHA platforms... */

#if MULTI_SP_AWARE
#if 0                           /* not redefined. It uses FD_SETUP_SPENV to access the SPEnv */
#define lexc_destructor(A1) lexc_destructor(HIDDEN_ARG, A1)
#endif
#define lex_chain_pair(A1,A2,A3) lex_chain_pair(HIDDEN_ARG, A1,A2,A3)
#define repair_atleast(A1,A2,A3,A4,A5,A6,A7,A8) repair_atleast(HIDDEN_ARG, A1,A2,A3,A4,A5,A6,A7,A8)
#define repair_among(A1,A2,A3,A4,A5) repair_among(HIDDEN_ARG, A1,A2,A3,A4,A5)
#define among_case(A1,A2,A3,A4,A5) among_case(HIDDEN_ARG, A1,A2,A3,A4,A5)
#define prune_among(A1,A2,A3,A4,A5,A6) prune_among(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)
#define prune_among_det(A1,A2,A3,A4,A5,A6) prune_among_det(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)
#define prune_among_force(A1,A2,A3,A4) prune_among_force(HIDDEN_ARG, A1,A2,A3,A4)
#define prune_between(A1,A2,A3,A4,A5,A6) prune_between(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)
#define entailed_between(A1,A2,A3,A4,A5) entailed_between(HIDDEN_ARG, A1,A2,A3,A4,A5)
#define entailed_among(A1,A2,A3) entailed_among(HIDDEN_ARG, A1,A2,A3)
#define prunable_among(A1,A2,A3,A4) prunable_among(HIDDEN_ARG, A1,A2,A3,A4)
#define compute_limit(A1,A2,A3,A4,A5,A6) compute_limit(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)
#define adjust_vector_up(A1,A2,A3,A4,A5,A6) adjust_vector_up(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)
#define adjust_vector_down(A1,A2,A3,A4,A5,A6) adjust_vector_down(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)
#if DynamicMallocBase           /* [PM] 3.10.1 */
#define lex_chain_leqc(A1,A2,A3,A4) lex_chain_leqc(HIDDEN_ARG, A1,A2,A3,A4)
#endif /* DynamicMallocBase */
#endif /* MULTI_SP_AWARE */

struct lexc_data {
  void (SPCDECL *destructor)(void *);
  void (SPCDECL *daemon)(HIDDEN_PROTO
			 void *,SP_globref,TAGGED); /* (handle,attr_ref,global) */
#if MULTI_SP_AWARE
  SPEnv *spenv;
#endif /* MULTI_SP_AWARE */
  SP_globref refbase;
  long stamp;			/* increases up to backtracking */
  int flag;
  int ntuples;
  int nvars;
  Dvar xvar;
  long *lb;
  long *ub;
  int *state;			/* 0x1 = tuple[i] OP tuple[i+1] entailed */
  int *stack;
  int *limit;			/* cached value of compute_limit */
  int top;
};

#define RefCAttr(T,V) (pdata->refbase + ((nvars*(T) + (V))<<1))
#define CVAR(T,V)     (pdata->xvar    + nvars*(T) + (V))
#define LB(T,V)       (pdata->lb[       nvars*(T) + (V)])
#define UB(T,V)       (pdata->ub[       nvars*(T) + (V)])

static void SPCDECL lexc_destructor(void *pdata_v)
{
  struct lexc_data *pdata = (struct lexc_data*)pdata_v;
  FD_SETUP_SPENV(pdata->spenv)

  SP_free_globrefs(pdata->refbase,pdata->ntuples*pdata->nvars<<1);
  SP_free(pdata);
}

enum cmpenum {
  cmp_l=1,
  cmp_e,
  cmp_le,
  cmp_g,
#if EAGER_ENTAILMENT
  cmp_lg,			/* too expensive */
#endif
  cmp_ge,
  cmp_lge
};

#define DECODE_SPEC(SPEC,L,U,SET)		\
DerefArg((SET),(SPEC),1);			\
(L) = GetSmall(SET);				\
DerefArg((SET),(SPEC),2);			\
(U) = GetSmall(SET);				\
DerefArg((SET),(SPEC),3);			\


static enum cmpenum dvar_cmp(Dvar dv1,Dvar dv2)
{
  long max_min = dvar_max_l(dv1) - dvar_min_l(dv2);
  long min_max = dvar_min_l(dv1) - dvar_max_l(dv2);

  if (dvar_is_integer(dv1) && dvar_is_integer(dv2))
    return (max_min<0 ? cmp_l : max_min>0 ? cmp_g : cmp_e);
  else if (dvar_is_alias(dv1,dv2))
    return cmp_e;
  else if (max_min<=0)		/* l,e,le */
    return (max_min<0 ? cmp_l :
	    min_max<0 ? cmp_le : cmp_e);
  else				/* g,ge,lge */
    return (min_max>0 ? cmp_g :
	    min_max==0 ? cmp_ge : cmp_lge);
}

static BOOL lex_chain_leqc MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                                  struct lexc_data *pdata,
                                  Dvar xv, Dvar yv, int c)
{
  int rc1 = dvar_fix_min_t(yv,dvar_min_t(xv)+IStep(c));
  int rc2 = dvar_fix_max_t(xv,dvar_max_t(yv)-IStep(c));

  if (rc1<0 || rc2<0)
    return FALSE;
  if (rc1+rc2>0)
    pdata->stack[pdata->top++] = xv-pdata->xvar;
  return TRUE;
}

static int lex_chain_pair MAGIC (HIDDEN_PROTO
				 struct lexc_data *pdata, int i, int mode)
{
  int ent=-1, iptr=0, jptr, kptr, c=0;
  int nvars = pdata->nvars;
  Dvar xvar = CVAR(i,0);
  Dvar yvar = CVAR(i+1,0);
  SP_globref xref = RefCAttr(i,0);
  SP_globref yref = RefCAttr(i+1,0);
  Dvar xv, yv;
  
 state1:
  if (iptr==nvars) {
    if (mode&1)
      goto ret_fail;
    else
      goto ret_succeed;
  }
  xv = xvar+iptr;
  yv = yvar+iptr;
  dvar_init(xv, xref+(iptr<<1), xref+(iptr<<1)+1);
  dvar_init(yv, yref+(iptr<<1), yref+(iptr<<1)+1);
  switch (dvar_cmp(xv,yv)) {
  case cmp_g:
    goto ret_fail;
  case cmp_ge:
    if (!lex_chain_leqc(pdata,xv,yv,0))
      goto ret_fail;
  case cmp_e:
    iptr++;
    goto state1;
#if EAGER_ENTAILMENT
  case cmp_lg:
    goto ret_xi_lt_yi;
#endif
  case cmp_l:
    goto ret_succeed;
  case cmp_lge:
    if (!lex_chain_leqc(pdata,xv,yv,0))
      goto ret_fail;
  case cmp_le:
    goto state2;
  }
 state2:
  jptr = iptr+1;
 state2b:
  if (jptr==nvars) {
    if (mode&1)
      goto ret_xi_lt_yi;
    else
      goto ret_xi_le_yi;
  }
  xv = xvar+jptr;
  yv = yvar+jptr;
  dvar_init(xv, xref+(jptr<<1), xref+(jptr<<1)+1);
  dvar_init(yv, yref+(jptr<<1), yref+(jptr<<1)+1);
  switch (dvar_cmp(xv,yv)) {
  case cmp_g:
    goto ret_xi_lt_yi;
  case cmp_l:
    goto ret_xi_le_yi;
  case cmp_e:
    jptr++;
    goto state2b;
  case cmp_le:
    goto state3;
  case cmp_ge:
    goto state4;
  default: /* cmp_lg, cmp_lge */;
    mode = (mode&0x1) + 2;
    ent = 0;
    goto ret;
  }
 state3:
  kptr = jptr+1;
 state3b:
  if (kptr==nvars) {
    if (mode&1)
      goto delay3;
    else
      goto ret_xi_le_yi;
  }
  xv = xvar+kptr;
  yv = yvar+kptr;
  dvar_init(xv, xref+(kptr<<1), xref+(kptr<<1)+1);
  dvar_init(yv, yref+(kptr<<1), yref+(kptr<<1)+1);
  switch (dvar_cmp(xv,yv)) {
  case cmp_l:
    goto ret_xi_le_yi;
  case cmp_e:
  case cmp_le:
    kptr++;
    goto state3b;
  default: /* cmp_g, cmp_ge, cmp_lg, cmp_lge */;
  delay3:
    mode = (mode&0x1) + 4;
    ent = 0;
    goto ret;
  }
 state4:
  kptr = jptr+1;
 state4b:
  if (kptr==nvars) {
    if (mode&1)
      goto ret_xi_lt_yi;
    else
      goto delay4;
  }
  xv = xvar+kptr;
  yv = yvar+kptr;
  dvar_init(xv, xref+(kptr<<1), xref+(kptr<<1)+1);
  dvar_init(yv, yref+(kptr<<1), yref+(kptr<<1)+1);
  switch (dvar_cmp(xv,yv)) {
  case cmp_g:
    goto ret_xi_lt_yi;
  case cmp_e:
  case cmp_ge:
    kptr++;
    goto state4b;
  default: /* cmp_l, cmp_le, cmp_lg, cmp_lge */;
  delay4:
    mode = (mode&0x1) + 6;
    ent = 0;
    goto ret;
  }
 ret_xi_lt_yi:
  c = -1;
 ret_xi_le_yi:
  xv = xvar+iptr;
  yv = yvar+iptr;
  if (!lex_chain_leqc(pdata,xv,yv,c))
    goto ret_fail;
  if (dvar_max_l(xv) > dvar_min_l(yv)+c)
    dvar_export_leqc(xv,yv,c,3);
 ret_succeed:
  ent = 1;
 ret:
 ret_fail:
  return ent;
}

static BOOL need_bounds(struct lexc_data *pdata, int i)
{
  if (pdata->flag&2)
    return TRUE;
  if (pdata->state[i]>0)
    return (i>=2 && pdata->state[i-2]==0 && pdata->state[i-1]==0);
  else
    return ((i+1<pdata->ntuples && pdata->state[i+1]==0) ||
	    (i>=1 && pdata->state[i-1]==0));
}

static BOOL
repair_atleast MAGIC (HIDDEN_PROTO
		      struct lexc_data *pdata,
		      int l, int u,
		      int countval,
		      TAGGED valueset,
		      BOOL ubp,
		      long *vec,
		      Dvar var)
{
  int j, countvar=0, alpha, slack, nforced;
  int nvars = pdata->nvars;
      
  for (j=nvars-1; j>=0; j--) {
    TAGGED elt = MakeSmall(vec[j]);
    int dom_intersects =
      (dvar_compare_set(var+j,valueset) != FDI_DISJOINT);
    
    if (fd_member(elt,valueset))
      countval--;
    if (dom_intersects)
      countvar++;
    if (countval+countvar>=l &&
	(ubp ? Tgt(elt,dvar_min_t(var+j)) : Tlt(elt,dvar_max_t(var+j)))) {
      if (countval+countvar==l && dom_intersects) {
	if (fd_compare((ubp ? fd_and_interval(dvar_set(var+j),Inf,elt-IStep(1))
			    : fd_and_interval(dvar_set(var+j),elt+IStep(1),Sup)),
		       valueset) != FDI_DISJOINT)
	  break;
      } else
	break;
    }
  }
  if (j<0)
    return FALSE;
  alpha = j;
  slack = countval+countvar-l;
  nforced = 0;
  for (j=nvars-1; j>=alpha; j--) {
    pdata->stack[j] = nforced;
    switch (dvar_compare_set(var+j,valueset)) {
    case FDI_EQUAL:
    case FDI_SUBSET:
      nforced++;
    }
  }
  for (j=alpha; j<nvars; j++) {
    TAGGED elt = MakeSmall(vec[j]);
    int dom_intersects =
      (dvar_compare_set(var+j,valueset) != FDI_DISJOINT);
    
    if (j==alpha) {
      if (slack==0 && dom_intersects)
	elt = ubp ? fd_and_max(fd_and_interval(dvar_set(var+j),Inf,elt-IStep(1)),
			       valueset)
	          : fd_and_min(fd_and_interval(dvar_set(var+j),elt+IStep(1),Sup),
			       valueset);
      else
	elt = ubp ? dvar_predecessor_t(var+j,elt)
	          : dvar_successor_t(var+j,elt);
    } else {
      if (slack==0 && dom_intersects)
	elt = ubp ? fd_and_max(dvar_set(var+j),valueset)
	          : fd_and_min(dvar_set(var+j),valueset);
      else if (countval+pdata->stack[j]==u)
	/* TODO: fd_subtract_{max,min} */
	elt = ubp ? fd_and_max(dvar_set(var+j),fd_complement(valueset))
	          : fd_and_min(dvar_set(var+j),fd_complement(valueset));
      else
	elt = ubp ? dvar_max_t(var+j)
	          : dvar_min_t(var+j);
    }
    vec[j] = GetSmall(elt);
    if (fd_member(elt,valueset))
      countval++;
    else if (dom_intersects)
      slack--;
  }
  return TRUE;
}

static BOOL
repair_among MAGIC (HIDDEN_PROTO
		    struct lexc_data *pdata,
		    TAGGED spec,
		    BOOL ubp,
		    long *vec,
		    Dvar var)
{
  int l, u, j, countval=0;
  int nvars = pdata->nvars;
  TAGGED valueset;

  DECODE_SPEC(spec,l,u,valueset);
  for (j=0; j<nvars; j++) {
    TAGGED elt = MakeSmall(vec[j]);
    if (fd_member(elt,valueset))
      countval++;
  }
  if (countval<l)
    return repair_atleast(pdata,l,u,countval,valueset,ubp,vec,var);
  else if (countval>u)
    return repair_atleast(pdata,nvars-u,nvars-l,nvars-countval,
			  fd_complement(valueset),ubp,vec,var);
  else
    return TRUE;
}

/* Compute d as the lexicographically smallest vector d such that
   lex_le(c,d) and d[i] in dom(X[i]) for all i and among(...) holds.
   If lt!=0, lex_lt(c,d) must hold.
   Returns -1 for failure, 0 for no change, 1 for change.
*/
static int
adjust_vector_up MAGIC (HIDDEN_PROTO
			struct lexc_data *pdata,
			int lt,
			TAGGED spec,
			long *c,
			long *d,
			Dvar var)
{
  int j, alpha = -1;
  int nvars = pdata->nvars;
  long prevlb;
      
  /* compute alpha = msp that we have to increment */
  j = 0;
  prevlb = c[j];
  while (j<nvars && dvar_contains_value_l(var+j,prevlb)) {
    if (prevlb < dvar_max_l(var+j))
      alpha = j;
    j++;
    prevlb = c[j];
  }
  if ((j<nvars && prevlb < dvar_max_l(var+j)) || (j==nvars && !lt))
    alpha = j;
  if (alpha<0)
    return -1;
  for (j=0; j<nvars; j++)
    d[j] = (j<alpha ? c[j] :
	    j>alpha ? dvar_min_l(var+j) :
	    dvar_successor_l(var+j,c[j]));
  if (spec!=atom_nil &&
      !repair_among(pdata,spec,FALSE,d,var))
    return -1;
  return (alpha<nvars);
}

/* Compute d as the lexicographically greatest vector d such that
   lex_le(d,c) and d[i] in dom(X[i]) for all i and among(...) holds.
   If lt!=0, lex_lt(d,c) must hold.
   Returns -1 for failure, 0 for no change, 1 for change.
*/
static int
adjust_vector_down MAGIC (HIDDEN_PROTO
			  struct lexc_data *pdata,
			  int lt,
			  TAGGED spec,
			  long *c,
			  long *d,
			  Dvar var)
{
  int j, alpha = -1;
  int nvars = pdata->nvars;
  long prevub;
      
  /* compute alpha = msp that we have to increment */
  j = 0;
  prevub = c[j];
  while (j<nvars && dvar_contains_value_l(var+j,prevub)) {
    if (prevub > dvar_min_l(var+j))
      alpha = j;
    j++;
    prevub = c[j];
  }
  if ((j<nvars && prevub > dvar_min_l(var+j)) || (j==nvars && !lt))
    alpha = j;
  if (alpha<0)
    return -1;
  for (j=0; j<nvars; j++)
    d[j] = (j<alpha ? c[j] :
	    j>alpha ? dvar_max_l(var+j) :
	    dvar_predecessor_l(var+j,c[j]));
  if (spec!=atom_nil &&
      !repair_among(pdata,spec,TRUE,d,var))
    return -1;
  return (alpha<nvars);
}

/* Compute largest limit such that either X[0,limit) = LB[0,limit) or
   X[0,limit) = UB[0,limit) must hold.
*/
static int 
compute_limit MAGIC (HIDDEN_PROTO
		     struct lexc_data *pdata,
		     long *lbrow,
		     long *ubrow,
		     int inf,
		     TAGGED spec,
		     Dvar var)
{
  int nvars = pdata->nvars;
  int sup = nvars;
  int j, mid;
  long *c;

  if (spec==atom_nil) {
    if (inf==nvars || dvar_successor_l(var+inf,lbrow[inf])<ubrow[inf])
      return inf;
    inf++;
    while (inf<nvars &&
	   dvar_min_l(var+inf)==ubrow[inf] &&
	   dvar_max_l(var+inf)==lbrow[inf])
      inf++;
  } else {			/* among(L,U,Set) */
    c = Malloc(nvars,long);
    while (inf<sup) {
      mid = (inf+sup)>>1;
      for (j=0; j<=mid; j++)
	c[j] = lbrow[j];
      c[mid]++;			/* increment last pos. by one */
      for (j=mid+1; j<nvars; j++)
	c[j] = dvar_min_l(var+j);
      adjust_vector_up(pdata,FALSE,spec,c,c,var);
      for (j=0; j<=mid; j++)
	if (c[j]!=ubrow[j])
	  break;
      if (j>mid)
	inf = mid+1;
      else
	sup = mid;
    }
    Free(c);
  }
  return inf;
}

#define CASE_MAYBE_GE 0
#define CASE_MAYBE_LE 1
#define CASE_MAYBE_GT 2
#define CASE_MAYBE_LT 3
#define CASE_MAYBE_ANY 4
#define CASE_INOUT_GE 5
#define CASE_INOUT_LE 6
#define CASE_INOUT_ANY 7

/* Case analysis for Prop. 1 & 2 */
static int
among_case MAGIC (HIDDEN_PROTO
		  int i,
		  long *bound,
		  BOOL ubp,
		  TAGGED valueset,
		  Dvar var)
{
  long min_of_intersection;
  long max_of_intersection;
  
  if (ubp) {
    switch (dvar_compare_set(var+i,valueset)) {
    case FDI_DISJOINT:
    case FDI_EQUAL:
    case FDI_SUBSET:
      if (dvar_min_l(var+i)==bound[i])
	return CASE_INOUT_GE;
      else
	return CASE_INOUT_ANY;
    }
    min_of_intersection = GetSmall(fd_and_min(dvar_set(var+i),valueset));
    if (min_of_intersection==bound[i])
      return CASE_MAYBE_GE;
    else if (min_of_intersection>bound[i])
      return CASE_MAYBE_GT;
    else
      return CASE_MAYBE_ANY;
  } else {
    switch (dvar_compare_set(var+i,valueset)) {
    case FDI_DISJOINT:
    case FDI_EQUAL:
    case FDI_SUBSET:
      if (dvar_max_l(var+i)==bound[i])
	return CASE_INOUT_LE;
      else
	return CASE_INOUT_ANY;
    }
    max_of_intersection = GetSmall(fd_and_max(dvar_set(var+i),valueset));
    if (max_of_intersection==bound[i])
      return CASE_MAYBE_LE;
    else if (max_of_intersection<bound[i])
      return CASE_MAYBE_LT;
    else
      return CASE_MAYBE_ANY;
  }
}

static BOOL
prune_among_force MAGIC (HIDDEN_PROTO
			 struct lexc_data *pdata,
			 int limit,
			 TAGGED valueset,
			 Dvar var)
{
  int nvars = pdata->nvars;
  BOOL pruned = FALSE;
  int j;

  for (j=limit; j<nvars; j++) {
    switch (dvar_compare_set(var+j,valueset)) {
    case FDI_DISJOINT:
    case FDI_EQUAL:
    case FDI_SUBSET:
      break;
    default:
      dvar_fix_set(var+j,valueset);
      pruned = TRUE;
    }
  }
  return pruned;
}

static void
prune_among_det MAGIC (HIDDEN_PROTO
		       struct lexc_data *pdata,
		       long *lbrow,
		       long *ubrow,
		       int limit, /* first nonground */
		       TAGGED spec,
		       Dvar var)
{
  int least, most, j;
  TAGGED valueset, complement;
  int nvars = pdata->nvars;
  int nb_in[2];
  int nb_out[2];
  int can_force_in, can_force_out;
  /* [0] - counts dom(X[i]) vs. valueset, i<limit
   * [1] - counts dom(X[i]) vs. valueset, i>=limit
   */

  DECODE_SPEC(spec,least,most,valueset);
  complement = fd_complement(valueset);
  nb_in[0] = nb_out[0] = 0;

  for (j=0; j<limit; j++) {
    int lbin = fd_member(MakeSmall(lbrow[j]),valueset);
    
    nb_in[0] += lbin;
    nb_out[0] += 1-lbin;
  }
 restart:
  nb_in[1] = nb_out[1] = 0;
  for (j=limit; j<nvars; j++) {
    switch (dvar_compare_set(var+j,valueset)) {
    case FDI_EQUAL:
    case FDI_SUBSET:
      nb_in[1]++;
      break;
    case FDI_DISJOINT:
      nb_out[1]++;
      break;
    }
  }

  can_force_in = (nb_out[0]+nb_out[1] == nvars-least-1);
  can_force_out = (nb_in[0]+nb_in[1] == most-1);
  if (nb_in[1]+nb_out[1] == nvars-limit) { /* no further pruning possible */
  } else if (nb_out[0]+nb_out[1] == nvars-least) {
    /* force all "maybe" to be "in" */
    prune_among_force(pdata,limit,valueset,var);
  } else if (nb_in[0]+nb_in[1] == most) {
    /* force all "maybe" to be "out" */
    prune_among_force(pdata,limit,complement,var);
  } else if (can_force_in||can_force_out) {
    int l, rule;

    /* Proposition 1&2&3, four variants */
    for (rule=0; rule<4; rule++) {
      long *bound = (rule&2) ? lbrow : ubrow;
      TAGGED set =  (rule&1) ? complement : valueset;
      BOOL cond =   (rule&1) ? can_force_out : can_force_in;
      int gtpos = -1;
      int token = CASE_MAYBE_GE; /* [PM] 3.10 shut up compiler, init value will not be used */
      
      if (cond) {
	for (l=limit; l+1<nvars; l++) {
	  token = among_case(l,bound,(rule<2),set,var);
	  switch (token) {
	  case CASE_MAYBE_GE:
	  case CASE_MAYBE_LE:
	  case CASE_INOUT_GE:
	  case CASE_INOUT_LE:
	    continue;
	  case CASE_MAYBE_GT:
	  case CASE_MAYBE_LT:
	    gtpos = l;
	  }
	  break;
	}
	if (l+1<nvars && (token==CASE_MAYBE_GT || token==CASE_MAYBE_LT)) {
	  if (prune_among_force(pdata,l+1,set,var))
	    goto restart;
	}
	for (j=limit; j<=l; j++) {
	  TAGGED feasible;
	  
	  switch (dvar_compare_set(var+j,set)) {
	  case FDI_DISJOINT:
	  case FDI_EQUAL:
	  case FDI_SUBSET:
	    continue;
	  }
	  feasible =
	    (rule&2) ? fd_or_interval(set,MakeSmall(lbrow[j]+(j<gtpos)),Sup) :
	               fd_or_interval(set,Inf,MakeSmall(ubrow[j]-(j<gtpos)));
	  switch (dvar_compare_set(var+j,feasible)) {
	  case FDI_EQUAL:
	  case FDI_SUBSET:
	    continue;
	  }
	  dvar_fix_set(var+j,feasible);
	  goto restart;
	}
      }
    }
    /* Proposition 4, two variants */
    if (limit+1>=nvars || lbrow[limit+1]<=ubrow[limit+1]+1)
      goto prop5;
    switch (dvar_compare_set(var+limit,valueset)) {
    case FDI_DISJOINT:
    case FDI_EQUAL:
    case FDI_SUBSET:
      goto prop5;
    }
    switch (dvar_compare_set(var+limit+1,valueset)) {
    case FDI_DISJOINT:
    case FDI_EQUAL:
    case FDI_SUBSET:
      goto prop5;
    }
    for (rule=0; rule<2; rule++) {
      TAGGED set =  (rule&1) ? complement : valueset;
      TAGGED cset = (rule&1) ? valueset : complement;
      BOOL cond =   (rule&1) ? can_force_out : can_force_in;
      TAGGED subset;

      if (cond) {
	subset = fd_subtract(dvar_set(var+limit),
			     fd_pair(MakeSmall(lbrow[limit]),
				     MakeSmall(ubrow[limit])));
	if (fd_compare(subset,set)==FDI_DISJOINT &&
	    dvar_prune_set(var+limit+1,
			   fd_and_interval(cset,MakeSmall(ubrow[limit+1]+1),
					   MakeSmall(lbrow[limit+1]-1)))!=0)
	  goto restart;
      }
    }
  prop5:
    /* if forcing all but one maybes to IN (OUT) causes contradiction,
       then at least one of the maybes must be OUT (IN) */
    for (rule=0; rule<2; rule++) {
      TAGGED set =  (rule&1) ? complement : valueset;
      BOOL cond =   (rule&1) ? can_force_out : can_force_in;

      if (cond) {
	Dvar var2 = Malloc(nvars,struct dvar);
	long *bound = Malloc(nvars,long);
	int maybes = nvars-nb_out[0]-nb_out[1]-nb_in[0]-nb_in[1];
	int pos = 0;
        BOOL force = FALSE;

	for (j=0; j<limit; j++) {
	  TAGGED t = MakeSmall(lbrow[j]);
	  
	  (var2+j)->min = t;
	  (var2+j)->max = t;
	  (var2+j)->set = fd_pair(t,t);
	  (var2+j)->flags = DV_PRUNED_VAL|DV_INTERVAL|DV_SET_OK;
	}
	for (j=limit; j<nvars; j++) {
	  dvar_assign(var2+j,var+j);
	  switch (dvar_compare_set(var2+j,set)) {
	  case FDI_DISJOINT:
	  case FDI_EQUAL:
	  case FDI_SUBSET:
	    continue;
	  }
	  if (maybes>1)
	    dvar_fix_set(var2+j,set);
	  else
	    pos = j;
	  maybes--;
	}
	if (adjust_vector_up(pdata,0,atom_nil,lbrow,bound,var2)<0)
	  force = TRUE;
	for (j=limit; j<nvars && !force; j++) {
	  if (bound[j]<ubrow[j])
	    break;
	  else if (bound[j]>ubrow[j])
	    force = TRUE;
	}
	Free(var2);
	Free(bound);
	if (force) {
	  dvar_fix_set(var+pos,set);
	  goto restart;
	}
      }
    }
  }
}

/* Filtering for among(L,U,Set) applied to row Row.
   We also know that either X[0,limit) = LB[0,limit)
   or X[0,limit) = UB[0,limit) holds.
*/
static void
prune_among MAGIC (HIDDEN_PROTO
		   struct lexc_data *pdata,
		   long *lbrow,
		   long *ubrow,
		   int limit,
		   TAGGED spec,
		   Dvar var)
{
  int j;
  int l;			/* first nonground pos. */
  int nvars = pdata->nvars;

  for (l=0; l<nvars; l++)
    if (!dvar_is_integer(var+l))
      break;
  if (limit<=l) {
    prune_among_det(pdata,lbrow,ubrow,l,spec,var);
  } else if (limit<nvars) {
    Dvar var_a = Malloc(2*nvars,struct dvar);
    Dvar var_b = var_a+nvars;
    long *bound = Malloc(nvars,long);
    int l;

    for (j=0; j<limit; j++) {
      dvar_assign(var_a+j,var+j);
      dvar_fix_value_l(var_a+j,lbrow[j]);
      bound[j] = lbrow[j];
    }
    l = limit;
    for (j=limit; j<nvars; j++) {
      dvar_assign(var_a+j,var+j);
      if (j==l) {
	dvar_fix_interval_t(var_a+j,MakeSmall(lbrow[j]),Sup);
	if (dvar_is_integer(var_a+j))
	  l++;
      }
      bound[j] = dvar_max_l(var+j);
    }
    repair_among(pdata,spec,TRUE,bound,var_a);
    prune_among_det(pdata,lbrow,bound,l,spec,var_a);
    for (j=0; j<limit; j++) {
      dvar_assign(var_b+j,var+j);
      dvar_fix_value_l(var_b+j,lbrow[j]);
      bound[j] = ubrow[j];
    }
    l = limit;
    for (j=limit; j<nvars; j++) {
      dvar_assign(var_b+j,var+j);
      if (j==l) {
	dvar_fix_interval_t(var_b+j,Inf,MakeSmall(ubrow[j]));
	if (dvar_is_integer(var_b+j))
	  l++;
      }
      bound[j] = dvar_min_l(var+j);
    }
    repair_among(pdata,spec,FALSE,bound,var_b);
    prune_among_det(pdata,bound,ubrow,l,spec,var_b);
    for (j=limit; j<nvars; j++)
      dvar_fix_set(var+j,fd_or(dvar_set(var_a+j),dvar_set(var_b+j)));
    Free(var_a);
    Free(bound);
  }
}


/* Require: lbrow[*] and ubrow[*] are _feasible_ bounds. */
static void
prune_between MAGIC (HIDDEN_PROTO
		     struct lexc_data *pdata,
		     long *lbrow,
		     long *ubrow,
		     TAGGED spec,
		     int row,
		     Dvar var)
{
  int nvars   = pdata->nvars;
  int j;

  for (j=0; j<nvars && lbrow[j]==ubrow[j]; j++)
    dvar_fix_value_l(var+j,lbrow[j]);
  if (j<nvars)
    dvar_fix_interval_l(var+j,lbrow[j],ubrow[j]);
  pdata->limit[row] = compute_limit(pdata,lbrow,ubrow,j,spec,var);
  for (; j<pdata->limit[row]; j++)
    dvar_fix_set(var+j,fd_pair(MakeSmall(lbrow[j]),MakeSmall(ubrow[j])));
  if (j<nvars && ubrow[j]+1<=lbrow[j]-1)
    dvar_prune_interval_l(var+j,ubrow[j]+1,lbrow[j]-1);
  if (pdata->flag&2)
    prune_among(pdata,lbrow,ubrow,pdata->limit[row],spec,var);
}

#if 0
static BOOL
entailed_between MAGIC (HIDDEN_PROTO
			struct lexc_data *pdata,
			long *lbrow,
			long *ubrow,
			TAGGED spec,
			Dvar xv)
{
  int nvars   = pdata->nvars;
  int j;

  for (j=0; j<nvars; j++) {
    long xmin = dvar_min_l(xv+j);

    if (lbrow[j]<xmin)
      break;
    else if (lbrow[j]>xmin)
      return FALSE;
  }
  for (j=0; j<nvars; j++) {
    long xmax = dvar_max_l(xv+j);

    if (ubrow[j]>xmax)
      break;
    else if (ubrow[j]<xmax)
      return FALSE;
  }
  if (spec!=atom_nil) {
    int nb_in=0;
    int nb_out=0;
    int least, most;
    TAGGED valueset;

    DECODE_SPEC(spec,least,most,valueset);
    for (j=0; j<nvars; j++)
      switch (dvar_compare_set(xv+j,valueset)) {
      case FDI_EQUAL:
      case FDI_SUBSET:
	nb_in++;
	break;
      case FDI_DISJOINT:
	nb_out++;
	break;
      }

    if (nb_in<least || nvars-nb_out>most)
      return FALSE;
  }
  return TRUE;
}
#endif

static BOOL
entailed_among MAGIC (HIDDEN_PROTO
		      struct lexc_data *pdata,
		      TAGGED spec,
		      Dvar xv)
{
  int nvars   = pdata->nvars;
  int j;
  int nb_in=0;
  int nb_out=0;
  int least, most;
  TAGGED valueset;

  DECODE_SPEC(spec,least,most,valueset);
  for (j=0; j<nvars; j++)
    switch (dvar_compare_set(xv+j,valueset)) {
    case FDI_EQUAL:
    case FDI_SUBSET:
      nb_in++;
      break;
    case FDI_DISJOINT:
      nb_out++;
      break;
    }

  return (nb_in>=least && nvars-nb_out<=most);
}

static BOOL
prunable_among MAGIC (HIDDEN_PROTO
		      struct lexc_data *pdata,
		      int limit,
		      TAGGED spec,
		      Dvar var)
{
  int least, most, j, l;
  TAGGED valueset;
  int nvars = pdata->nvars;
  int nb_in=0, nb_out=0;

  for (l=0; l<nvars; l++)
    if (!dvar_is_integer(var+l))
      break;
  if (limit>l)
    return TRUE;
  DECODE_SPEC(spec,least,most,valueset);
  for (j=limit; j<nvars; j++) {
    switch (dvar_compare_set(var+j,valueset)) {
    case FDI_EQUAL:
    case FDI_SUBSET:
      nb_in++;
      break;
    case FDI_DISJOINT:
      nb_out++;
      break;
    }
  }
  return (nb_out>=nvars-least-1 || nb_in>=most-1);
}

static void SPCDECL 
lexc_daemon MAGIC (HIDDEN_PROTO
		   void *vdata,
		   SP_globref attr_ref,
		   TAGGED global)
{
  struct lexc_data *pdata = (struct lexc_data *)vdata;
  TAGGED tstate;
  int ar, state_stamp;
  WAMENV;

  if (RefMutable(CTagToArg(global,3)) & IStep(1)) /* already enqueued? */
    return;
  tstate = RefMutable(CTagToArg(global,1));
  ar = Arity(TagToHeadfunctor(tstate));
  state_stamp = GetSmall(CTagToArg(tstate,ar));
  if (pdata->stamp==state_stamp) { /* incremental */
    /* valid: state, lb, ub */
    int ntuples = pdata->ntuples;
    int nvars = pdata->nvars;
    int refoffset = (attr_ref-pdata->refbase)>>1;
    int row = refoffset/nvars;
    int elt = refoffset%nvars;
    TAGGED among_spec = CTagToArg(tstate,5);
    int j;

    /* no incremental handling of lex_chain_pair yet */

    if (need_bounds(pdata,row)) {
      Dvar var = CVAR(row,0);
      SP_globref ref = RefCAttr(row,0);
      
      dvar_init(var+elt, attr_ref, attr_ref+1);
      if (!dvar_contains_value_l(var+elt,LB(row,elt)) ||
	  !dvar_contains_value_l(var+elt,UB(row,elt)))
	goto push;
      if (pdata->flag&2) {
	for (j=0; j<nvars; j++) {
	  if (j!=elt)
	    dvar_init(var+j, ref, ref+1);
	  ref += 2;
	}
	if (prunable_among(pdata,pdata->limit[row],among_spec,var))
	  goto push;
      }
    } else if ((row<ntuples-1 &&
		pdata->state[row]==0 && pdata->state[row+1]>0) ||
	       (row>0 &&
		pdata->state[row-1]==0 && pdata->state[row]>0))
      goto push;
    return;
  }
 push:
  fd_enqueue_global(w, global, 0x7/* DOM, append*/);   
}

/*
  '$fd_lex_chain'(+State0, +State, -Actions).
  State = state(Tuples,NT,NV,Flag,Among,EntFlags,Handle,Stamp)
  |EntFlags| = NT
  EntFlags[i]==1 iff Tuples[i] OP Tuples[i+1] is entailed.
  If EntFlags[i]==1, Tuples[i] is _separated_ from Tuples[i+1].
  (Flag & 0x1) means lexicographic (strict) less than.
  (Flag & 0x2), Among=among(L,U,Vals) means each vector must take
  at least L and at most U values out of Vals.

  We handle three cases:
  1) Tuple i is separated from both i-1 and i+1.
     Then tuple i is ignored altogether.
  2) Tuples i and i+1 are not separated, but are separated from i-1 and i+2.
     Then tuples i and i+1 are handled by the tuple-pair algorithm.
  3) Groups of 3 or more nonseparated tuples are handled by:
     a) compute feasible upper bounds
     b) compute feasible lower bounds
     c) prune wrt. both bounds
*/
void SPCDECL
prolog_fd_lex_chain MAGIC (HIDDEN_PROTO
			   SP_term_ref State0,
			   SP_term_ref State,
			   SP_term_ref Actions)
{
  int ent = -1;			/* initially disentailed */
  TAGGED handle, t1, telt, trow, tmat;
  int ntuples, nvars, nelts, i, j, flag;
  struct lexc_data *pdata;
  BOOL committed, post;
  long state_stamp;
  char *ptr;
  TAGGED among_spec;
  WAMENV;

/*    X(0) = RefTerm(State0); */
  (void)State0;                 /* [PM] 3.9b5 avoid -Wunused */
  w->numstack_end = NULL;
  dvar_export_start(3);
  RefTerm(State) = unify_output_state(w,&handle,&state_stamp,&committed);
  if (!IsVar(handle)) {		/* got [Flag | '$free'(Ptr)] */
    post = FALSE;
    pdata = Pdata(struct lexc_data,handle);
    ntuples = pdata->ntuples;
    nvars = pdata->nvars;
    nelts = ntuples*nvars;
  } else {			/* build persistent state */
    post = TRUE;
    DerefArg(telt,X(0),2);	/* get Ntuples */
    ntuples = GetSmall(telt);
    DerefArg(telt,X(0),3);	/* get Nvars */
    nvars = GetSmall(telt);
    nelts = ntuples*nvars;
    pdata = Palloc(struct lexc_data,
		   nelts*(sizeof(struct dvar)+2*sizeof(long)+sizeof(int)) +
		   2*ntuples*sizeof(int),
		   handle);
    pdata->destructor = lexc_destructor;
    pdata->daemon = lexc_daemon;
    FD_STORE_SPENV(pdata->spenv);
    pdata->refbase = SP_alloc_globrefs(nelts<<1);
    pdata->stamp = state_stamp-1; /* DON'T trust initially */
    DerefArg(telt,X(0),4);	/* get Flag */
    pdata->flag = GetSmall(telt);
    pdata->ntuples = ntuples;
    pdata->nvars = nvars;
    ptr = (char *)(pdata+1);
    pdata->xvar = (Dvar)ptr;
    ptr += nelts*sizeof(struct dvar);
    pdata->lb = (long *)ptr;
    ptr += nelts*sizeof(long);
    pdata->ub = (long *)ptr;
    ptr += nelts*sizeof(long);
    pdata->stack = (int *)ptr;
    ptr += nelts*sizeof(int);
    pdata->state = (int *)ptr;
    ptr += ntuples*sizeof(int);
    pdata->limit = (int *)ptr;
    ptr += ntuples*sizeof(int);
    {
      SP_globref ref = pdata->refbase;
      
      DerefArg(tmat,X(0),1);	/* get X */
      for (i=0; i<ntuples; i++) {
	DerefCar(trow,tmat);
	DerefCdr(tmat,tmat);
	for (j=0; j<nvars; j++) {
	  DerefCar(telt,trow);
	  DerefCdr(trow,trow);
	  DerefArg(t1,telt,1);	/* get domain var */
	  RefGlob(ref+1) = t1;
	  DerefArg(t1,telt,2);	/* get attribute */
	  RefGlob(ref) = t1;
	  ref += 2;
	}
      }
    }
  }

  /* RESUME HERE */

  flag = pdata->flag;
  pdata->top = 0;
  pdata->stamp = state_stamp+1;
  DerefArg(trow,X(0),6);	/* get EntFlags */
  for (i=0; i<ntuples-1; i++) {
    DerefCar(telt,trow);
    DerefCdr(trow,trow);
    pdata->state[i] = !IsVar(telt);
  }
  pdata->state[i] = 1;
  among_spec = CTagToArg(X(0),5);

  for (i=0; i<ntuples; i++)
    if (need_bounds(pdata,i)) {
      SP_globref ref = RefCAttr(i,0);
      Dvar xv = CVAR(i,0);
      for (j=0; j<nvars; j++) {
	dvar_init(xv+j, ref, ref+1);
	ref += 2;
      }
    }

  /* compute all upper bounds */
  
  for (i=ntuples-1; i>=0; i--) {
    if (!need_bounds(pdata,i)) {
    } else if (pdata->state[i]>0) {
      for (j=0; j<nvars; j++)
	UB(i,j) = dvar_max_l(CVAR(i,j));
      if ((flag&2) &&
	  !repair_among(pdata,among_spec,TRUE,&UB(i,0),CVAR(i,0)))
	goto ret_fail;
    } else if (adjust_vector_down(pdata,(flag&1),among_spec,&UB(i+1,0),&UB(i,0),CVAR(i,0))<0)
      goto ret_fail;
  }

  /* compute all lower bounds */
  
  for (i=0; i<ntuples; i++) {
    if (!need_bounds(pdata,i)) {
    } else if (i==0 || pdata->state[i-1]>0) {
      for (j=0; j<nvars; j++)
	LB(i,j) = dvar_min_l(CVAR(i,j));
      if ((flag&2) && !repair_among(pdata,among_spec,FALSE,&LB(i,0),CVAR(i,0)))
	goto ret_fail;
    } else if (adjust_vector_up(pdata,(flag&1),among_spec,&LB(i-1,0),&LB(i,0),CVAR(i,0))<0)
      goto ret_fail;
  }
  
  /* prune all tuples wrt. lower and upper bounds */

  for (i=0; i<ntuples; i++)
    if (need_bounds(pdata,i)) {
      Dvar xvar = pdata->xvar + i*nvars;
      long *lbrow = pdata->lb + i*nvars;
      long *ubrow = pdata->ub + i*nvars;
      int l=nvars;		/* first nonground pos. */

      prune_between(pdata,lbrow,ubrow,among_spec,i,xvar);
      for (j=0; j<nvars; j++) {
	if (l==nvars && !dvar_is_integer(xvar+j))
	  l = j;
      }
#if 0
      /* doesn't seem to ever contribute any pruning */
      if (l<nvars &&
	  (flag&2) &&
	  !entailed_between(pdata,lbrow,ubrow,among_spec,xvar)) {
	/* constructive disjunction based on most significant pos.
	   after ground prefix */
	Dvar cd_var = Malloc(3*nvars,struct dvar);
	long *vec1 = Malloc(2*nvars,long);
	long *vec2 = vec1+nvars;

	for (j=0; j<nvars; j++) {
	  dvar_assign(cd_var+j,xvar+j);
	  dvar_assign(cd_var+nvars+j,xvar+j);
	  if (l>=r)
	    dvar_assign(cd_var+2*nvars+j,xvar+j);
	}

	/* branch 1: assume x[l]=a[l] */
	for (j=0; j<nvars; j++)
	  vec1[j] = j<=l ? lbrow[j] : dvar_max_l(xvar+j);
	adjust_vector_down(pdata,FALSE,among_spec,vec1,vec1,cd_var);
	prune_between(pdata,lbrow,vec1,among_spec,i,cd_var);

	/* branch 2: assume x[l]=b[l] */
	for (j=0; j<nvars; j++)
	  vec1[j] = j<=l ? ubrow[j] : dvar_min_l(xvar+j);
	adjust_vector_up(pdata,FALSE,among_spec,vec1,vec1,cd_var+nvars);
	prune_between(pdata,vec1,ubrow,among_spec,i,cd_var+nvars);

	if (l>=r) {
	  /* branch 3: assume a[l]<x[l]<b[l] */
	  for (j=0; j<nvars; j++)
	    vec1[j] = j<l ? lbrow[j] :
	              j==l ? lbrow[j]+1 :
	              dvar_min_l(xvar+j);
	  adjust_vector_up(pdata,FALSE,among_spec,vec1,vec1,cd_var+2*nvars);
	  for (j=0; j<nvars; j++)
	    vec2[j] = j<l ? ubrow[j] :
	              j==l ? ubrow[j]-1 :
	              dvar_max_l(xvar+j);
	  adjust_vector_down(pdata,FALSE,among_spec,vec2,vec2,cd_var+2*nvars);
	  prune_between(pdata,vec1,vec2,among_spec,i,cd_var+2*nvars);
	}

	for (j=l; j<nvars; j++)
	  if (l<r) {
	    if (dvar_fix_set(xvar+j,
			     fd_or(dvar_set(cd_var+j),
				   dvar_set(cd_var+nvars+j)))<0)
	      break;
	  } else {
	    if (dvar_fix_set(xvar+j,
			     fd_or(dvar_set(cd_var+j),
				   fd_or(dvar_set(cd_var+nvars+j),
					 dvar_set(cd_var+2*nvars+j))))<0)
	      break;
	  }
	Free(cd_var);
	Free(vec1);
	if (j<nvars)
	  goto ret_fail;
      }
#endif
      for (j=0; j<nvars; j++)
	dvar_pruning_done(xvar+j);
      continue;
    }
    else if (pdata->state[i]==0) {
      switch (lex_chain_pair(pdata,i,flag&1)) {
      case -1:
	goto ret_fail;
      case  1:
	pdata->state[i] = 1;
      }
    }

  /* export prunings */
  
  for (i=0; i<ntuples; i++)
    if (need_bounds(pdata,i)) {
      Dvar xv = CVAR(i,0);
      for (j=0; j<nvars; j++)
	dvar_export(xv+j, 3);
    }
  i = pdata->top;
  while (i>0) {
    j = pdata->stack[--i];
    dvar_export(pdata->xvar+j, 3);
    dvar_export(pdata->xvar+nvars+j, 3);
  }

  /* update states */

  ent = 1;
  DerefArg(trow,X(0),6);	/* get Flags */
  for (i=0; i<ntuples-1; i++) {
    DerefCar(telt,trow);
    DerefCdr(trow,trow);
    if (pdata->state[i]==0) {
      for (j=0; j<nvars; j++) {
	long cmp = dvar_max_l(CVAR(i,j))-dvar_min_l(CVAR(i+1,j));
	if (cmp<0) goto sep;
	if (cmp>0) goto nosep;
      }
      if (!(flag&1)) goto sep;
    nosep:			/* tuple i can be lex gt tuple i+1  */
      ent = 0;
      continue;
    }
  sep:				/* tuple i can't be lex gt tuple i+1  */
    if (IsVar(telt)) {
      LetShadowregs;
      BindHVA(telt,TaggedOne);
    }
  }
  if (ent==1 && (flag&2)) {	/* check entailment of among as well */
    for (i=0; i<ntuples && ent==1; i++)
      ent = entailed_among(pdata,among_spec,CVAR(i,0));
  }
  if (ent==0 && post) {
    SP_globref ref = pdata->refbase;
    Dvar var = pdata->xvar;
    for (j=0; j<nelts; j++) {
      dvar_init(var+j, ref, ref+1);
      if (!dvar_is_integer(var+j))
	attach_daemon(pdata, ref, ref+1, X(1), fd.functor_dom, 3);
      ref += 2;
    }
  }

 ret_fail:
  Pfree(ent==1,handle,pdata);
  dvar_export_done(Actions, ent, 3);
}

