/* Copyright(C) 1999, Swedish Institute of Computer Science */

#include "fd.h"
#include "dvars.h"

#if MULTI_SP_AWARE
#define tell_eq(A1,A2,A3) tell_eq(HIDDEN_ARG, A1,A2,A3)
#define negdom(A1) negdom(HIDDEN_ARG, A1)
#endif /* MULTI_SP_AWARE */

static int tell_eq MAGIC (HIDDEN_PROTO
			  Dvar dvx, Dvar dvy,
			  int intersecting)
{
  TAGGED inters;
  TAGGED x = RefGlob(dvx->var_ref);
  TAGGED y = RefGlob(dvy->var_ref);
  
  switch (intersecting) {
  case FDI_DISJOINT:
    return -1;
  case FDI_EQUAL:
    if (!IsVar(x))
      return 1;
    else {
      dvar_export_equal(dvx, dvy, 3);
      return 1;
    }
  case FDI_SUBSET:
    if (!IsVar(x)) {
      dvar_fix_value_t(dvy, x);
      return 1;
    } else {
      dvar_export_equal(dvx, dvy, 3);
      return 1;
    }
  case FDI_SUPERSET:
    if (!IsVar(y)) {
      dvar_fix_value_t(dvx, y);
      return 1;
    } else {
      dvar_export_equal(dvx, dvy, 3);
      return 1;
    }
  default:
    inters = fd_and(dvar_set(dvx),dvar_set(dvy));
    if (fd_singleton(inters)) {
      TAGGED min = fd_min(inters);
      if (IsVar(x))
	dvar_fix_value_t(dvx, min);
      if (IsVar(y))
	dvar_fix_value_t(dvy, min);
      return 1;
    } else {
      dvar_export_equal(dvx, dvy, 3);
      return 1;
    }
  }
}

/* '$fd_in_set_iff'(+State, -NewState, -Actions) :- X in_set Set iff B.
   State is f(X,XMut,Set,B,BMut),
   Actions is a list of prunings etc.
*/
void SPCDECL
prolog_fd_in_set_iff MAGIC (HIDDEN_PROTO
			    SP_term_ref State,
			    SP_term_ref NewState,
			    SP_term_ref Actions)
{
  WAMENV;
  int ent = 0;		/* neither entailed nor dis- */
  int intersecting;
  int i;
  TAGGED x, set, b, t;
  SP_globref ref = SP_alloc_globrefs(4);
  struct dvar localvar[2];
  Dvar dvx = &localvar[0];
  Dvar dvb = &localvar[1];
  
  w->numstack_end = NULL;
/*    X(0) = RefTerm(State); */
  RefTerm(NewState) = RefTerm(State);
  dvar_export_start(3);
  DerefNonvar(X(0));
  DerefArg(x,X(0),1);		/* X */
  RefGlob(ref+1) = x;
  DerefArg(t,X(0),2);		/* XMut */
  RefGlob(ref+0) = t;
  DerefArg(set,X(0),3);		/* Set */
  DerefArg(b,X(0),4);		/* B */
  RefGlob(ref+3) = b;
  DerefArg(t,X(0),5);		/* BMut */
  RefGlob(ref+2) = t;
  dvar_init(dvx, ref+0, ref+1);
  dvar_init(dvb, ref+2, ref+3);
  intersecting = dvar_compare_set(dvx,set);
  if (IsVar(b)) {
    switch (intersecting) {
    case FDI_SUBSET:		/* [B=1,exit] */
    case FDI_EQUAL:		/* [B=1,exit] */
      ent = 1;
      dvar_fix_value_t(dvb, TaggedOne);
      break;
    case FDI_DISJOINT:		/* [B=0,exit] */
      ent = 1;
      dvar_fix_value_t(dvb, TaggedZero);
      break;
    }
  } else if (Teqz(b)) {
    if (IsVar(x)) {
				/* [X in_set \set,exit] */
      switch (intersecting) {
      case FDI_SUBSET:
      case FDI_EQUAL:
	ent = -1; break;
      case FDI_DISJOINT:
	ent = 1; break;
      default:
	ent = 1;
	dvar_prune_set(dvx, set);
      }
    } else if (fd_member(x,set)) {
      ent = -1;			/* [fail] */
    } else {
      ent = 1;			/* [exit] */
    }
  } else {
    if (IsVar(x)) {
				/* [X in_set set,exit] */
      switch (intersecting) {
      case FDI_SUBSET:
      case FDI_EQUAL:
	ent = 1; break;
      case FDI_DISJOINT:
	ent = -1; break;
      default:
	ent = 1;
	dvar_fix_set(dvx, set);
      }
    } else if (fd_member(x,set)) {
      ent = 1; /* [exit] */
    } else {
      ent = -1; /* [fail] */
    }
  }
  for (i=0; i<2; i++)
    dvar_pruning_done(&localvar[i]);
  for (i=0; i<2; i++)
    dvar_export(&localvar[i], 3);
  dvar_export_done(Actions, ent, 3);
  SP_free_globrefs(ref,4);
}


/* '$fd_eq_iff'(+State, -Actions) :- X#=Y iff B.
   State is f(X,XMut,Y,YMut,B,BMut),
   Actions is a list of prunings etc.
*/
void SPCDECL
prolog_fd_eq_iff MAGIC (HIDDEN_PROTO
			SP_term_ref State,
			SP_term_ref NewState,
			SP_term_ref Actions)
{
  WAMENV;
  int ent = 0;		/* neither entailed nor dis- */
  int intersecting;
  int i;
  TAGGED x, y, b, t;
  SP_globref ref = SP_alloc_globrefs(6);
  struct dvar localvar[3];
  Dvar dvx = &localvar[0];
  Dvar dvy = &localvar[1];
  Dvar dvb = &localvar[2];
  
  w->numstack_end = NULL;

/*    X(0) = RefTerm(State); */
  RefTerm(NewState) = RefTerm(State);
  dvar_export_start(3);
  DerefNonvar(X(0));
  DerefArg(x,X(0),1);		/* X */
  RefGlob(ref+1) = x;
  DerefArg(t,X(0),2);		/* XMut */
  RefGlob(ref+0) = t;
  DerefArg(y,X(0),3);		/* Y */
  RefGlob(ref+3) = y;
  DerefArg(t,X(0),4);		/* YMut */
  RefGlob(ref+2) = t;
  DerefArg(b,X(0),5);		/* B */
  RefGlob(ref+5) = b;
  DerefArg(t,X(0),6);		/* BMut */
  RefGlob(ref+4) = t;
  dvar_init(dvx, ref+0, ref+1);
  dvar_init(dvy, ref+2, ref+3);
  dvar_init(dvb, ref+4, ref+5);
  intersecting = fd_compare(dvar_set(dvx),dvar_set(dvy));
  if (IsVar(b)) {
    switch (intersecting) {
    case FDI_DISJOINT:		/* [B=0,exit] */
      ent = 1;
      dvar_fix_value_t(dvb, TaggedZero);
      break;
    case FDI_EQUAL:
      if (!IsVar(y)) {		/* [B=1,exit] */
	ent = 1;
	dvar_fix_value_t(dvb, TaggedOne);
      }
      break;
    }
  } else if (Teqz(b)) {
    if (intersecting==FDI_DISJOINT)
      ent = 1;
    else if (!IsVar(x) && !IsVar(y))
      ent = -1;
    else if (!IsVar(y)) {
      ent = 1;
      dvar_prune_value_t(dvx, y);
    } else if (!IsVar(x)) {
      ent = 1;
      dvar_prune_value_t(dvy, x);
    }
  } else      
    ent = tell_eq(dvx,dvy,intersecting);
  for (i=0; i<3; i++)
    dvar_pruning_done(&localvar[i]);
  for (i=0; i<3; i++)
    dvar_export(&localvar[i], 3);
  dvar_export_done(Actions, ent, 3);
  SP_free_globrefs(ref,6);
}


/* '$fd_le_iff'(+State, -Actions) :- X #=< Y iff B.
   State is f(X,XMut,Y,YMut,B,BMut),
   Actions is a list of prunings etc.
*/
void SPCDECL
prolog_fd_le_iff MAGIC (HIDDEN_PROTO
			SP_term_ref State,
			SP_term_ref NewState,
			SP_term_ref Actions)
{
  WAMENV;
  int ent = 0;		/* neither entailed nor dis- */
  int i;
  TAGGED x, y, b, t, xmin, xmax, ymin, ymax;
  SP_globref ref = SP_alloc_globrefs(6);
  struct dvar localvar[3];
  Dvar dvx = &localvar[0];
  Dvar dvy = &localvar[1];
  Dvar dvb = &localvar[2];
  
  w->numstack_end = NULL;

/*    X(0) = RefTerm(State); */
  RefTerm(NewState) = RefTerm(State);
  dvar_export_start(3);
  DerefNonvar(X(0));
  DerefArg(x,X(0),1);		/* X */
  RefGlob(ref+1) = x;
  DerefArg(t,X(0),2);		/* XMut */
  RefGlob(ref+0) = t;
  DerefArg(y,X(0),3);		/* Y */
  RefGlob(ref+3) = y;
  DerefArg(t,X(0),4);		/* YMut */
  RefGlob(ref+2) = t;
  DerefArg(b,X(0),5);		/* B */
  RefGlob(ref+5) = b;
  DerefArg(t,X(0),6);		/* BMut */
  RefGlob(ref+4) = t;
  dvar_init(dvx, ref+0, ref+1);
  dvar_init(dvy, ref+2, ref+3);
  dvar_init(dvb, ref+4, ref+5);
  xmin = dvar_min_t(dvx);
  xmax = dvar_max_t(dvx);
  ymin = dvar_min_t(dvy);
  ymax = dvar_max_t(dvy);
  if (IsVar(b)) {
    if (FDlt(ymax,xmin)) {	/* [B=0,exit] */
      ent = 1;
      dvar_fix_value_t(dvb, TaggedZero);
    } else if (!FDgt(xmax,ymin)) { /* [B=1,exit] */
      ent = 1;
      dvar_fix_value_t(dvb, TaggedOne);
    }
  } else if (Teqz(b)) {		/* enforce X #> Y */
    if (FDlt(ymax,xmin))
      ent = 1;
    else if (!FDgt(xmax,ymin))
      ent = -1;
    else {
      TAGGED ymin1 = ymin+IStep(1);
      TAGGED xmax1 = xmax-IStep(1);
      
      if (TagIsSmall(ymin) && FDlt(xmin,ymin1))
	dvar_fix_min_t(dvx, ymin1);
      if (TagIsSmall(xmax) && FDgt(ymax,xmax1))
	dvar_fix_max_t(dvy, xmax1);
      ent = !IsVar(x) || !IsVar(y) || xmax==ymin1;
    }
  } else {			/* enforce X #=< Y */
    if (FDlt(ymax,xmin))
      ent = -1;
    else if (!FDgt(xmax,ymin))
      ent = 1;
    else {
      if (TagIsSmall(ymax) && FDgt(xmax,ymax))
	dvar_fix_max_t(dvx, ymax);
      if (TagIsSmall(xmin) && FDlt(ymin,xmin))
	dvar_fix_min_t(dvy, xmin);
      ent = !IsVar(x) || !IsVar(y) || ymax==xmin;
    }
  }
  for (i=0; i<3; i++)
    dvar_pruning_done(&localvar[i]);
  for (i=0; i<3; i++)
    dvar_export(&localvar[i], 3);
  dvar_export_done(Actions, ent, 3);
  SP_free_globrefs(ref,6);
}



/* '$fd_oneof'(+State, -Actions) :- X#=Y #\/ X#=Z.
   State is f(X,XMut,Y,YMut,Z,ZMut),
   Actions is a list of prunings etc.
*/
void SPCDECL
prolog_fd_oneof MAGIC (HIDDEN_PROTO
		       SP_term_ref State,
		       SP_term_ref NewState,
		       SP_term_ref Actions)
{
  WAMENV;
  int ent = 0;		/* neither entailed nor dis- */
  int xcapz, ycapz;
  int i;
  TAGGED x, y, z, t, xdom, ydom, zdom;
  SP_globref ref = SP_alloc_globrefs(6);
  struct dvar localvar[3];
  Dvar dvx = &localvar[0];
  Dvar dvy = &localvar[1];
  Dvar dvz = &localvar[2];
  
  w->numstack_end = NULL;

/*    X(0) = RefTerm(State); */
  RefTerm(NewState) = RefTerm(State);
  dvar_export_start(3);
  DerefNonvar(X(0));
  DerefArg(x,X(0),1);		/* X */
  RefGlob(ref+1) = x;
  DerefArg(t,X(0),2);		/* XMut */
  RefGlob(ref+0) = t;
  DerefArg(y,X(0),3);		/* Y */
  RefGlob(ref+3) = y;
  DerefArg(t,X(0),4);		/* YMut */
  RefGlob(ref+2) = t;
  DerefArg(z,X(0),5);		/* Z */
  RefGlob(ref+5) = z;
  DerefArg(t,X(0),6);		/* ZMut */
  RefGlob(ref+4) = t;
  dvar_init(dvx, ref+0, ref+1);
  dvar_init(dvy, ref+2, ref+3);
  dvar_init(dvz, ref+4, ref+5);
  xdom = dvar_set(dvx);
  ydom = dvar_set(dvy);
  zdom = dvar_set(dvz);
  xcapz = fd_compare(xdom,zdom);
  ycapz = fd_compare(ydom,zdom);
  if (xcapz==FDI_DISJOINT)
    ent = tell_eq(dvy,dvz,ycapz);
  else if (ycapz==FDI_DISJOINT)
    ent = tell_eq(dvx,dvz,xcapz);
  else {
    TAGGED xcupy = fd_or(xdom,ydom);

    switch (fd_compare(zdom,xcupy)) {
    case FDI_INTERSECT:
    case FDI_SUPERSET:
      dvar_fix_set(dvz, xcupy);
      /* entailed if at most one variable left */
      ent = (IsVar(x) + IsVar(y) + (!dvar_is_integer(dvz)) <= 1);
      break;
    case FDI_DISJOINT:
      ent = -1;
      break;
    }
  }
  for (i=0; i<3; i++)
    dvar_pruning_done(&localvar[i]);
  for (i=0; i<3; i++)
    dvar_export(&localvar[i], 3);
  dvar_export_done(Actions, ent, 3);
  SP_free_globrefs(ref,6);
}


/* computes -dom(d2) */
/* ripped off from indexical.c */
static TAGGED negdom MAGIC (HIDDEN_PROTO
			    TAGGED d2)
{
  int j;
  TAGGED t2, r2, tail, *h, *array;

  for (j=0, t2=d2; t2!=EmptySet; j++)
    t2 = CTagToCdr(t2);

				/* create j intervals, then merge */

  array = numstack_alloc(4*j);
  h = array + 4*j;
  tail = EmptySet;
  for (t2=d2; t2!=EmptySet;) {
    h -= 4;
    r2 = CTagToCar(t2); t2 = CTagToCdr(t2);
    h[0] = MakeList(h+2);
    h[1] = tail;
    h[2] = safe_minus(TaggedZero,RangeMax(r2));
    h[3] = safe_minus(TaggedZero,RangeMin(r2));
    tail = MakeList(h);
  }
  return tail;
}


/* '$fd_abs'(+State, -Actions) :- abs(X) #= Y.
   State is f(X,XMut,Y,YMut),
   Actions is a list of prunings etc.
*/
void SPCDECL
prolog_fd_abs MAGIC (HIDDEN_PROTO
		     SP_term_ref State,
		     SP_term_ref NewState,
		     SP_term_ref Actions)
{
  WAMENV;
  int ent = 0;		/* neither entailed nor dis- */
  int i;
  TAGGED x, y, t, xdom, ydom;
  SP_globref ref = SP_alloc_globrefs(4);
  struct dvar localvar[2];
  Dvar dvx = &localvar[0];
  Dvar dvy = &localvar[1];
  
  w->numstack_end = NULL;

/*    X(0) = RefTerm(State); */
  RefTerm(NewState) = RefTerm(State);
  dvar_export_start(3);
  DerefNonvar(X(0));
  DerefArg(x,X(0),1);		/* X */
  RefGlob(ref+1) = x;
  DerefArg(t,X(0),2);		/* XMut */
  RefGlob(ref+0) = t;
  DerefArg(y,X(0),3);		/* Y */
  RefGlob(ref+3) = y;
  DerefArg(t,X(0),4);		/* YMut */
  RefGlob(ref+2) = t;
  dvar_init(dvx, ref+0, ref+1);
  dvar_init(dvy, ref+2, ref+3);
  xdom = dvar_set(dvx);
  ydom = dvar_set(dvy);
  if (!FDlt(fd_min(xdom),TaggedZero))
    ent = tell_eq(dvx,dvy,fd_compare(xdom,ydom));
  else {
    TAGGED ydom0 = ydom;
    TAGGED xndom, yndom, xdom1, ydom1;
    
    if (FDlt(fd_min(ydom),TaggedZero))
      ydom0 = fd_and_interval(ydom0,TaggedZero,Sup);
    xndom = negdom(xdom);
    yndom = negdom(ydom0);
    xdom1 = fd_or(ydom0,yndom);
    ydom1 = fd_or(xdom,xndom);
    if (FDlt(fd_min(ydom),TaggedZero))
      ydom1 = fd_and_interval(ydom1,TaggedZero,Sup);
    switch (fd_compare(xdom,xdom1)) {
    case FDI_SUBSET:
    case FDI_EQUAL:
      break;
    case FDI_DISJOINT:
      ent = -1;
      goto ret;
    case FDI_INTERSECT:
      xdom1 = fd_and(xdom,xdom1);
    case FDI_SUPERSET:
      dvar_fix_set(dvx, xdom1);
    }
    switch (fd_compare(ydom,ydom1)) {
    case FDI_SUBSET:
    case FDI_EQUAL:
      ent = fd_singleton(ydom1);
      break;
    case FDI_DISJOINT:
      ent = -1;
      break;
    case FDI_INTERSECT:
      ydom1 = fd_and(ydom,ydom1);
    case FDI_SUPERSET:
      ent = fd_singleton(ydom1);
      dvar_fix_set(dvy, ydom1);
    }
  }
  for (i=0; i<2; i++)
    dvar_pruning_done(&localvar[i]);
  for (i=0; i<2; i++)
    dvar_export(&localvar[i], 3);
 ret:
  dvar_export_done(Actions, ent, 3);
  SP_free_globrefs(ref,4);
}

