/* Copyright(C) 1999, Swedish Institute of Computer Science */

#include "fd.h"
#include "dvars.h"

#define CSTATE_ONLY 0

#if MULTI_SP_AWARE
#define ord_list_to_set(A1,A2) ord_list_to_set(HIDDEN_ARG, A1,A2)
#define element_feasible(A1,A2,A3,A4) element_feasible(HIDDEN_ARG, A1,A2,A3,A4)
#define qsort_taggedswap(A1,A2,A3,A4) qsort_taggedswap(HIDDEN_ARG, A1,A2,A3,A4)
#define qsort_taggedmed3(A1,A2,A3) qsort_taggedmed3(HIDDEN_ARG, A1,A2,A3)
#define qsort_tagged(A1,A2) qsort_tagged(HIDDEN_ARG, A1,A2)
#if 0                           /* not redefined. It uses FD_SETUP_SPENV to access the SPEnv */
#define element_destructor(A1) element_destructor(HIDDEN_ARG, A1)
#endif
#endif /* MULTI_SP_AWARE */

static TAGGED ord_list_to_set MAGIC (HIDDEN_PROTO
				     TAGGED *vec,
				     int k)
{
  int i;
  FDCONS cons;

  fdcons_init(&cons);
  for (i=0; i<k; i++)
    fdcons_add(&cons,vec[i]);
  return fdcons_set(&cons);
}


static int cmp_tagged(TAGGED *x, TAGGED *y)
{
  long val1 = *x;
  long val2 = *y;

  return CMP(val1,val2);
}


#define QType TAGGED
#define QCmp  cmp_tagged
#define QSort qsort_tagged
#include "qsort.ic"



/*
   '$fd_relation'(+State0, -State, -Actions) :-
   State0 is f(N,X-XMut,Pairs,Y-YMut) where
   Pairs is a list of N possible X-Y values,
   State  similarly,
   Actions is a list of prunings etc.

   Max. heap need, except Actions:
   5     for State
   2N    for new Pairs
   ====
   2N+5 total
*/
void SPCDECL
prolog_fd_relation MAGIC (HIDDEN_PROTO
			  SP_term_ref State0,
			  SP_term_ref State,
			  SP_term_ref Actions)
{
  WAMENV;
  int k, n, ent = -1;		/* disentailed */
  TAGGED pair, pairs;		/* Pair, Pairs */
  TAGGED xpair;			/* X-XMut */
  TAGGED xdom;			/* XDom */
  TAGGED *xvals;		/* filtered X values */
  TAGGED ypair;			/* Y-YMut */
  TAGGED ydom;			/* YDom */
  TAGGED *yvals;		/* filtered Y values */
  TAGGED *shared, *newtail, outarg, oldtail, *h, x, y, t1;
  BOOL yorder = TRUE;
  SP_globref refs = SP_alloc_globrefs(4);
  struct dvar dvx, dvy;
  
  w->numstack_end = NULL;
/*    X(0) = RefTerm(State0); */
  (void)State0;
  dvar_export_start(3);
  DerefNonvar(X(0));
  DerefArg(t1,X(0),1);  n = GetSmall(t1);

  RequireHeap(2*n+5, 3);

  DerefArg(xpair,X(0),2)	/* get X-XMut */
  DerefArg(pairs,X(0),3)	/* get Pairs */
  DerefArg(ypair,X(0),4);	/* get Y-YMut */
  DerefArg(t1,xpair,1);		/* get and protect X */
  RefGlob(refs+1) = t1;
  DerefArg(t1,xpair,2);		/* get and protect XMut */
  RefGlob(refs  ) = t1;
  DerefAttribute(xdom,t1);	/* get dom/4 */
  xdom = DomainSet(xdom);	/* get FD set */
  DerefArg(t1,ypair,1);		/* get and protect Y */
  RefGlob(refs+3) = t1;
  DerefArg(t1,ypair,2);		/* get and protect YMut */
  RefGlob(refs+2) = t1;
  DerefAttribute(ydom,t1);	/* get dom/4 */
  ydom = DomainSet(ydom);	/* get FD set */
  dvar_init(&dvx, refs,   refs+1);
  dvar_init(&dvy, refs+2, refs+3);

  xvals = Calloc(n, TAGGED);	/* get space for x values */
  yvals = Calloc(n, TAGGED);	/* get space for y values */
  k = 0;			/* k filtered pairs */
  shared = newtail = &outarg;
  oldtail = pairs;
  h = w->global_top;
  while (TagIsLST(pairs)) {
    DerefCar(pair,pairs);
    DerefCdr(pairs,pairs);
    DerefArg(x,pair,1);
    DerefArg(y,pair,2);
    if (fd_member(x,xdom) && fd_member(y,ydom)) { /* O(N^2) cost here! */
      xvals[k] = x;
      yvals[k] = y;
      if (k>0 && Tgt(yvals[k-1],y))
	yorder = FALSE;
      k++;
      *newtail = MakeList(h);
      *h = pair;
      newtail = h+1;
      h += 2;
    } else {
      oldtail = pairs;
      shared = newtail;
      w->global_top = h;
    }
  }
  if (k==0)
    goto ret;
  *shared = oldtail;
  h = w->global_top;
  X(0) = MakeStructure(h);
  *h++ = TagToHeadfunctor(X(0));
  *h++ = MakeSmall(k);
  *h++ = xpair;
  *h++ = outarg;
  *h++ = ypair;
  w->global_top = h;
  RefTerm(State) = X(0);
  
  if (!yorder)
    qsort_tagged(yvals, k);
  dvar_prune_set(&dvx, ord_list_to_set(xvals,k));
  dvar_prune_set(&dvy, ord_list_to_set(yvals,k));
  dvar_pruning_done(&dvx);
  dvar_pruning_done(&dvy);
  dvar_export(&dvx, 3);
  dvar_export(&dvy, 3);
  ent = 0;
  if (xvals[0]==xvals[k-1] || yvals[0]==yvals[k-1])
    ent = 1;
ret:
  SP_free_globrefs(refs,4);
  dvar_export_done(Actions, ent, 3);
}



struct element_data {
  void (SPCDECL *destructor)(void *);
  void (SPCDECL *daemon)(HIDDEN_PROTO
			 void *,SP_globref,TAGGED); /* (handle,attr_ref,global) */
#if MULTI_SP_AWARE
  SPEnv *spenv;
#endif /* MULTI_SP_AWARE */

  long stamp;
  SP_globref refbase;		/* first 4 for XMut, X, YMut, Y */
  int nrefs;			/* including first 4 */
  int nelts;
  int first_feas;
  int first_infeas;
#if CSTATE_ONLY
  int first_feas_save;
  int first_infeas_save;
#endif
  struct {
    TAGGED *xmin, *xmax;
    TAGGED *ymin, *ymax;
    SP_globref *mutref;
  } elt;
};

#define XMIN(E) (pdata->elt.xmin[E])
#define XMAX(E) (pdata->elt.xmax[E])
#define YMIN(E) (pdata->elt.ymin[E])
#define YMAX(E) (pdata->elt.ymax[E])
#define MUTREF(E) (pdata->elt.mutref[E])


static void SPCDECL element_destructor(void *pdata_v)
{
  struct element_data *pdata = (struct element_data *)pdata_v;
  FD_SETUP_SPENV(pdata->spenv)

  SP_free_globrefs(pdata->refbase,pdata->nrefs);
  SP_free(pdata);
}


static BOOL element_feasible MAGIC (HIDDEN_PROTO
				    struct element_data *pdata,
				    int edata,
				    Dvar dvx, Dvar dvy)
{
  BOOL ycond;
  TAGGED edom;

  if (XMIN(edata)==XMAX(edata)) {
    ycond = dvar_contains_value_t(dvx,XMIN(edata));
  } else {
    ycond = (dvar_compare_interval_t(dvx,XMIN(edata),XMAX(edata))!=FDI_DISJOINT);
  }
  if (ycond) {
    if (MUTREF(edata)) {		/* refresh */
      TAGGED tmp;
      DerefAttribute(edom,RefGlob(MUTREF(edata)));
				/* compute min&max of domain intersection */
      switch (dvar_compare_set(dvy,DomainSet(edom))) {
      case FDI_EQUAL:
      case FDI_SUPERSET:
	YMIN(edata) = DomainMin(edom);
	YMAX(edata) = DomainMax(edom);
	break;
      case FDI_SUBSET:
	YMIN(edata) = dvar_min_t(dvy);
	YMAX(edata) = dvar_max_t(dvy);
	break;
      case FDI_INTERSECT:
	tmp = fd_and(dvar_set(dvy),DomainSet(edom));
	YMIN(edata) = fd_min(tmp);
	YMAX(edata) = fd_max(tmp);
	break;
      case FDI_DISJOINT:
	ycond = FALSE;
	break;
      }
    } else
      ycond = dvar_contains_value_t(dvy,YMIN(edata));
  }
  return ycond;
}


static void SPCDECL 
element_daemon MAGIC (HIDDEN_PROTO
		      void *vdata,
		      SP_globref attr_ref,
		      TAGGED global)
{
  struct element_data *pdata = (struct element_data *)vdata;
  TAGGED tstate;
  int ar, state_stamp;
  WAMENV;

  (void)attr_ref;
  tstate = RefMutable(CTagToArg(global,1));
  ar = Arity(TagToHeadfunctor(tstate));
  state_stamp = GetSmall(CTagToArg(tstate,ar));
  if (pdata->stamp==state_stamp) { /* incremental */
    struct dvar dvx, dvy;
    SP_globref refbase = pdata->refbase;
    TAGGED ymin, ymax;
    int i;
    int mask = 0x0;		/* accumulates what we've seen
				   0x1 = support for min(Y)
				   0x2 = support for max(Y) */
    

    dvar_init(&dvx, refbase,   refbase+1);
    dvar_init(&dvy, refbase+2, refbase+3);
    ymin = dvar_min_t(&dvy);
    ymax = dvar_max_t(&dvy);
    if (dvar_is_integer(&dvx))
      mask = 0x7;
    for (i=pdata->first_feas; i<pdata->first_infeas && mask<0x3; i++) {
      if (dvar_compare_interval_t(&dvx,XMIN(i),XMAX(i))!=FDI_DISJOINT) {
	/* bin i is feasible wrt. X */
	if (!(mask & 0x1) && Tge(ymin,YMIN(i)) && Tle(ymin,YMAX(i)))
	  mask |= 0x1;		/* support for min(Y) */
	if (!(mask & 0x2) && Tge(ymax,YMIN(i)) && Tle(ymax,YMAX(i)))
	  mask |= 0x2;		/* support for max(Y) */
      }
    }
    if (mask==0x3)
      return;
  }
  fd_enqueue_global(w, global, 0x4/* MINMAX, append*/);   
}

/* '$fd_element'(+State0, -State, -Actions) :-
   State0 is f(X-XMut,[E1,E2,...],Y-YMut,LeftInfeas,RightInfeas,Handle,Stamp) where
   State  is State0 with updated Stamp,
   Actions is a list of prunings etc.
*/
void SPCDECL
prolog_fd_element MAGIC (HIDDEN_PROTO
			 SP_term_ref State0,
			 SP_term_ref State,
			 SP_term_ref Actions)
{
  WAMENV;
  int hit, miss, ent = -1; /* disentailed */
  int nelts=0, nvars=0, edata=0;
  long state_stamp;
  BOOL committed, post;
  TAGGED elt, elts;		/* Elt, Elts */
  TAGGED xelt;			/* X-XMut */
  TAGGED yelt;			/* Y-YMut */
  TAGGED handle;		/* perm data handle */
  TAGGED emin=0, emax=0, edom;
  TAGGED t1, ti, elt0, xmin, xmax, xset;
  SP_globref eref, refbase;
  struct element_data *pdata;
  int i, j, k;
  char *ptr;
  struct dvar dvx, dvy, dve;
  
  w->numstack_end = NULL;

/*    X(0) = RefTerm(State0); */
  (void)State0;
  dvar_export_start(3);
  RefTerm(State) = unify_output_state(w,&handle,&state_stamp,&committed);

  if (!IsVar(handle)) {		/* got [Flag | '$free'(Ptr)] */
    post = FALSE;
    pdata = Pdata(struct element_data,handle);
    refbase = pdata->refbase;
    nelts = pdata->nelts;
  } else {
    post = TRUE;
    DerefArg(elts,X(0),2);	/* get Elts */
    elt0 = 0;
    while (TagIsLST(elts)) {
      DerefCar(elt,elts);
      DerefCdr(elts,elts);
      if (elt!=elt0) {
	nelts++;
	if (IsVar(elt)) nvars++;
	elt0 = elt;
      }
    }
    pdata = Palloc(struct element_data,
		   nelts*5*sizeof(TAGGED),
		   handle);
    /* store header data */
    pdata->destructor = element_destructor;
    pdata->daemon = element_daemon;
    FD_STORE_SPENV(pdata->spenv);
    pdata->nrefs = 2*nvars + 4;
    pdata->refbase = refbase = SP_alloc_globrefs(2*nvars + 4);
    pdata->stamp = state_stamp-1; /* DON'T trust initially */
    pdata->nelts = nelts;
#if CSTATE_ONLY
    pdata->first_feas = pdata->first_feas_save = 0;
    pdata->first_infeas = pdata->first_infeas_save = nelts;
#endif
    ptr = (char *)(pdata+1);
    pdata->elt.xmin = (TAGGED *)ptr; ptr += nelts*sizeof(TAGGED);
    pdata->elt.xmax = (TAGGED *)ptr; ptr += nelts*sizeof(TAGGED);
    pdata->elt.ymin = (TAGGED *)ptr; ptr += nelts*sizeof(TAGGED);
    pdata->elt.ymax = (TAGGED *)ptr; ptr += nelts*sizeof(TAGGED);
    pdata->elt.mutref = (SP_globref *)ptr; ptr += nelts*sizeof(SP_globref);
    /* decode and store X and Y data */
    DerefArg(xelt,X(0),1)	/* get X-XMut */
    DerefArg(yelt,X(0),3);	/* get Y-YMut */
    DerefArg(t1,xelt,1);	/* get and protect X */
    RefGlob(refbase+1) = t1;
    DerefArg(t1,xelt,2);	/* get and protect XMut */
    RefGlob(refbase  ) = t1;
    DerefArg(t1,yelt,1);	/* get and protect Y */
    RefGlob(refbase+3) = t1;
    DerefArg(t1,yelt,2);	/* get and protect YMut */
    RefGlob(refbase+2) = t1;
    /* decode and store Elts data */
    DerefArg(elts,X(0),2);	/* get Elts */
    eref = refbase+4;
    elt0 = 0;
    ti = TaggedZero;
    while (TagIsLST(elts)) {
      ti += IStep(1);
      DerefCar(elt,elts);
      DerefCdr(elts,elts);
      if (elt!=elt0) {
	elt0 = elt;
	XMIN(edata) = ti;
	XMAX(edata) = ti;
	if (IsVar(elt)) {
	  RefGlob(eref+1) = elt; /* protect elt var */
	  t1 = check_argument(w,elt,Inf,Sup,Sup);
	  RefGlob(eref) = t1;	/* protect elt attribute */
	  DerefAttribute(edom,t1);
	  MUTREF(edata) = eref;
	  eref += 2;
	} else {
	  YMIN(edata) = elt;
	  YMAX(edata) = elt;
	  MUTREF(edata) = 0;
	}
	edata++;
      } else {
	XMAX(edata-1) = ti;
      }
    }
  }
#if CSTATE_ONLY
  if (state_stamp != pdata->stamp) { /* trust nothing */
    pdata->first_feas = pdata->first_feas_save;
    pdata->first_infeas = pdata->first_infeas_save;
  }
#else
  DerefArg(elt,X(0),4);
  pdata->first_feas = GetSmall(elt);
  DerefArg(elt,X(0),5);
  pdata->first_infeas = nelts - GetSmall(elt);
#endif
  pdata->stamp = state_stamp+1;
  
  dvar_init(&dvx, refbase,   refbase+1);
  dvar_init(&dvy, refbase+2, refbase+3);
  xmin = dvar_min_t(&dvx);
  xmax = dvar_max_t(&dvx);
  xset = dvar_set(&dvx);

  /* dichotomic search for first item that is not before X */
  i = pdata->first_feas;
  k = pdata->first_infeas;
  while (i<k) {
    j = (i+k)>>1;
    if (Tgt(xmin,XMAX(j)))
      i = j+1;
    else
      k = j;
  }
  pdata->first_feas = i;

  /* scan the feasible items */
  hit = -1;
  miss = i-1;
  for (k=pdata->first_infeas; i<k; i++) {
    edata = i;

    if (Tlt(xmax,XMIN(edata)))
      break;
    if (element_feasible(pdata,edata,&dvx,&dvy)) {
      /*
      if (XMIN(edata) == xmin)
	xmin = XMAX(edata) + IStep(1);
      else
	xset = fd_subtract_interval(xset,XMIN(edata),XMAX(edata));
      */
      if (hit == -1) {		/* form union(Edom) */
	emin = YMIN(edata);
	emax = YMAX(edata);
      } else {			/* form interval hull */
	if (FDgt(emin,YMIN(edata)))
	  emin = YMIN(edata);
	if (FDlt(emax,YMAX(edata)))
	  emax = YMAX(edata);
      }
      hit = i;
    } else {
      if (hit == -1) {
	pdata->first_feas = i+1;
      } else if (miss<hit) {
	if (XMIN(miss+1) == xmin)
	  xmin = XMIN(edata);
	else
	  xset = fd_subtract_interval(xset,
				      XMIN(miss+1),
				      XMIN(edata)-IStep(1));
      }
      miss = i;
    }
  }
  if (hit == -1) {
    goto ret;
  } else if (miss<hit) {
    if (XMIN(miss+1) == xmin)
      xmin = XMAX(hit)+IStep(1);
    else
      xset = fd_subtract_interval(xset, XMIN(miss+1), XMAX(hit));
  }
  pdata->first_infeas = hit+1;
#if CSTATE_ONLY
  if (committed) {
    pdata->first_feas_save = pdata->first_feas;
    pdata->first_infeas_save = hit+1;
  }
#else
  CTagToArg(X(0),4) = MakeSmall(pdata->first_feas);
  CTagToArg(X(0),5) = MakeSmall(nelts - pdata->first_infeas);
#endif  
  if (FDlt(emin,dvar_min_t(&dvy)))
    emin = dvar_min_t(&dvy);
  if (FDgt(emax,dvar_max_t(&dvy)))
    emax = dvar_max_t(&dvy);
  if (Tle(xmin,xmax))
    dvar_prune_set(&dvx, fd_and_interval(xset,xmin,xmax));
  dvar_fix_interval_t(&dvy, emin, emax);
  ent = (emin==emax && pdata->first_feas==hit);
  dvar_pruning_done(&dvx);
  dvar_pruning_done(&dvy);
  dvar_export(&dvx, 3);
  dvar_export(&dvy, 3);
  if (pdata->first_feas==hit) {	/* can prune E var too */
    eref = MUTREF(hit);
    if (eref) {
      dvar_init(&dve, eref, eref+1);
      dvar_fix_interval_t(&dve, emin, emax);
      dvar_pruning_done(&dve);
      dvar_export(&dve, 3);
      if (!ent) {
	dvar_export_equal(&dve, &dvy, 3); /* do this AFTER any dvar_export! */
	ent = 1;
      }
    }
  }
  if (ent==0 && post)
    attach_daemon(pdata, refbase,   refbase+1, X(1), fd.functor_dom, 3);
ret:
  Pfree(ent==1,handle,pdata);
  dvar_export_done(Actions, ent, 3);
}

