/* Copyright(C) 1999, Swedish Institute of Computer Science */

#include "fd.h"

#if MULTI_SP_AWARE
#define run_all_daemons() run_all_daemons(HIDDEN_ARG)
#define fd_enqueue_list_gc(A1,A2,A3,A4) fd_enqueue_list_gc(HIDDEN_ARG, A1,A2,A3,A4)
#define fd_enqueue_list(A1,A2,A3,A4) fd_enqueue_list(HIDDEN_ARG, A1,A2,A3,A4)
#define fd_enqueue_val(A1,A2,A3,A4,A5) fd_enqueue_val(HIDDEN_ARG, A1,A2,A3,A4,A5)
#define compress_trail(A1,A2) compress_trail(HIDDEN_ARG, A1,A2)
#endif /* MULTI_SP_AWARE */

/* If defined, grounded variables prepend constraints in the non-VAL
   lists onto the propagation queue.  If undefined, they get
   appended.  In any case, constraints in the VAL list get prepended.
   The jury is still out.  Throughout SP 3.x, we have defined it.
*/
#define DEPTH_FIRST_STRATEGY 1

/* Returns FD attribute for a dvar or integer.
   Heap consumption bounds:
   existing dvar - 0
   integer - INT_ATTRIBUTE_SIZE
   non-domain variable - FD_ATTRIBUTE_SIZE + ARITYLIMIT + 4
*/
TAGGED check_argument MAGIC (HIDDEN_PROTO
			     Argdecl,
			     TAGGED argument,
			     TAGGED min, TAGGED max, TAGGED size)
{
  TAGGED *h, t1, attr;
  int j;

  DerefSwitch(argument,;);
  switch (TagOf(argument))
    {
    case TAG_CVA:
      if ((t1=get_attributes(argument,fd.fd_module)))
	return t1;
    case TAG_SVA:
    case TAG_HVA:
      h = w->global_top;
      attr = MakeStructure(h);
      for (j=0; j<FD_ATTRIBUTE_SIZE; j++) {
	t1 = fd_attribute[j];
	if (t1&1)
	  t1 += TagHVA(w->global_top)-1;
	*h++ = t1;
      }
      w->global_top[FD_ATTR_MIN_OFFSET] = min;
      w->global_top[FD_ATTR_MAX_OFFSET] = max;
      w->global_top[FD_ATTR_SIZE_OFFSET] = size;
      w->global_top = h;
      put_attributes(argument,attr,fd.fd_module);
      return attr;
    case TAG_SIN:
      h = w->global_top;
      attr = MakeStructure(h);
      *h++ = fd_attribute[0];
      *h++ = fd_attribute[1];
      *h++ = fd_attribute[2];
      *h++ = attr + WD(5);
      *h++ = TaggedZero;
/* 5*/*h++ = fd.functor_mutable;
      *h++ = attr + WD(8);
      *h++ = TaggedZero;
/* 8*/*h++ = fd.functor_dom4;
      *h++ = MakeList(w->global_top+13);
/*10*/*h++ = argument;
      *h++ = argument;
      *h++ = TaggedOne;
/*13*/*h++ = MakeList(w->global_top+10);
      *h++ = atom_nil;
      w->global_top = h;
      return attr;
    default:
      return ERRORTAG;
    }
}


/* '$fd_arg_attribute'(+Var, +Finitep, -Attr)
*/
void SPCDECL
prolog_fd_arg_attribute MAGIC (HIDDEN_PROTO
			       SP_term_ref Var,
			       long finitep,
			       SP_term_ref Attr)
{
  WAMENV;
  TAGGED attr, domain;

  attr = check_argument(w,RefTerm(Var),Inf,Sup,Sup);
  if (attr && finitep) {
    DerefAttribute(domain,attr); /* dom/4 term */
    if (!AreSmall(DomainMin(domain),DomainMax(domain)))
      attr = ERRORTAG;
  }
  if (attr)
    RefTerm(Attr) = attr;
  else
    SP_fail();  
}

/* '$fd_dvar_list'(+List, +Finitep)
*/
void SPCDECL
prolog_fd_dvar_list MAGIC (HIDDEN_PROTO
			   SP_term_ref List,
			   long finitep)
{
  WAMENV;
  TAGGED list, attr, domain;
  int i = 0;

  DEREF(list,RefTerm(List));
  while (TagIsLST(list)) {
    DerefCar(attr,list);
    DerefCdr(list,list);
    if (IsVar(attr))
      i++;
  }
  if (list!=atom_nil)		/* type error? */
    goto fail;
  if (i>0 && !finitep) {
    RequireHeap(i*(FD_ATTRIBUTE_SIZE + ARITYLIMIT + 4),0);
  }
  DEREF(list,RefTerm(List));
  while (TagIsLST(list)) {
    DerefCar(attr,list);
    DerefCdr(list,list);
    if (!TagIsSmall(attr)) {
      attr = check_argument(w,attr,Inf,Sup,Sup);
      if (!attr)
	goto fail;
      if (finitep) {
	DerefAttribute(domain,attr); /* dom/4 term */
	if (!AreSmall(DomainMin(domain),DomainMax(domain)))
	  goto fail;
      }
    }
  }
  return;
 fail:
  SP_fail();
}

/* '$fd_coref'(+List)
   Succeeds if List contains F1(X) and F2(X) for some X.
*/
void SPCDECL
prolog_fd_coref MAGIC (HIDDEN_PROTO SP_term_ref List)
{
  WAMENV;
  TAGGED list, var;

  DEREF(list,RefTerm(List));
  while (TagIsLST(list)) {
    DerefCar(var,list);
    DerefCdr(list,list);
    var = CTagToArg(var,1);
    DerefHeapSwitch(var,goto bind;);
    if (var==atom_nil)		/* F(Var), 2nd occurrence */
      return;
    else
      continue;			/* F(Integer) */
  bind:				/* F(Var), 1st occurrence */
    TrailPushCheck(var);
    CTagToPointer(var) = atom_nil;
  }
  SP_fail();			/* No coreference found. */
}


BOOL fd_member MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                      TAGGED x, TAGGED set)
{
  TAGGED range;

  while (set!=EmptySet)
    {
      range = CTagToCar(set); 
      set = CTagToCdr(set);   
      switch (val_vs_range(x,range))
	{
	case CMP_INSIDE: return TRUE;
	case CMP_BEFORE: return FALSE;
	}
    }
  return FALSE;
}



BOOL fd_check_overflow MAGIC (HIDDEN_PROTO_VOID)
{
  if (fd.fd_overflow)
    {
      SP_term_ref excp = SP_new_term_ref();
      SP_term_ref a2 = SP_new_term_ref();
      SP_term_ref a3 = SP_new_term_ref();

      SP_put_integer(a2, 0);
      SP_put_string(a3, "CLPFD integer overflow");
      (void)SP_cons_functor(excp,SP_atom_from_string("representation_error"),3,
			    a2,a2,a3);
      SP_raise_exception(excp);
      SP_reset_term_refs(excp);
      return FALSE;
    }
  return TRUE;
}

/* Build a copy of old on the heap.
   Precondition: old is built entirely on the numstack.
*/
TAGGED fd_globalize MAGIC (HIDDEN_PROTO
			   Argdecl,
			   TAGGED old, long req, int ar)
{
  TAGGED d1, r1, b, e, *h, value;
  TAGGED *valuep = &value;

  d1 = old;
  while (d1!=EmptySet)
      req += 4,
      d1 = CTagToCdr(d1);
  RequireHeap(req,ar);
  h = w->global_top;
  d1 = old;
  while (d1!=EmptySet) {
    r1 = CTagToCar(d1);
    d1 = CTagToCdr(d1);
    b = RangeMin(r1);
    e = RangeMax(r1);
    *valuep = MakeList(h);
    *h++ = *valuep+WD(2);
    valuep = h++;
    if (b==Sup || e==Inf)
      fd.fd_overflow = TRUE;
    *h++ = b;
    *h++ = e;
  }
  *valuep = atom_nil;
  w->global_top = h;
  return value;
}



void update_mutable MAGIC (HIDDEN_PROTO
			   struct worker *w,
			   TAGGED new_value, TAGGED mutable)
{
  TAGGED *h, *arg, value;
  LetShadowregs;
  
  value=RefMutable(mutable);
  while (TagIsSTR(value) && TagToHeadfunctor(value)==fd.functor_mutable) {
    mutable=value; value=RefMutable(value);
  }							
  arg = TagToArg(mutable,0);
  if (arg[2] < TrailToInt(w->node->trail_top) &&
      arg < GLOBAL_UNCOND)
    {
      h = w->trail_top; /* trail mutable if need be */
      *h++ = mutable;
      *h++ = arg[1];
      *h++ = arg[2];
      arg[1] = new_value;	/* must come BEFORE choice_overflow */
      arg[2] = TrailToInt(w->trail_top);
      w->trail_top = h;
      if (ChoiceYounger(ChoiceOffset(w->node,CHOICEPAD),w->trail_top))
	choice_overflow(CHOICEPAD);
    }
  else
    arg[1] = new_value;
}

void fd_link MAGIC (HIDDEN_PROTO
		    struct worker *w,
		    TAGGED var,
		    long key,
		    TAGGED item)
{
  TAGGED mutable;
  TAGGED *h;
  TAGGED *s=NULL;
  TAGGED queues;
  long mask;
  LetShadowregs;

  AttrToSuspM(check_argument(w,var,Inf,Sup,Sup),mutable);
  queues = RefMutable(mutable);
  h = w->global_top;
  
  if (TagToSTR(queues) < GLOBAL_UNCOND)/* can't smash */
    {
      int i;
	  
      for (s = TagToSTR(queues), i=0; i<8; i++)
	*h++ = *s++;
      queues = MakeStructure(h-8);
      update_mutable(w,queues,mutable);
    }
  if (key==fd.functor_dom)
    mask = IStep(MASK_DOM),
    s = TagToArg(queues,3);
  else if (key==fd.functor_min)
    mask = IStep(MASK_MIN),
    s = TagToArg(queues,4);
  else if (key==fd.functor_max)
    mask = IStep(MASK_MAX),
    s = TagToArg(queues,5);
  else if (key==fd.functor_minmax)
    mask = IStep(MASK_MINMAX),
    s = TagToArg(queues,6);
  else
    {
				/* patch item for globals suspending on val */
      if (TagToHeadfunctor(item)!=functor_minus)
	{
	  HeapPush(h,functor_minus);
	  HeapPush(h,atom_nil);
	  HeapPush(h,item);
	  item = MakeStructure(h-3);
	}
      mask = IStep(MASK_VAL),
      s = TagToArg(queues,7);
    }
  CTagToArg(queues,1) += IStep(1);
  CTagToArg(queues,2) |= mask;
  HeapPush(h,item);
  HeapPush(h,*s);
  *s = MakeList(h-2);
  w->global_top = h;
}


/*** support for queues of indexicals and globals ***/

#define CLPFD_MUTABLE RefTerm(5) /* xref Emulator/sicstus.c */

void fd_sync MAGIC (HIDDEN_PROTO Argdecl)
{
  TAGGED ptr = RefMutable(CLPFD_MUTABLE) & -4L; /* clear GC bits */
  struct propagator *current = (struct propagator *)TermToStkPointerOrNull(ptr);
  struct propagator *cur = fd.current_propagator;

  while (cur && cur!=current) {
    int i;
    for (i=0; i<FD_NB_QUEUES; i++) {
      cur->queue[i].first = 0;
      cur->queue[i].last = 0;
    }
    fd.current_propagator = cur->next;
    cur->next = fd.free_propagators;
    fd.free_propagators = cur;
    cur = fd.current_propagator;
  }
  /* [MC 3.11.1] see SPRM 7785.  Shouldn't happen, but does.  Exact cause
     not understood.  Hypothesis: a combination of freeze and
     backtracking causes contents of CLPFD_MUTABLE to be outside the
     fd.current_propagator chain.  Rescue op: pop fd.free_propagators
     stack until found. */

  while (cur!=current) {
    cur = fd.free_propagators;
    fd.free_propagators = cur->next;
    cur->next = fd.current_propagator;
    fd.current_propagator = cur;
  }
}


/* '$fd_begin'
*/
void SPCDECL
prolog_fd_begin MAGIC (HIDDEN_PROTO_VOID)
{
  WAMENV;
  struct propagator *cur;
  int i;
  
  fd_sync(Arg);
  if (fd.free_propagators) {
    cur = fd.free_propagators;
    fd.free_propagators = cur->next;
  } else {
    cur = (struct propagator *)sp_checkalloc(sizeof(struct propagator), TRUE);
    for (i=0; i<FD_NB_QUEUES; i++) {
      cur->queue[i].first = 0;
      cur->queue[i].last = 0;
      cur->queue[i].size = 4;
      cur->queue[i].items = sp_checkalloc(4*sizeof(TAGGED), TRUE);
    }
  }
  cur->next = fd.current_propagator;
  fd.current_propagator = cur;
  update_mutable(w,StkPointerToTermOrZero(cur),CLPFD_MUTABLE);
}


/* '$fd_end'
*/
void fd_end MAGIC (HIDDEN_PROTO Argdecl)
{
  struct propagator *cur;

  fd_sync(Arg);
  cur = fd.current_propagator;
  fd.current_propagator = cur->next;
  cur->next = fd.free_propagators;
  fd.free_propagators = cur;
  update_mutable(w,StkPointerToTermOrZero(fd.current_propagator),CLPFD_MUTABLE);
}


void
fd_dealloc MAGIC (HIDDEN_PROTO_VOID)
{
  WAMENV;
  struct propagator *cur;
  
  RefMutable(CLPFD_MUTABLE) = TaggedZero;
  fd_sync(Arg);
  while ((cur=fd.free_propagators)) {
    int i;
    for (i=0; i<FD_NB_QUEUES; i++)
      sp_checkdealloc(cur->queue[i].items, cur->queue[i].size*sizeof(TAGGED), TRUE);
    fd.free_propagators = cur->next;
    sp_checkdealloc((TAGGED *)cur, sizeof(struct propagator), TRUE);
  }
}  


/* support function for stack shifting and gc */
/* defines copied from heapgc.c */

#define GC_FIRSTMASK 0x1
#define GC_MARKMASK 0x2

#define gc_UnmarkM(x)  ((x)&=(~GC_MARKMASK))

#define gc_PutValueFirst(p,x) Deposit(p,POINTERMASK|GC_FIRSTMASK,x)

#define intoRelocationChain(j,curr) \
(*(curr) = gc_PutValueFirst(*(j),*(curr)), \
 *(j)    = gc_PutValueFirst(AddrToTag(curr)|GC_FIRSTMASK,*(j)))

void SPCDECL
fd_manager_hook(SP_HOOKS_COOKIE_PROTO,Argdecl,int msg,TAGGED *ptr)
{
  long reloc;
  struct propagator *cur;
  TAGGED t, *p;
  int i, j;
  
  SP_HOOKS_COOKIE_USE;
  switch (msg) {
  case 1:			/* stack shifter */
    reloc = (long)ptr;
    fd_sync(Arg);
    cur = fd.current_propagator;
    while (cur) {
      for (j=0; j<FD_NB_QUEUES; j++) {
	int last = cur->queue[j].last;
	for (i=cur->queue[j].first; i!=last; ) {
	  cur->queue[j].items[i] += reloc;
	  i++;
	  if (i==cur->queue[j].size)
	    i = 0;
	}
      }
      cur = cur->next;
    }
    break;
  case 2:			/* gc, mark phase */
    fd_sync(Arg);
    cur = fd.current_propagator;
    while (cur) {
      for (j=0; j<FD_NB_QUEUES; j++) {
	int last = cur->queue[j].last;
	for (i=cur->queue[j].first; i!=last; ) {
	  markVariable(&cur->queue[j].items[i]);
	  i++;
	  if (i==cur->queue[j].size)
	    i = 0;
	}
      }
      cur = cur->next;
    }
    break;
  case 3:			/* gc, sweep phase */
    cur = fd.current_propagator;
    while (cur) {
      for (j=0; j<FD_NB_QUEUES; j++) {
	int last = cur->queue[j].last;
	for (i=cur->queue[j].first; i!=last; ) {
	  gc_UnmarkM(cur->queue[j].items[i]);
	  t = cur->queue[j].items[i];
	  p = TagToPointer(t);
	  if (OffHeaptop(p,ptr))
	    intoRelocationChain(p,&cur->queue[j].items[i]);
	  i++;
	  if (i==cur->queue[j].size)
	    i = 0;
	}
      }
      cur = cur->next;
    }
    break;
  }
}



struct daemon_frame {
  void (SPCDECL *destructor)(void *);
  void (SPCDECL *daemon)(HIDDEN_PROTO
			 void *,SP_globref,TAGGED); /* (handle,attr_ref,global) */
};


/* attaching a daemon */
/* Term = daemon(Global,AttrRef,StatusM,Ent,RawHandle) */
void attach_daemon MAGIC (HIDDEN_PROTO
			  void *handle,
			  SP_globref attr_ref, 
			  SP_globref var_ref, 
			  TAGGED global,
			  TAGGED list_functor,
			  int ar)
{
  TAGGED *h, thandle, var;
  WAMENV;

  X(ar) = global;
  RequireHeap(256+38+9,ar);	/* see prolog_fd_global_told */
  thandle = PointerToTerm(handle);
  h = w->global_top;
  h[0] = fd.functor_daemon5;
  h[1] = X(ar);
  h[2] = StkPointerToTermOrZero(attr_ref);
  h[3] = CTagToArg(X(ar),3);	/* status mutable */
  h[4] = CTagToArg(X(ar),4);	/* entailment variable */
  h[5] = thandle;
  w->global_top = h+6;
  var = RefGlob(var_ref);
  DerefSwitch(var,;);
  fd_link(w,var,list_functor,MakeStructure(h));
}

/* Term = daemon(Global,AttrRef,StatusM,Ent,RawHandle) */
static void run_all_daemons MAGIC (HIDDEN_PROTO_VOID)
{
  struct propagator *cur = fd.current_propagator;
  TAGGED t1;
  SP_globref attr_ref;
  struct daemon_frame *handle;

  while (cur->queue[4].first != cur->queue[4].last) {
    TAGGED item = cur->queue[4].items[cur->queue[4].first++];
    if (cur->queue[4].first==cur->queue[4].size)
      cur->queue[4].first = 0;
    t1 = CTagToArg(item,2);
    attr_ref = TermToStkPointerOrNull(t1);
    t1 = CTagToArg(item,5);
    glue_get_integer(t1,(long *)&handle);
    (*handle->daemon)(HIDDEN_ARG_COMMA
		      handle, attr_ref, CTagToArg(item,1));
  }
}


/* where=0x0/1 -- prepend/append to indexical queue
         0x2/3 -- prepend/append to 'val' global queue
         0x4/5 -- prepend/append to 'minmax' global queue
         0x6/7 -- prepend/append to 'dom' global queue
         0x8/9 -- prepend/append to daemon queue
*/
void fd_enqueue MAGIC (HIDDEN_PROTO
		       Argdecl,
		       TAGGED item, int where)
{
  struct propagator *cur = fd.current_propagator;
  int ix = where>>1;
  int i, pop, size;
  
  /* mark as enqueued, STATUS FIELD */
  if (ix<4) {
    TAGGED mutable = CTagToArg(item,3);
    
    update_mutable(w,RefMutable(mutable)|IStep(1),mutable);
  }
  
  /* room for item? */
  pop = cur->queue[ix].last-cur->queue[ix].first;
  size = cur->queue[ix].size;
  if (pop<0)
    pop += size;
  if (pop+1 >= size) { /* grow */
    cur->queue[ix].items = sp_checkrealloc(cur->queue[ix].items,
					   size*sizeof(TAGGED),
					   size*sizeof(TAGGED)<<1, TRUE);
    cur->queue[ix].size += size;
    if (cur->queue[ix].first>cur->queue[ix].last) {
      for (i=size-1; i>=cur->queue[ix].first; i--)
	cur->queue[ix].items[i+size] = cur->queue[ix].items[i];
      cur->queue[ix].first += size;
    }
    size += size;
  }
  if (where & 0x1) {
    int pos = cur->queue[ix].last;
    
    cur->queue[ix].items[pos++] = item;
    if (pos==size) pos = 0;
    cur->queue[ix].last = pos;
  } else {
    int pos = cur->queue[ix].first;
    
    if (pos==0) pos = size;
    cur->queue[ix].items[--pos] = item;
    cur->queue[ix].first = pos;
  }
}

/* enqueue if not already in the queue */
/* where is as above */
void fd_enqueue_global MAGIC (HIDDEN_PROTO
			      Argdecl,
			      TAGGED item, int where)
{
  TAGGED t1 = CTagToArg(item,3); /* status mutable */
  TAGGED status = RefMutable(t1);
  
  if (!(status & IStep(1))) {
    fd_enqueue(w,item,where); /* DOM or MINMAX */
  }
}


/* 0 -- empty queue
   1 -- dequeued indexical
   2 -- dequeued global
*/
int fd_dequeue MAGIC (HIDDEN_PROTO TAGGED *item)
{
  struct propagator *cur = fd.current_propagator;
  TAGGED t1;
  int i;

 retry:
  for (i=0; i<FD_NB_QUEUES-1; i++)
    if (cur->queue[i].first != cur->queue[i].last) {
      int rc = (i==0 ? 1 : 2);
      int pos = cur->queue[i].first;
      
      *item = t1 = cur->queue[i].items[pos++];
      if (pos==cur->queue[i].size) pos = 0;
      cur->queue[i].first = pos;
      t1 = CTagToArg(t1,4);	/* Ent, always a HVA */
      DerefSwitch(t1,return rc;); /* skip if (dis)entailed */
      goto retry;
    }
  return 0;
}

static void fd_enqueue_list_gc MAGIC (HIDDEN_PROTO
				      struct worker *w,
				      int index,
				      TAGGED filter, TAGGED lists_loc)
{
  TAGGED *lists = TagToArg(RefMutable(lists_loc),0);
  TAGGED list, new, item, status, prev=0, t1, *h;
  int gc=0, decr=0;
  LetShadowregs;

  list = lists[index];
  new = list;
  while (TagIsLST(list))
    {
      item = CTagToCar(list);
      DerefArg(t1,item,4);
      if (!IsVar(t1))		/* entailed */
	{
	  list = CTagToCdr(list);
	  if (prev==0)
	    decr++, new = list;
	  else if (TagToLST(prev) >= GLOBAL_UNCOND)
	    decr++, CTagToCdr(prev) = list;
	  else
	    decr++, gc++;
	}
      else
	{
	  prev = list;
	  list = CTagToCdr(list);
	}
      if (!IsVar(t1) || t1==filter)
	;
      else if (TagToHeadfunctor(item)==fd.functor_daemon5)
	{
	  t1 = CTagToArg(item,3); /* status mutable */
	  if (!(RefMutable(t1) & IStep(8)))
	    fd_enqueue(w, item, 0x9);
	}
      else
	{
	  t1 = CTagToArg(item,3); /* status mutable */
	  status = RefMutable(t1);
	  if (!(status & IStep(1)))
	    {
	      if (TagToHeadfunctor(item)==fd.functor_ix7)
		fd_enqueue(w, item, 0x1);
	      else
		fd_enqueue(w, item, (index==3 ? 0x7 : 0x5)); /* DOM or MINMAX */
	    }
	}
    }
  h = w->global_top;
  if (decr>0)
    {
      TAGGED next, *nextp = &next;
      
      while (gc>0)
	{
	  item = CTagToCar(new);
	  new = CTagToCdr(new);
	  DerefArg(t1,item,4);
	  if (!IsVar(t1))		/* entailed */
	    --gc;
	  else
	    {
	      *nextp = MakeList(h);
	      HeapPush(h,item);
	      nextp = h++;
	    }
	}
      *nextp = new;
      /* next is the new list, decrement count by decr */
      if (lists < GLOBAL_UNCOND)/* can't smash */
	{
	  int i;
	  
	  for (i=0; i<8; i++)
	    *h++ = lists[i];
	  lists = h-8;
	  update_mutable(w,MakeStructure(lists),lists_loc);
	}
      lists[1] -= IStep(decr);
      lists[index] = next;
      if (next==atom_nil)
	{
	  index = GetSmall(lists[2]) & MASK_VAL;
	  if (TagIsLST(lists[3])) index |= MASK_DOM;
	  if (TagIsLST(lists[4])) index |= MASK_MIN;
	  if (TagIsLST(lists[5])) index |= MASK_MAX;
	  if (TagIsLST(lists[6])) index |= MASK_MINMAX;
	  lists[2] = MakeSmall(index);
	}
    }
  w->global_top = h;
}


static void fd_enqueue_list MAGIC (HIDDEN_PROTO
				   Argdecl,
				   int index,
				   TAGGED filter, TAGGED lists)
{
  TAGGED list, item, status, t1;

  list = CTagToArg(lists,index);
  while (TagIsLST(list))
    {
      item = CTagToCar(list);
      list = CTagToCdr(list);
      DerefArg(t1,item,4);
      if (!IsVar(t1) || t1==filter)
	;
      else if (TagToHeadfunctor(item)==fd.functor_daemon5)
	{
	  t1 = CTagToArg(item,3); /* status mutable */
	  if (!(RefMutable(t1) & IStep(8)))
	    fd_enqueue(w, item, 0x8);
	}
      else
	{
	  t1 = CTagToArg(item,3); /* status mutable */
	  status = RefMutable(t1);
	  if (!(status & IStep(1)))
	    {
#if DEPTH_FIRST_STRATEGY
	      if (TagToHeadfunctor(item)==fd.functor_ix7)
		fd_enqueue(w, item, 0x0);
	      else
		fd_enqueue(w, item, (index==3 ? 0x6 : 0x4)); /* DOM or MINMAX */
#else
	      if (TagToHeadfunctor(item)==fd.functor_ix7)
		fd_enqueue(w, item, 0x1);
	      else
		fd_enqueue(w, item, (index==3 ? 0x7 : 0x5)); /* DOM or MINMAX */
#endif
	    }
	}
    }
}


static int fd_enqueue_val MAGIC (HIDDEN_PROTO
				 struct worker *w,
				 int index,
				 TAGGED filter,
				 TAGGED lists,
				 TAGGED *stack)
{
  TAGGED list0, list, pair, item, functor, vars, var, t1, *h;
  int n=0;
  int iff_count=0;
  LetShadowregs;

  list = CTagToArg(lists,index);
  list0 = list;
  h = w->global_top;
  while (TagIsLST(list))
    {
      pair = CTagToCar(list);
      list = CTagToCdr(list);
      vars = CTagToArg(pair,1); DerefNonvar(vars);
      item = CTagToArg(pair,2); DerefNonvar(item);
      functor = TagToHeadfunctor(item);
      DerefArg(t1,item,4);
      if (IsVar(t1) && t1!=filter)
	{
	  var = atom_nil;
	  while (TagIsLST(vars) && !IsVar(var))
	    {
	      DerefCar(var,vars);
	      DerefCdr(vars,vars);
	    }
	  if (IsVar(var))
	    {			/* fd_val_link(t1, vars, item) later */
	      n++;
	      iff_count += (functor==fd.functor_iff4);
	      HeapPush(h,pair);
	      HeapPush(h,*stack);
	      *stack = MakeList(h-2);
	    }
	  else if (functor==fd.functor_daemon5)
	    {
	      t1 = CTagToArg(item,3); /* status mutable */
	      if (!(RefMutable(t1) & IStep(8)))
		fd_enqueue(w, item, 0x8);
	    }
	  else if (functor==fd.functor_ix7)
	    {
	      fd_enqueue(w, item, 0x0);
	    }
	  else if (functor==fd.functor_iff4)
	    {			/* iff(Ix,B,Key,A) */
	      TAGGED bvar, key;

	      iff_count++;
	      bvar = CTagToArg(item,2);
	      key = CTagToArg(item,3);
	      DerefNonvar(bvar);
	      if (bvar==key)
		{
		  DerefArg(item,item,1);
		  fd_enqueue(w, item, 0x0);
		}
	    }
	  else
	    {
	      t1 = CTagToArg(item,3); /* status mutable */
	      var = RefMutable(t1);
	      if (!(var & IStep(1))) /* can be in queue if global */
		{
		  fd_enqueue(w, item, 0x2);
		}
	    }
	}
      if (iff_count==0)
	list0 = list;
    }
  w->global_top = h;
  /* now disable all iff(Ix,B,Key,A) items we encountered */
  while (TagIsLST(list0))
    {
      pair = CTagToCar(list0);
      list0 = CTagToCdr(list0);
      item = CTagToArg(pair,2); DerefNonvar(item);
      DerefArg(t1,item,4);
      if (IsVar(t1) && t1!=filter && TagToHeadfunctor(item)==fd.functor_iff4)
	{			/* iff(Ix,B,Key,A) */
	  TAGGED key = CTagToArg(item,3);

	  BindHVA(t1,key);
	}
    }
  return n;
}


/* assuming bits > 0 */
/* X(1), X(2) must survive over this */
void fd_enqueue_all MAGIC (HIDDEN_PROTO
			   Argdecl,
			   int bits,
			   TAGGED filter, TAGGED lists_loc)
{
  TAGGED t1;
  int nsusp;
  
				/* each head and tail of each list 
				   is dereferenced */
  t1 = RefMutable(lists_loc);	/* get suspension lists */
  t1 = CTagToArg(t1,1);		/* get suspension count */
  nsusp = GetSmall(t1);
  X(3) = filter;
  X(4) = lists_loc;
  RequireHeap((nsusp<<1)+8, 5); /* 2 * #suspensions + lists/7 */
  filter = X(3);
  lists_loc = X(4);
  if (bits & MASK_SINGLETON)
    {
      TAGGED lists = RefMutable(lists_loc);
      int n=0;

      /* 3.9: enqueue all constraints when ground,
	 (a) to maximize entailment detection,
	 (b) to handle co-references */
      bits |= GetSmall(CTagToArg(lists,2)); /* bitmask of susp. lists */
      X(3) = atom_nil;		/* accumulator for fd_enqueue_val */
      /* we are about to prepend---prepend slowest first! */
#if DEPTH_FIRST_STRATEGY
      if (bits & MASK_DOM)
	fd_enqueue_list(w, 3, filter, lists);
      if (bits & MASK_MIN)
	fd_enqueue_list(w, 4, filter, lists);
      if (bits & MASK_MAX)
	fd_enqueue_list(w, 5, filter, lists);
      if (bits & MASK_MINMAX)
	fd_enqueue_list(w, 6, filter, lists);
#else
      if (bits & MASK_MINMAX)
	fd_enqueue_list(w, 6, filter, lists);
      if (bits & MASK_MAX)
	fd_enqueue_list(w, 5, filter, lists);
      if (bits & MASK_MIN)
	fd_enqueue_list(w, 4, filter, lists);
      if (bits & MASK_DOM)
	fd_enqueue_list(w, 3, filter, lists);
#endif
      if (bits & MASK_VAL)
	n = fd_enqueue_val(w, 7, filter, lists, &X(3));
      if (n>0)
	{
	  TAGGED pair, var, vars, item;
	  TAGGED *h;

	  RequireHeap(13*n, 4);
	  while (TagIsLST(X(3)))
	    {
	      pair = CTagToCar(X(3));  
	      X(3) = CTagToCdr(X(3)); 
	      vars = CTagToArg(pair,1); DerefNonvar(vars);
	      item = CTagToArg(pair,2); DerefNonvar(item);
			    /* unwrap any iff(Ix,B,Key,A) item */
	      if (TagToHeadfunctor(item)==fd.functor_iff4)
		{
		  TAGGED bvar, key;

		  bvar = CTagToArg(item,2);
		  key = CTagToArg(item,3);
		  DerefNonvar(bvar);
		  if (bvar!=key)
		    continue;
		  DerefArg(item,item,1);
		}

	      do
		{
		  DerefCar(var,vars);
		  DerefCdr(vars,vars);
		}
	      while (!IsVar(var));
	      h = w->global_top;
	      HeapPush(h,functor_minus);
	      HeapPush(h,vars);
	      HeapPush(h,item);
	      w->global_top = h;
	      fd_link(w, var, fd.functor_val, MakeStructure(h-3));
	    }
	}
    }
  else
    {
      if (bits & MASK_MINMAX)
	fd_enqueue_list_gc(w, 6, filter, lists_loc);
      if (bits & MASK_MAX)
	fd_enqueue_list_gc(w, 5, filter, lists_loc);
      if (bits & MASK_MIN)
	fd_enqueue_list_gc(w, 4, filter, lists_loc);
      if (bits & MASK_DOM)
	fd_enqueue_list_gc(w, 3, filter, lists_loc);
    }
  run_all_daemons();
}


/* implies $fd_begin */
void SPCDECL
prolog_fd_global_enqueue MAGIC (HIDDEN_PROTO SP_term_ref TermR)
{
  WAMENV;
  TAGGED term=RefTerm(TermR);
  TAGGED mutable;

  SP_MANGLE(prolog_fd_begin)(HIDDEN_ARG);
  DerefNonvar(term);
  mutable = CTagToArg(term,3);
  update_mutable(w,RefMutable(mutable)|IStep(9),mutable); /* STATUS FIELD, disable resumption */
  /*** done in Prolog
  fd_sync(Arg);
  fd_enqueue(w, term, 0x3);
  ***/
}


/* implies $fd_begin */
void SPCDECL
prolog_fd_enqueue_all MAGIC (HIDDEN_PROTO SP_term_ref ListsM)
{
  WAMENV;
  TAGGED t1;
  int bits;

/*    X(0) = RefTerm(ListsM); */
  (void)ListsM;
  SP_MANGLE(prolog_fd_begin)(HIDDEN_ARG);
  DerefNonvar(X(0));
  t1 = RefMutable(X(0));	/* get suspension lists */
  bits = GetSmall(CTagToArg(t1,2)); /* get filter */
  if (bits > 0) {
    fd_sync(Arg);
    fd_enqueue_all(w,bits+MASK_SINGLETON,atom_nil,X(0));
  }
}

/* Support for managing incumbents:
   '$fd_update_incumbent'(+Ptr, +Value, +Vertex).
	store a new incumbent vertex and value
   '$fd_incumbent_bound'(+Ptr, -Value).
	retrieve the current incumbent value.
*/

void SPCDECL
prolog_fd_update_incumbent MAGIC (HIDDEN_PROTO
				  void *vins,
				  SP_term_ref ValueR,
				  SP_term_ref VertexR)
{
  WAMENV;
  struct instance *ins = vins;
  int i, no_cells;
  TAGGED vertex;

  no_cells = ((ins->objsize-sizeof(struct instance))>>LogSizeOfWord) + 1;
  /* value is at offset no_cells-4
     vertex is at offset no_cells-7, -9, ...
  */
  DEREF(ins->code[no_cells-4],RefTerm(ValueR));
  DEREF(vertex,RefTerm(VertexR));
  for (i=no_cells-7; TagIsLST(vertex); i-=2) {
    DerefCar(ins->code[i],vertex);
    DerefCdr(vertex,vertex);
  }
}


void SPCDECL
prolog_fd_incumbent_bound MAGIC (HIDDEN_PROTO
				 void *vins,
				 SP_term_ref ValueR)
{
  WAMENV;
  struct instance *ins = vins;
  int no_cells = ((ins->objsize-sizeof(struct instance))>>LogSizeOfWord) + 1;
  /* value is at offset no_cells-4
     vertex is at offset no_cells-7, -9, ...
  */
  RefTerm(ValueR) = ins->code[no_cells-4];
}


void SPCDECL prolog_fd_minint_maxint MAGIC (HIDDEN_PROTO
					    long *minp, long *maxp)
{
  (void)HIDDEN_ARG_OR_ZERO;
  *minp = -HighInt;
  *maxp = HighInt-1L;
}


/* Heap routines from Cormen et al. */

void heap_init(struct heap *h)
{
  h->size = 0;
}


void heap_insert(struct heap *h,
		 void *item,
		 HeapFun cmpfun)
{
  int i = ++h->size;

  while (i>1 && (*cmpfun)(h->item[i>>1],item) > 0) {
    h->item[i] = h->item[i>>1];
    i >>= 1;
  }
  h->item[i] = item;
}


/* Repair the heap property
   i.e. h->item[i>>1].key <= h->item[i].key
*/
void heapify(struct heap *h,
	     HeapFun cmpfun)
{
  int i = 1;

  for (;;) {
    int l = i<<1;
    int r = l+1;
    int smallest;
    
    if (l <= h->size && (*cmpfun)(h->item[l],h->item[i]) < 0)
      smallest = l;
    else
      smallest = i;
    if (r <= h->size && (*cmpfun)(h->item[r],h->item[smallest]) < 0)
      smallest = r;
    if (smallest != i) {
      void *tmp = h->item[i];
      
      h->item[i] = h->item[smallest];
      h->item[smallest] = tmp;
      i = smallest;
    } else
      break;
  }
}


void *heap_extract_min(struct heap *h,
		       HeapFun cmpfun)
{
  void *it = h->item[1];

  if (h->size==0)
    return NULL;
  h->item[1] = h->item[h->size];
  --h->size;
  heapify(h,cmpfun);

  return it;
}

/* Support for Palloc/Pfree. */

void *fd_perm_alloc MAGIC (HIDDEN_PROTO
			   Argdecl,
			   int nbytes,
			   TAGGED handle) /* HVA to bind */
{
  void *ptr = SP_malloc(nbytes);
  TAGGED tptr = PointerToTerm(ptr);
  TAGGED *h = w->global_top; 
  TAGGED inner, outer;
  LetShadowregs;

  inner = MakeStructure(h);
  *h++ = fd.functor_free;
  *h++ = tptr;
  outer = MakeList(h);
  Load0HVA(h);
  *h++ = inner;
  w->global_top = h;
  TrailPushCheck(outer);
  BindHVA(handle,outer);
  return ptr;
}


void *fd_perm_data MAGIC (HIDDEN_PROTO
			  TAGGED handle) /* [Flag | '$free'(Ptr)] */
{
  void *ptr;
  
  handle = CTagToCdr(handle);
  handle = CTagToArg(handle,1);
  /* ptr = TermToPointer(handle); */
  glue_get_integer(handle,(long *)&ptr);
  return ptr;
}


/* A mutable item on the trail is redundant if:
   1. the mutable is in H+, or
   2. there is a previous item in TR+, or
   3. the new and old values are the same.
   */
#define RedundantMutableItem(Mutp,Item) \
((Mutp) >= GLOBAL_UNCOND || (Item)[2] >= TrailToInt(w->node->trail_top) || (Mutp)[1]==(Item)[1])

#define PrevMutItem(Item) \
(w->trail_start + GetSmall((Item)[2]))

/* Compress the trail from tr and up, removing all zeroes.
   Any mutable timestamps pointing into the compressed region
   must also be updated.
   RIPPED OFF FROM EMULATOR.
*/
static void compress_trail MAGIC (HIDDEN_PROTO Argdecl,
				  TAGGED *tr)
{
  TAGGED *trlim = w->trail_top;
  TAGGED *h = tr;
  TAGGED ref, *mutstr, *mutitem;

  while (TrailYounger(trlim,tr)) {
    ref = tr[0];
    if (TagIsSTR(ref) && TagToHeadfunctor(ref)==fd.functor_mutable) {
      /* About to move a mutable item.  First, remove any redundant items. */
      mutstr = TagToSTR(ref);
      mutitem = PrevMutItem(mutstr);
      while (RedundantMutableItem(mutstr,mutitem) && mutitem >= tr) {
	mutstr[2] = mutitem[2];
	mutitem[0] = mutitem[1] = mutitem[2] = 0;
	mutitem = PrevMutItem(mutstr);
      }
      if (mutitem >= tr) {	/* some mutitems still in the compressed region */
	while (mutstr[2] != TrailToInt(tr))
	  mutstr = PrevMutItem(mutstr);
	mutstr[2] -= IStep(tr-h);	/* = TrailToInt(h);  */
	h[0] = ref;
	h[1] = tr[1];
	h[2] = tr[2];
	h+=3;
      }
      tr+=3;
    } else {
      if (ref) 
	*h++ = ref;
      tr++;
    }
  }
  w->trail_top = h;
}

void fd_perm_free MAGIC (HIDDEN_PROTO
			 Argdecl,
			 TAGGED handle)	/* [Flag | '$free'(Ptr)] */
{
  struct node *nd;
  TAGGED *ptr;
  void (SPCDECL *destructor)(void *);
  ANYPOINTER frame;
  
  DerefHeapSwitch(handle,;);
  ptr = TagToLST(handle);
  for (nd = w->node;
       ptr < nd->global_top;
       nd = ChoiceptPrevious(nd))
    ;
  if (nd==w->node) {
    for (ptr=w->trail_top; ptr[-1]!=handle; ptr--)
      ;
    ptr[-1] = 0;		/* for compress_trail */
    compress_trail(w,ptr-1);
    /* frame = TermToPointer(CTagToArg(CTagToCdr(handle),1)); */
    glue_get_integer(CTagToArg(CTagToCdr(handle),1),(long *)&frame);
    destructor = *(void (SPCDECL **)(void*))frame;
    (*destructor)(frame);
  } else {
    for (;
	 ChoiceYounger(nd,w->choice_start) && !ChoiceptTestCleanup(nd);
	 nd = ChoiceptPrevious(nd))
      ChoiceptMarkCleanup(nd);
    *ptr = TaggedOne;
    TrailPushCheck(TagHVA(ptr));
  }
}    

/* Most propagators have arguments (+State0, -State), ... where
   State0 = F(......,Handle,Stamp), is left dereferenced,
   State  = copy of State0 with Stamp incremented.
   Also, check if this execution step can be backtracked over or not.
*/
TAGGED unify_output_state MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                                 Argdecl,
                                 TAGGED *phandle,
                                 long *pstamp,
                                 BOOL *pcommitted)
{
  TAGGED handle, t1, *s, *h;
  int ar, i;
  LetShadowregs;
  
  DerefNonvar(X(0));
  ar = Arity(TagToHeadfunctor(X(0)));
  DerefArg(handle,X(0),ar-1);
  *phandle = handle;
  DerefArg(t1,X(0),ar);
  *pstamp = GetSmall(t1);
  s = TagToArg(X(0),0);
  if (s >= GLOBAL_UNCOND) {
    s[ar] += IStep(1);		/* increment stamp */
  } else {
    h = w->global_top;
    for (i=0; i<ar+1; i++)
      h[i] = s[i];
    w->global_top = h+ar+1;
    h[ar] += IStep(1);		/* increment stamp */
    X(0) = MakeStructure(h);
  }
  *pcommitted = (IsVar(handle) ? TRUE : TagToLST(handle) >= GLOBAL_UNCOND);
  return X(0);
}

