/**************************************************************************
 * Filename:	perl.c
 * Author:	Jesper Eskilson <jojo@sics.se>
 * Created:	Thu Feb 18 18:39:16 1999
 * Description: Foreign resource C-code for library(perl)
 * CVS:         $Id: perl.c,v 1.1 1999/09/03 14:00:12 jojo Exp $
 * Last-Update:	Time-stamp: <1999-03-06 1156 jojo>
 **************************************************************************
 */

#include <EXTERN.h>
#include <perl.h>

#include "sicstus.h"

static PerlInterpreter *my_perl;

void perl_init() 
{
  char *embedding[] = { "", "-e", "0" };

  my_perl = perl_alloc();
  perl_construct( my_perl );
  
  perl_parse(my_perl, NULL, 3, embedding, NULL);
  perl_run(my_perl);
  
  /** Treat $a as an integer **/
  perl_eval_pv("$a = 3; $a **= 2", TRUE);
  sp_ttyprintf("a = %d\n", SvIV(perl_get_sv("a", FALSE)));
  
  /** Treat $a as a float **/
  perl_eval_pv("$a = 3.14; $a **= 2", TRUE);
  sp_ttyprintf("a = %f\n", SvNV(perl_get_sv("a", FALSE)));
  
  /** Treat $a as a string **/
  perl_eval_pv("$a = 'rekcaH lreP rehtonA tsuJ'; $a = reverse($a);", TRUE);
  sp_ttyprintf("a = %s\n", SvPV(perl_get_sv("a", FALSE), na));
}

void perl_deinit()
{
  perl_destruct(my_perl);
  perl_free(my_perl);
  perl_construct(my_perl);
}

/** my_perl_eval_sv(code, error_check)
** kinda like perl_eval_sv(),
** but we pop the return value off the stack
**/
SV* sp_my_perl_eval_sv(SV *sv, I32 croak_on_error)
{
  dSP;
  SV* retval;

  PUSHMARK(SP);
  perl_eval_sv(sv, G_SCALAR);

  SPAGAIN;
  retval = POPs;
  PUTBACK;

  if (croak_on_error && SvTRUE(GvSV(errgv)))
    croak(SvPVx(GvSV(errgv), na));

  return retval;
}

/** match(string, pattern)
**
** Used for matches in a scalar context.
**
** Returns 1 if the match was successful; 0 otherwise.
**/

I32 sp_match(SV *string, char *pattern)
{
  SV *command = NEWSV(1099, 0), *retval;

  sv_setpvf(command, "my $string = '%s'; $string =~ %s",
	    SvPV(string,na), pattern);

  retval = my_perl_eval_sv(command, TRUE);
  SvREFCNT_dec(command);

  return SvIV(retval);
}

/** substitute(string, pattern)
**
** Used for =~ operations that modify their left-hand side (s/// and tr///)
**
** Returns the number of successful matches, and
** modifies the input string if there were any.
**/

I32 sp_substitute(SV **string, char *pattern)
{
  SV *command = NEWSV(1099, 0), *retval;
  
  sv_setpvf(command, "$string = '%s'; ($string =~ %s)",
	    SvPV(*string,na), pattern);
  
  retval = my_perl_eval_sv(command, TRUE);
  SvREFCNT_dec(command);
  
  *string = perl_get_sv("string", FALSE);
  return SvIV(retval);
}

/** matches(string, pattern, matches)
**
** Used for matches in an array context.
**
** Returns the number of matches,
** and fills in **matches with the matching substrings
**/

I32 sp_matches(SV *string, char *pattern, AV **match_list)
{
  SV *command = NEWSV(1099, 0);
  I32 num_matches;

  sv_setpvf(command, "my $string = '%s'; @array = ($string =~ %s)",
	    SvPV(string,na), pattern);

  my_perl_eval_sv(command, TRUE);
  SvREFCNT_dec(command);

  *match_list = perl_get_av("array", FALSE);
  num_matches = av_len(*match_list) + 1; /** assume $[ is 0 **/

  return num_matches;
}






