/* Copyright (C) 1996, Swedish Institute of Computer Science. */

#define SP_MSDOS __DJGPP__

#ifdef __WATCOMC__
#ifdef __NT__
#define SP_WIN32 1
#endif
#ifdef __OS2__
#define SP_OS2 1
#endif
#endif

#ifdef __BORLANDC__
#define SP_WIN32 1
#endif

#ifdef _MSC_VER
#define SP_WIN32 1
#endif

#ifdef __WIN32__
#define SP_WIN32 1
#endif

#if __EMX__
#define SP_OS2 1
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#if SP_WIN32
#include <windows.h>
#endif
#include <sicstus/sicstus.h>

#ifndef _MAX_PATH
#define _MAX_PATH 260
#endif

static int breakargs(char *program, char *line, char **argv);
static void spexit(char *call, char *msg);
static char *exception_to_str(void);

#ifdef SPRTW
int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInst,
		   LPSTR lpszCmdLine, int nCmdShow)
#else
void main(int argc, char **argv)    
#endif
{
  SP_pred_ref pred;
  char *p, buf[_MAX_PATH], *bootpath;
#ifdef SPRTW
  int argc;
  char *argv[128];

  GetModuleFileName(hInstance, buf, sizeof(buf));
  argc = breakargs(buf, lpszCmdLine, argv);
#else
  strcpy(buf, argv[0]);
#endif
#if SP_WIN32
  bootpath = NULL;
#else
  bootpath = buf;
#endif
  p = buf + strlen(buf);
  while (*p != '\\')
    p--;
  *p = '\0';

  #if SP_INHIBIT_IMPLICIT_APIPROCINIT /* [PM] 3.10.2 set to work around HPUX cc bug */
  SetupSICStusDISPATCH();
  #endif/* SP_INHIBIT_IMPLICIT_APIPROCINIT */

  if (SP_initialize(argc, argv, bootpath) != SP_SUCCESS)
    spexit("Initialization failed", SP_error_message(SP_errno));

  strcat(buf, "\\main.sav");
  switch (SP_restore(buf))
    {
    case SP_SUCCESS: break;
    case SP_FAILURE: 
      spexit("Restoring main.sav", "failed");
    case SP_ERROR:
      spexit("Restoring main.sav raised exception", exception_to_str());
    };

  if ((pred = SP_predicate("main",0,"")) == NULL)
    spexit("Looking up main/0 failed", SP_error_message(SP_errno));

  switch (SP_query(pred))
    {
    case SP_SUCCESS: exit(0);
    case SP_FAILURE: 
      spexit("Running main", "failed");
    case SP_ERROR:
      spexit("Running main raised exception", exception_to_str());
    };
}

#ifdef SPRTW
static int breakargs(char *program, char *line, char **argv)
{
  int argc = 1;
  argv[0] = program;

  while(*line)
    { 
      while(*line && isspace(*line))
	line++;

      if (*line == '"')		/* quoted arguments */
	{
	  char *start = line+1;
	  char *end = start;

	  while( *end && *end != '"' )
	    end++;
	  if (*end == '"')
	    {
	      *end = '\0';
	      argv[argc++] = start;
	      line = end+1;
	      continue;
	    }
	}

      if (*line)
	{
	  argv[argc++] = line;
	  while(*line && !isspace(*line))
	    line++;
	  if (*line)
	    *line++ = '\0';
	}
    }
  argv[argc] = NULL;		/* add trailing NULL pointer to argv */
  return argc;
}
#endif

static void spexit(char *str, char *msg)
{
#ifdef SPRTW
  char buf[_MAX_PATH];

  sprintf(buf, "%s: %s\n", str, msg);
  MessageBox(NULL, buf, "sprtw", MB_OK|MB_TASKMODAL);
#else
  fprintf(stderr, "%s: %s\n", str, msg);
#endif
  exit(EXIT_FAILURE);           /* [PM] 3.9 was -1 (EXIT_FAILURE == 1) */
}

/*----------------------------------------------------------------------*/
/* All this just to print the exception ! */

struct open_chars {
  char *chars;
  int index;
  int size;
};

#define INIT_BUFSIZE 512

static int SPCDECL chputc(c, buf)
     int c;
     struct open_chars *buf;
{
  if (buf->index == buf->size)
    {
      buf->size *= 2;
      buf->chars = (char *)SP_realloc(buf->chars, buf->size);
    }
  return (buf->chars[buf->index++] = c);
}

static char *empty_str = "";

static char *exception_to_str(void)
{
  SP_pred_ref pred;
  SP_term_ref
    tExcp = SP_new_term_ref(),
    tStreamC = SP_new_term_ref(),
    tStream = SP_new_term_ref();
  struct open_chars *buf;
  SP_stream *stream;

  if (!(SP_exception_term(tExcp) &&
	(buf=(struct open_chars *)SP_malloc(sizeof(struct open_chars))) &&
	SP_make_stream(buf, NULL, chputc, NULL, NULL, NULL, NULL, &stream) &&
	(buf->chars=(char *)SP_malloc(INIT_BUFSIZE))))
    return empty_str;

  buf->size = INIT_BUFSIZE;
  buf->index = 0;
  SP_put_address(tStreamC, (void *)stream);
  SP_put_variable(tStream);
  if (!((pred = SP_predicate("stream_code", 2, "")) &&
	SP_query(pred, tStream, tStreamC) == SP_SUCCESS &&
	(pred = SP_predicate("write", 2, "")) &&
	SP_query_cut_fail(pred, tStream, tExcp) == SP_SUCCESS))
    return empty_str;

  buf->chars[buf->index++] = '\0';
  return buf->chars;
}
